/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <tulip/PropertyAlgorithm.h>
#include <tulip/BooleanProperty.h>
#include <tulip/StringProperty.h>
#include <tulip/ColorProperty.h>
#include <tulip/LayoutProperty.h>
#include <tulip/SizeProperty.h>

#include <portgraph/PortGraphRule.h>
#include <portgraph/PortGraphModel.h>
#include <portgraph/Bridge.h>
#include <portgraph/Trace.h>
#include <portgraph/PorgyTlpGraphStructure.h>

#include <boost/foreach.hpp>

using namespace std;
using namespace tlp;

const static std::string SAVEID="saveId";

static node findNode(const int n, IntegerProperty *prop, Graph *g) {
    node n2;
    forEach(n2, g->getNodes()) {
        if(prop->getNodeValue(n2)==n)
            break;
    }
    return n2;
}

class ApplyRule:public tlp::Algorithm {

public:

    PLUGININFORMATION(PorgyConstants::APPLY_RULE, "Bruno Pinaud", "09/04/10", "Comments", "1.0.1", "Porgy_internal")
    ApplyRule(const PluginContext *context):Algorithm(context) {
        addDependency(PorgyConstants::REDRAW_GRAPH, "1.0");
        addDependency(PorgyConstants::REDRAW_TRACE, "1.0");

        addInParameter<bool>(PorgyConstants::applyConnectedComponentPacking, "", "true");
        addInParameter<bool>(PorgyConstants::useM, "", "true");
        addInParameter<BooleanProperty>(PorgyConstants::POSITION, "", "", false);
        addInParameter<BooleanProperty>(PorgyConstants::BAN,"", "", false);
        addInParameter<bool>(PorgyConstants::isCondition, "", "false");
        addInParameter<bool>(PorgyConstants::layoutNewModel, "", "true");
        addInParameter<bool>("isProba", "", "false");
        addOutParameter<Graph *>(PorgyConstants::newModel, "","",false);
    }

    bool check(string &errorMsg) {
        return (PorgyTlpGraphStructure::isPorgyGraph(graph, errorMsg)&&PorgyTlpGraphStructure::isModelGraph(graph->getSuperGraph()));
    }

    ///////////////
    //graph => l'instance du membre gauche de la règle à appliquer
    bool run() {
        pluginProgress->showPreview(false);
        pluginProgress->setComment("rewriting graph " + graph->getSuperGraph()->getName());

        Graph *old_graph = graph->getSuperGraph();

        //Est-ce que cette instance de la règle a déjà été appliquée avec les mêmes paramètres
        Trace traceobj(PorgyTlpGraphStructure::getMainTrace(graph));
        string new_model_name = traceobj.existEdge(old_graph->getName(), graph->getName());
        if(!new_model_name.empty()) {
            dataSet->set(PorgyConstants::newModel, PorgyTlpGraphStructure::findModel(graph, new_model_name));
            cerr << __PRETTY_FUNCTION__ << ": Rule already applied before!" << endl;
            return true;
        }

        Graph *new_graph = PorgyTlpGraphStructure::clone(old_graph, dataSet);
        Graph *_rule = PorgyTlpGraphStructure::findRule(graph,PorgyTlpGraphStructure::getRuleNameForSubGraph(graph));
        assert(_rule!=NULL);
        //the saveId property is used to save the id of the RHS nodes to have a linked between RHS and the graph being rewritten
        Observable::holdObservers();
        IntegerProperty *saveId=_rule->getLocalProperty<IntegerProperty>(SAVEID);
        node n;
        forEach(n, _rule->getNodes()) {
            saveId->setNodeValue(n, n.id);
        }
        edge e;
        forEach(e, _rule->getEdges()) {
            saveId->setEdgeValue(e, e.id);
        }

        PortGraphRule PGMrule(_rule);
        BooleanProperty newelt(_rule);
        newelt.setAllNodeValue(false);
        newelt.setAllEdgeValue(false);
        //select all elements of RHS
        forEach(n, _rule->getNodes()) {
            if(PortGraphRule::isMember(n, _rule, PorgyConstants::SIDE_RIGHT))
                newelt.setNodeValue(n,true);
        }
        forEach(e, _rule->getEdges()) {
            if(PortGraphRule::isMember(e, _rule, PorgyConstants::SIDE_RIGHT))
                newelt.setEdgeValue(e,true);
        }
        BooleanProperty *newmodel_newelt = new_graph->getLocalProperty<BooleanProperty>(PorgyConstants::NEW);
        //copy RHS to the new graph
        tlp::copyToGraph(new_graph, _rule, &newelt, newmodel_newelt);

        BooleanProperty *Pos=NULL;
        BooleanProperty *Ban=NULL;
        bool useM=false;
        bool isCond=false;
        if(dataSet!=NULL) {
            dataSet->get<BooleanProperty *>(PorgyConstants::POSITION, Pos);
            dataSet->get<BooleanProperty *>(PorgyConstants::BAN, Ban);
            dataSet->get<bool>(PorgyConstants::useM, useM);
            dataSet->get(PorgyConstants::isCondition, isCond);
        }

        //restore interface
        IntegerProperty *oldId = new_graph->getLocalProperty<IntegerProperty>(SAVEID);
        IntegerProperty *tag_prop = graph->getProperty<IntegerProperty>(PorgyConstants::TAG);
        PortGraphModel lhs_instance_portgraph(graph);
        Iterator<PortNode *> *it=lhs_instance_portgraph.getPortNodes();
        while(it->hasNext()) {
            PortNode *pn = it->next();
            Iterator<Port::PortPtr> *it2 = pn->getPorts();
            while(it2->hasNext()) {
                Port::PortPtr p2 = it2->next();
                //copying properties of ports in STATE_NOTUSED
                node port_rule_left(tag_prop->getNodeValue(p2->getNode()));
                //check if the tag is valid (it should be a node in the rule)
                assert(_rule->isElement(port_rule_left));
                StringProperty* state_rule = _rule->getProperty<StringProperty>(PorgyConstants::STATE);
                if(state_rule->getNodeValue(port_rule_left)==PorgyConstants::STATE_NOTUSED) {
                    //copying state, color and texture
                    node right_equiv_model = findNode(PGMrule.findRightEquiv(port_rule_left),oldId, new_graph);

                    StringProperty* state_new_graph = new_graph->getProperty<StringProperty>(PorgyConstants::STATE);
                    ColorProperty* color_new_graph = new_graph->getProperty<ColorProperty>("viewColor");
                    StringProperty* texture_new_graph=new_graph->getProperty<StringProperty>("viewTexture");
                    StringProperty* state=graph->getProperty<StringProperty>(PorgyConstants::STATE);
                    ColorProperty* color=graph->getProperty<ColorProperty>("viewColor");
                    StringProperty* texture=graph->getProperty<StringProperty>("viewTexture");
                    if(Pos!=NULL) {
                        BooleanProperty *newP = new_graph->getProperty<BooleanProperty>(Pos->getName());
                        PortGraphModel pgm(new_graph);
                        //if M is true, add node in the active P property
                        if(useM) {
                            BooleanProperty* newM = new_graph->getLocalProperty<BooleanProperty>(PorgyConstants::M);
                            if(newM->getNodeValue(right_equiv_model)) {
                                //newP->setNodeValue(right_equiv_model, true);
                                pgm.getPortContainer(right_equiv_model)->getParentPortNode()->select(true, newP);
                            }
                        }
                        else {  //otherwise, always add node to the active P property
                            pgm.getPortContainer(right_equiv_model)->getParentPortNode()->select(true, newP);
                          //  newP->setNodeValue(right_equiv_model, true);
                        }
                    }

                    //copy the state if the RHS node state is empty
                    if((state_new_graph->getNodeValue(right_equiv_model).empty())||(state_new_graph->getNodeValue(right_equiv_model)==PorgyConstants::STATE_NOTUSED))
                        state_new_graph->copy(right_equiv_model, p2->getNode(),state);
                    color_new_graph->copy(right_equiv_model, p2->getNode(),color);
                    texture_new_graph->copy(right_equiv_model, p2->getNode(),texture);

                }
                //looking for edges
                //if the difference between the degree if the LHS instance nodes and the model nodes is greater than 0, there is edges to save
                if(old_graph->deg(p2->getNode())-graph->deg(p2->getNode())>0) {
#ifndef NDEBUG
                    cerr << "There are edges connected to port " << p2->getNode() << " in the LHS instance interface to preserve" << endl;
#endif
                    Iterator<edge> *itp = old_graph->getInOutEdges(p2->getNode());
                    while(itp->hasNext()) {
                        edge e(itp->next());
                        //we do not want edges between center and port and edges in the LHS instance
                        const pair<node, node> &ends = old_graph->ends(e);
                        node src(ends.first), dst(ends.second);
                        if((PortNode::isCenter(src, old_graph))||(PortNode::isCenter(dst, old_graph))||(graph->isElement(e))) {
                            continue;
                        }
#ifndef NDEBUG
                        cerr << "restoring LHS interface for edge " << e << endl;
#endif
                        bool green_edge=false;
                        if(graph->isElement(src)) {
                            if(graph->isElement(dst)) {
#ifndef NDEBUG
                                cerr << "edge (both source and target) is in the LHS instance but it is not part of the LHS instance" << endl;
#endif
                                src = findNode(PGMrule.findRightEquiv(node(tag_prop->getNodeValue(src))), oldId, new_graph);
                                dst = findNode(PGMrule.findRightEquiv(node(tag_prop->getNodeValue(dst))), oldId, new_graph);
                            }
                            else {
#ifndef NDEBUG
                                cerr << "Only edge source is in the LHS instance" << endl;
#endif
                                node n1(PGMrule.findRightEquiv(node(tag_prop->getNodeValue(src))));
                                if(PGMrule.isMember(n1, _rule, PorgyConstants::SIDE_LEFT)) {
                                    cerr << "green edge detected" << endl;
                                    green_edge=true;
                                    src = findNode(n1, tag_prop, graph);
                                    forEach(n, old_graph->getInOutNodes(src)) {
                                        if(PortBase::isPort(n, old_graph)) {
                                            src = n;
                                            break;
                                        }
                                    }
                                    if(new_graph->hasEdge(src, dst, false)) {
                                        cerr << "edge already restored, skipping..." << endl;
                                        continue;
                                    }
                                }
                                else
                                    src = findNode(n1,oldId, new_graph);
                            }
                        }
                        else {
#ifndef NDEBUG
                            cerr << "Only edge target is in the LHS instance " << endl;
#endif
                            node n1(PGMrule.findRightEquiv(node(tag_prop->getNodeValue(dst))));
                            if(PGMrule.isMember(n1, _rule, PorgyConstants::SIDE_LEFT)) {
                                cerr << "green edge detected" << endl;
                                green_edge=true;
                                dst = findNode(n1,tag_prop, graph);
                                node n;
                                forEach(n, old_graph->getInOutNodes(dst)) {
                                    if(PortBase::isPort(n, old_graph)) {
                                        dst = n;
                                        break;
                                    }
                                }
                                if(new_graph->hasEdge(src, dst, false)) {
                                    cerr << "edge already restored, skipping..." << endl;
                                    continue;
                                }
                            }
                            else
                                dst = findNode(n1,oldId, new_graph);
                        }
                        edge new_e = new_graph->addEdge(src, dst);
#ifndef NDEBUG
                        cerr << "Adding edge " << new_e << " between " << src << " and " << dst << endl;
#endif
                        if(green_edge) {
                            newmodel_newelt->setEdgeValue(new_e, true);
                            green_edge=false;
                        }
                        else {
                            PropertyInterface *prop;
                            forEach(prop, new_graph->getObjectProperties()) {
                                prop->copy(new_e,e,prop);
                            }
                        }
                    }
                    delete itp;
                }
            }
            delete it2;
        }
        delete it;

        set<node> setnewnode;
        forEach(n, newmodel_newelt->getNodesEqualTo(true)) {
            if(PortNode::isCenter(n, new_graph))
                setnewnode.insert(n);
        }

        LayoutProperty *layout_old_graph = graph->getProperty<LayoutProperty>("viewLayout");
        SizeProperty *size_old_graph = graph->getProperty<SizeProperty>("viewSize");
        StringProperty *label_old_graph = graph->getProperty<StringProperty>("viewLabel");

        StringProperty *label_new_graph = new_graph->getProperty<StringProperty>("viewLabel");
        SizeProperty *size_new_graph = new_graph->getProperty<SizeProperty>("viewSize");
        LayoutProperty *layout_new_graph = new_graph->getProperty<LayoutProperty>("viewLayout");

        //instead of leaving new nodes at (0,0,0), put them where the old ones are if a LHS node linked to a RHS node
        BooleanProperty used(graph);
        used.setAllNodeValue(false);
        Iterator<PortNode *> *itpo = lhs_instance_portgraph.getPortNodes();
        while(itpo->hasNext()) {
            PortNode *p = itpo->next();  //old node
            Iterator<Port::PortPtr> *itpop = p->getPorts();
            while(itpop->hasNext()) {
                Port::PortPtr po = itpop->next();
                node p_rhs(PGMrule.findRightEquiv(node(tag_prop->getNodeValue(po->getNode()))));
                if(p_rhs.isValid()) {
                    const PortNode* pp = PGMrule.getPortNode(p_rhs);
                    node new_node(findNode(pp->getCenter(), oldId, new_graph));  //replacing node from RHS
                    layout_new_graph->setNodeValue(new_node, layout_old_graph->getNodeValue(p->getCenter()));
                    size_new_graph->setNodeValue(new_node, size_old_graph->getNodeValue(p->getCenter()));
                    used.setNodeValue(p->getCenter(), true);
                    setnewnode.erase(new_node);
                    break;
                }
            }
            delete itpop;
        }
        delete itpo;

        //LHS and RHS are not linked. Try to put each new node at the same position of an old node with the same label
        vector<set<node>::iterator> toremove;
        for(set<node>::iterator it=setnewnode.begin();it!=setnewnode.end();++it) {
            node no(*it);
            //looking for an old node with the same label
            stableForEach(n, used.getNodesEqualTo(false)) {
                if(lhs_instance_portgraph.isCenter(n)) {
                    if(label_old_graph->getNodeValue(n)==label_new_graph->getNodeValue(no)) {
                        layout_new_graph->setNodeValue(no, layout_old_graph->getNodeValue(n));
                        size_new_graph->setNodeValue(no, size_old_graph->getNodeValue(n));
                        used.setNodeValue(n, true);
                        toremove.push_back(it);
                        break;
                    }
                }
            }
        }
        //to avoid invalidating iterators, setnewnode must be updated after iterating on it
        BOOST_FOREACH(set<node>::iterator it, toremove)
                setnewnode.erase(it);
        toremove.clear();

        //for the remaining nodes, put them at the same position of an old node
        for(set<node>::iterator it=setnewnode.begin();it!=setnewnode.end();++it) {
            node no(*it);
            stableForEach(n, used.getNodesEqualTo(false)) {
                layout_new_graph->setNodeValue(no, layout_old_graph->getNodeValue(n));
                size_new_graph->setNodeValue(no, size_old_graph->getNodeValue(n));
                used.setNodeValue(n, true);
                toremove.push_back(it);
                break;
            }
        }
        BOOST_FOREACH(set<node>::iterator it, toremove)
                setnewnode.erase(it);
        if(!setnewnode.empty())
            cerr << __PRETTY_FUNCTION__ << ":" << __LINE__ << ": TODO!! " << endl;

        //removing remaining nodes from the LHS instance (edges will be removed automatically)
        Iterator<node> *itn = graph->getNodes();
        new_graph->delNodes(itn);
        delete itn;
        Observable::unholdObservers();
        //nettoyage
        new_graph->delLocalProperty(SAVEID);
        _rule->delLocalProperty(SAVEID);
        if(new_graph->existLocalProperty(PorgyConstants::ARITY))
            new_graph->delLocalProperty(PorgyConstants::ARITY);
        if(new_graph->existLocalProperty(PorgyConstants::M))
            new_graph->delLocalProperty(PorgyConstants::M);
        if(new_graph->existLocalProperty(PorgyConstants::RULE_SIDE))
            new_graph->delLocalProperty(PorgyConstants::RULE_SIDE);

        if(!isCond) {
            node tgt = traceobj.addMetaNode(new_graph);
            bool isProba=false;
            dataSet->get("isProba", isProba);
            double proba=0;
            dataSet->get("Proba", proba);
            if(!isProba)
                traceobj.addSimpleTransformationEdge(traceobj.findNode(old_graph->getName()),tgt,Pos,Ban,graph,_rule->getId());
            else
                traceobj.addProbabilisticTransformationEdge(traceobj.findNode(old_graph->getName()),tgt,Pos,Ban,graph,_rule->getId(), proba);

            bool draw_model=true;
            if(dataSet!=NULL)
                dataSet->get<bool>(PorgyConstants::layoutNewModel, draw_model);
            if(draw_model){
                string errMsg;
                DataSet ds;
                ds.set("Layout", PGMrule.getLayoutAlgorithm());
#ifndef NDEBUG
                cerr << __PRETTY_FUNCTION__ << ": Layouting " << new_graph->getName() << endl;
#endif
                if(!new_graph->applyPropertyAlgorithm(PorgyConstants::REDRAW_GRAPH, new_graph->getProperty<LayoutProperty>("viewLayout"),errMsg,pluginProgress,&ds))
                    cerr << "Echec du layout du nouveau modèle : " << errMsg << endl;
            }
        }
#ifndef NDEBUG
        else
            cerr << __PRETTY_FUNCTION__ << ": if or while test condition detected => no update to the trace graph" << endl;
#endif
        dataSet->set(PorgyConstants::newModel, new_graph);
        return true;
    }
};

PLUGIN(ApplyRule)
