/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <portgraph/PorgyTlpGraphStructure.h>
#include <portgraph/Trace.h>
#include <portgraph/PorgyConstants.h>

#include <tulip/PropertyAlgorithm.h>
#include <tulip/ForEach.h>
#include <tulip/BooleanProperty.h>

#include <boost/random/mersenne_twister.hpp>
#include <boost/random/uniform_int_distribution.hpp>

#include <ctime>

using namespace tlp;
using namespace std;

using namespace boost::random;

class CheckApply:public tlp::Algorithm {
    string rulename;
    static mt19937 rng;

public:
PLUGININFORMATION(PorgyConstants::CHECK_APPLY_RULE, "Bruno Pinaud", "5/11/09", "Comments", "1.0", "Porgy_internal")
CheckApply(const PluginContext *context):Algorithm(context), rulename("") {
    addDependency(PorgyConstants::APPLY_RULE, "1.0.1");
    addDependency(PorgyConstants::CHECK_RULE, "1.0.1");

    addInParameter<string>(PorgyConstants::RuleName, "", "rule_6");
    addInParameter<bool>(PorgyConstants::checkRule, "", "true");
    addInParameter<int>(PorgyConstants::maximum_number_of_instances, "", "-1");
    addInParameter<unsigned int>(PorgyConstants::minimum_number_of_instances, "", "1");
    addInParameter<BooleanProperty>(PorgyConstants::POSITION, "", "", false);
    addInParameter<BooleanProperty>(PorgyConstants::BAN, "", "", false);
    addInParameter<bool>(PorgyConstants::applyConnectedComponentPacking, "", "true");
    addInParameter<bool>(PorgyConstants::useM, "", "true");
    addInParameter<bool>(PorgyConstants::isCondition, "", "false");
    addInParameter<bool>(PorgyConstants::isProba,"", "", false);

    addOutParameter<node>(PorgyConstants::lastModel, "", "", false);
    addOutParameter<Graph *>(PorgyConstants::newModel, "","",false);
    addOutParameter<double>(PorgyConstants::Proba, "","",false);
}

bool check(string &errMsg) {
    if(!PorgyTlpGraphStructure::isModelGraph(graph)) {
        errMsg = "Active graph is not a model !";
        pluginProgress->setError(errMsg);
        return false;
    }

    if(dataSet!=NULL) {
        dataSet->get<string>(PorgyConstants::RuleName, rulename);
    }

    //je cherche si la règle existe
    if(PorgyTlpGraphStructure::findRule(graph, rulename)==NULL) {
        errMsg = rulename+": no such rule !";
        pluginProgress->setError(errMsg);
        return false;
    }
    return true;
}

bool run() {
    //pas de preview
    pluginProgress->showPreview(false);

    int maximum_number=0; //-1 si on veut toutes les possibilités
    bool check=false;
    BooleanProperty *position=NULL, *ban=NULL;

    if(dataSet!=NULL) {
        dataSet->get<int>(PorgyConstants::maximum_number_of_instances, maximum_number);
        dataSet->get<bool>(PorgyConstants::checkRule, check);
        dataSet->get<BooleanProperty *>(PorgyConstants::POSITION, position);
        dataSet->get<BooleanProperty *>(PorgyConstants::BAN, ban);
    }

    pluginProgress->setComment("Applying rule "+rulename);

    //faut-il vérifier si la règle est applicable ?
    pluginProgress->progress(0,maximum_number+1);
    string errMsg;
    Trace t(PorgyTlpGraphStructure::getMainTrace(graph));
    if(check) {
        if(!graph->applyAlgorithm(PorgyConstants::CHECK_RULE, errMsg, dataSet, pluginProgress)) {
            if(pluginProgress->getError()==PorgyConstants::all_instances_found)
                return true;
            node n = t.addFailureNode(graph->getName(),PorgyTlpGraphStructure::findRule(graph,rulename)->getId(), position,ban);
            dataSet->set(PorgyConstants::lastModel, n);
            dataSet->set(PorgyConstants::newModel, graph);
#ifndef NDEBUG
            cerr << __PRETTY_FUNCTION__ << " (" << __LINE__ << ") Rule not applicable: " << errMsg << endl;
            cerr << __PRETTY_FUNCTION__ << " (" << __LINE__ << ") last model: " << n.id << endl;
#endif
            pluginProgress->setError(PorgyConstants::FAILURE_NODE_ADDED);
            t.redraw(errMsg, pluginProgress);
            return false;
        }
    }
    pluginProgress->progress(1,maximum_number+1);
    vector<Graph *> vector_sub;
    if(maximum_number!=0) {
        Graph *sub;
        forEach(sub, graph->getSubGraphs()) {
            if((PorgyTlpGraphStructure::getRuleNameForSubGraph(sub)==rulename)&&(PorgyTlpGraphStructure::getPPropertyNameAttribute(sub)==((position!=NULL)?position->getName():PorgyConstants::NO_P))) {
                vector_sub.push_back(sub);
            }
        }
        //choose maximum_number of instanciation uniformely at random
        if(maximum_number==-1)
            maximum_number=vector_sub.size();
        bool isProba=false;
        dataSet->get(PorgyConstants::isProba, isProba);
        if(isProba)
            dataSet->set<double>(PorgyConstants::Proba, 1.0/vector_sub.size());
        for(int i=0;i<maximum_number;++i) {
            uniform_int_distribution<> num(0,vector_sub.size()-1);
            int index = num(rng);
            Graph *s = vector_sub.at(index);
            vector_sub.erase(vector_sub.begin()+index);
            if(!s->applyAlgorithm(PorgyConstants::APPLY_RULE, errMsg, dataSet, pluginProgress)) {
                cerr << __PRETTY_FUNCTION__  << "(" << __LINE__ << "): " << errMsg;
                return false;
            }
            ++i;
            pluginProgress->progress(i+1,maximum_number+1);
        }
    }
    t.redraw(errMsg, pluginProgress);
    return true;
}

};

mt19937 CheckApply::rng(static_cast<unsigned int>(time(NULL)));

PLUGIN(CheckApply)
