/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <portgraph/PorgyTlpGraphStructure.h>
#include <portgraph/PortBase.h>
#include <portgraph/PortNodeBase.h>
#include <portgraph/Bridge.h>
#include <portgraph/PortGraphRule.h>

#include <tulip/ForEach.h>
#include <tulip/PropertyAlgorithm.h>
#include <tulip/StringCollection.h>
#include <tulip/BooleanProperty.h>
#include <tulip/LayoutProperty.h>

using namespace tlp;
using namespace std;

namespace {
const char * paramHelp[] = {
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "StringCollection" ) \
    HTML_HELP_DEF( "default", "\"rule_1;rule_1\"" )	\
    HTML_HELP_CLOSE(),
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "bool" ) \
    HTML_HELP_DEF( "default", "false" )	\
    HTML_HELP_BODY()						\
    "If yes, the layout algorithm will be applied to the new rule." \
    HTML_HELP_CLOSE(),
};
}

class Parallel:public tlp::Algorithm {

    vector<Graph *> rulesg;
    StringCollection rules;
    string parallel_rulename;

public:
    PLUGININFORMATION(PorgyConstants::CREATE_PARALLEL_RULE, "Bruno Pinaud", "22/07/10", "Comments", "1.0.1", "Porgy_internal")
    Parallel(const PluginContext* context):Algorithm(context) {
        addDependency(PorgyConstants::REDRAW_RULE, "1.0");

        addInParameter<StringCollection>("Rule string", paramHelp[0], "rule_1;rule_1");
        addInParameter<bool>("Layout new rule?", paramHelp[1], "true");
        addOutParameter<string>(PorgyConstants::PARALLEL_RULENAME);
    }

    bool check(string &errMsg) {

        if (dataSet==NULL) {
            errMsg = "a dataSet is mandatory";
            return false;
        }
        dataSet->get("Rule string", rules);

        if(rules.size()<2) {
            errMsg = "Need at least 2 rules for a parallel application";
            return false;
        }

        for(unsigned i=0;i<rules.size();++i) {
            Graph *g =PorgyTlpGraphStructure::findRule(graph, rules[i]);
            if(g!=NULL) {
                rulesg.push_back(g);
            }
            else {
                errMsg = rules[i] + ": no such rule!!";
                return false;
            }
        }

        return true;
    }

    bool run() {
        //pas de preview
        pluginProgress->showPreview(false);

        //building parallel rule name
        parallel_rulename="parallel_";
        for(unsigned i=0;i!=rules.size();++i) {
            parallel_rulename+=rules[i];
            if(i<rules.size()-1)
                parallel_rulename+="_";
        }

        //Is it already existing?
        if(PorgyTlpGraphStructure::findRule(graph, parallel_rulename)!=NULL)
            return true;

        //Building rule: it is the union of all LHS and RHS
        Graph *parallel_rule=PorgyTlpGraphStructure::getRulesRoot(graph)->addSubGraph(parallel_rulename);
        parallel_rule->setAttribute(PorgyConstants::IS_PARALLEL_RULE, true);

        //Copy last rule entirely (need a bridge)
        Graph *g = rulesg.back();
        rulesg.pop_back();
        Observable::holdObservers();
        tlp::copyToGraph(parallel_rule, g);
        //for all other rules, only LHS, RHS and bridge ports have to be copied
        for(vector<Graph *>::const_iterator it = rulesg.begin();it!=rulesg.end();++it) {
            Graph *r = *it;
            BooleanProperty sel(r);
            sel.setAllNodeValue(false);
            sel.setAllEdgeValue(false);

            Iterator<node> *itn = r->getNodes();
            while(itn->hasNext()) {
                node n = itn->next();
                if((PortGraphRule::isMember(n, r, PorgyConstants::SIDE_LEFT))||(PortGraphRule::isMember(n, r, PorgyConstants::SIDE_RIGHT))) {
                    sel.setNodeValue(n, true);
                }
                if(Bridge::isBridge(n,r)&&(PortBase::isPort(n, r))) {
                    sel.setNodeValue(n, true);
                }
            }
            delete itn;
            Iterator<edge> *ite = r->getEdges();
            while(ite->hasNext()) {
                edge e = ite->next();
                if((PortGraphRule::isMember(e,r, PorgyConstants::SIDE_LEFT))||(PortGraphRule::isMember(e,r, PorgyConstants::SIDE_RIGHT))||(PortGraphRule::isMember(e,r, PorgyConstants::SIDE_BRIDGE))||(PortGraphRule::isMember(e,r, PorgyConstants::SIDE_BRIDGE_OPP))) {
                    sel.setEdgeValue(e, true);
                }
            }
            delete ite;
            tlp::copyToGraph(parallel_rule, r, &sel);

        }

        //je rattache les ports orphelins des bridges
        IntegerProperty *side_parallel = PortGraphRule::getSideProperty(parallel_rule);
        node bridge_center;

        //je cherche le centre du bridge
        node n;
        forEach(n, parallel_rule->getNodes()) {
            if(Bridge::isBridge(n, parallel_rule)&&(PortNodeBase::isCenter(n, parallel_rule)))
                bridge_center=n;
        }
        forEach(n, parallel_rule->getNodes()) {
            if((Bridge::isBridge(n, parallel_rule))&&(PortBase::isPort(n, parallel_rule))) {
                edge e = parallel_rule->addEdge(bridge_center, n);
                side_parallel->setEdgeValue(e, PorgyConstants::SIDE_BRIDGE_PORT);
            }
        }

        //settings redraw algorithm attribute: use the algorithm of the first rule used
        string str;
        g->getAttribute(PorgyConstants::LAYOUTALGORITHM, str);
        parallel_rule->setAttribute(PorgyConstants::LAYOUTALGORITHM, str);

        //layouting the nule rule
        bool layout=true;
        dataSet->get<bool>("Layout new rule?", layout);
        if(layout) {
            string errMsg;
            parallel_rule->applyPropertyAlgorithm(PorgyConstants::REDRAW_RULE,parallel_rule->getProperty<LayoutProperty>("viewLayout"), errMsg, pluginProgress);
        }
        Observable::unholdObservers();

        dataSet->set(PorgyConstants::PARALLEL_RULENAME, parallel_rulename);
        return true;
    }
};
PLUGIN(Parallel)

