/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "RunStrategyElement.h"

#include <tulip/Graph.h>
#include <tulip/StringCollection.h>
#include <tulip/BooleanProperty.h>

#include <portgraph/PorgyTlpGraphStructure.h>
#include <portgraph/PortGraphModel.h>
#include <portgraph/Trace.h>

#include <boost/random/uniform_real.hpp>
#include <boost/random/variate_generator.hpp>
#include <boost/random/uniform_int_distribution.hpp>
#include <boost/fusion/include/at_c.hpp>
#include <boost/fusion/include/vector.hpp>

#include <numeric>
#include <ctime>

using namespace tlp;
using namespace std;
using namespace boost;

using fusion::at_c;

/////////////
namespace {
static bool run_rule(const string &rulename, Graph *g, DataSet *ds, const bool isCond, PluginProgress *pp) {
    //variables pas encore utilisées
    //    bool emptyM=true;
#ifndef NDEBUG
    cerr << "Running " << rulename << " on " << g->getName() << endl;
    BooleanProperty *pos=NULL;
    ds->get(PorgyConstants::POSITION, pos);
    cerr << "Property used for position: " << pos->getName() << endl;
    ds->get(PorgyConstants::BAN, pos);
    cerr << "Property used for Ban: " << ((pos!=NULL)?pos->getName():"NULL") << endl;
#endif
    ds->set<string>(PorgyConstants::RuleName, rulename);
    ds->set<unsigned int>(PorgyConstants::minimum_number_of_instances, 1);
    ds->set<bool>(PorgyConstants::checkRule, true);
    ds->set<bool>(PorgyConstants::lastModel, true);
    ds->set<bool>(PorgyConstants::applyConnectedComponentPacking, false);
    ds->set<bool>(PorgyConstants::isCondition, isCond);

    //    if(emptyM) // appliquer la règle en ne prenant pas en compte M
    ds->set<bool>(PorgyConstants::useM, false);
    //    else
    //        dataSet->set<bool>(PorgyConstants::useM, true);

    string errMsg;
    return g->applyAlgorithm(PorgyConstants::CHECK_APPLY_RULE, errMsg, ds, pp);

}

//////////////////////////////////////
static bool run_parallel_rule(const StrategyElement::parallel_datastruct &rule, DataSet *ds, PluginProgress *pp, Graph *g, const bool isCond) {
    //fabrication de la règle parallèle
    string errMsg;
    vector<string> rulesv(fusion::at_c<1>(rule));
    rulesv.insert(rulesv.begin(),fusion::at_c<0>(rule));
    StringCollection rulesc(rulesv);
    ds->set<StringCollection>("Rule string", rulesc);

    if(!g->applyAlgorithm(PorgyConstants::CREATE_PARALLEL_RULE, errMsg, ds, pp)) {
        cerr << "Failed to create the parallel rule: " << errMsg << endl;
        pp->setError("error");
        return false;
    }
    string parallel_rulename;
    ds->get(PorgyConstants::PARALLEL_RULENAME, parallel_rulename);
    //exécution de la règle parallèle
    return run_rule(parallel_rulename, g, ds, isCond, pp);
}

}

void runStrategyElement::FromOp(BooleanProperty &check, BooleanProperty *prop_old) {
    PorgyConstants::focus_operation from_op=PorgyConstants::PPROPERTY;
    _ds->get<PorgyConstants::focus_operation>("From", from_op);
    switch(from_op) {
    case PorgyConstants::CRTGRAPH:
        check.setAllNodeValue(true);
        check.setAllEdgeValue(true);
        break;
    case PorgyConstants::CRTPOS:
    case PorgyConstants::CRTBAN:
    {
        check = *prop_old;
        break;
    }
    case PorgyConstants::PPROPERTY:
    case PorgyConstants::PROPERTY:
    case PorgyConstants::FPROPERTY:
    case PorgyConstants::NEXTNGB:
    case PorgyConstants::ALLNGB:
    case PorgyConstants::ONENGB:
        cerr << __PRETTY_FUNCTION__ << ":" << __LINE__ << ": TODO!!!" << endl;
        assert(false);
        break;
    }
}
BooleanProperty* runStrategyElement::FocusOperation(const bool isBan, Graph *new_model) {

    BooleanProperty *fold=NULL, *ban_old=NULL;
    BooleanProperty *prop_new=NULL;
    _ds->get<BooleanProperty *>(PorgyConstants::POSITION, fold);
    _ds->get<BooleanProperty *>(PorgyConstants::BAN, ban_old);

    //if fold/ban_old are NULL, need to create a property first
    if(isBan&&(ban_old==NULL)) {
        ban_old = _g->getLocalProperty<BooleanProperty>("Ban");
        _ds->set<BooleanProperty *>(PorgyConstants::BAN, ban_old);
    }
    if(!isBan&&(fold==NULL)) {
        fold = _g->getLocalProperty<BooleanProperty>("Focus");
        _ds->set<BooleanProperty *>(PorgyConstants::POSITION, fold);
    }

    PortGraphModel pg(new_model);

    PorgyConstants::focus_operation operation=PorgyConstants::PPROPERTY;
    _ds->get<PorgyConstants::focus_operation>("Operation", operation);
    if(operation!=PorgyConstants::PPROPERTY) {
        if(!isBan) {
            prop_new = new BooleanProperty(new_model, fold->getName());
        }
        else {
            prop_new = new BooleanProperty(new_model, ban_old->getName());
        }
        prop_new->setAllNodeValue(false);
        prop_new->setAllEdgeValue(false);
    }
    BooleanProperty check(new_model);
    switch(operation) {
    case PorgyConstants::CRTPOS:
    {
        PortNode *pn;
        forEach(pn, pg.getPortNodes()) {
            if(fold->getNodeValue(pn->getCenter())) {
                pn->select(true, prop_new);
            }
        }
        edge e;
        forEach(e, pg.getEdges()) {
            if(fold->getEdgeValue(e)) {
                prop_new->setEdgeValue(e, true);
            }
        }
    }
        break;
    case PorgyConstants::CRTBAN:
    {
        _ds->get<BooleanProperty *>(PorgyConstants::BAN, ban_old);
        PortNode *pn;
        forEach(pn, pg.getPortNodes()) {
            if(ban_old->getNodeValue(pn->getCenter())) {
                pn->select(true, prop_new);
            }
        }
        edge e;
        forEach(e, pg.getEdges()) {
            if(ban_old->getEdgeValue(e)) {
                prop_new->setEdgeValue(e, true);
            }
        }
        break;
    }
    case PorgyConstants::CRTGRAPH:
        prop_new->setAllNodeValue(true);
        prop_new->setAllEdgeValue(true);
        break;
    case PorgyConstants::PROPERTY:
    {
        string check_property, value;
        StrategyElement::Graph_Element elt;
        StrategyElement::relational_operators op;
        _ds->get<string>("Property to check", check_property);
        _ds->get<string>("Value", value);
        _ds->get<StrategyElement::Graph_Element>("Element to check", elt);
        _ds->get<StrategyElement::relational_operators>("Operator",op);
        FromOp(check, isBan?ban_old:fold);

        //getting property
        PropertyInterface *prop = _g->getProperty(check_property);

        //find elements
        switch(elt) {
        case StrategyElement::NODE:
        {
            prop_new->setAllNodeValue(false);
            PortNode *pn;
            forEach(pn, pg.getPortNodes()) {
                node c = pn->getCenter();
                switch(op) {
                case StrategyElement::EQ:
                {
                    if((check.getNodeValue(c))&&(prop->getNodeStringValue(c)==value)) {
                        pn->select(true, prop_new);
                    }
                    break;
                }
                case StrategyElement::NEQ:
                {
                    if((check.getNodeValue(c))&&(prop->getNodeStringValue(c)!=value)) {
                        pn->select(true, prop_new);
                    }
                    break;
                }
                case StrategyElement::GT:
                {
                    if((check.getNodeValue(c))&&(prop->getNodeStringValue(c)>value)) {
                        pn->select(true, prop_new);
                    }
                    break;
                }
                case StrategyElement::GTE:
                {
                    if((check.getNodeValue(c))&&(prop->getNodeStringValue(c)>=value)) {
                        pn->select(true, prop_new);
                    }
                    break;
                }
                case StrategyElement::LTE:
                {
                    if((check.getNodeValue(c))&&(prop->getNodeStringValue(c)<=value)) {
                        pn->select(true, prop_new);
                    }
                    break;
                }
                case StrategyElement::LT:
                {
                    if((check.getNodeValue(c))&&(prop->getNodeStringValue(c)>value)) {
                        pn->select(true, prop_new);
                    }
                    break;
                }
                }
            }

            break;
        }
        case StrategyElement::FUNCTION: //should never happen
            assert(false);
            break;
        case StrategyElement::PORT:
        case StrategyElement::EDGE:
            cerr << __PRETTY_FUNCTION__ << ":" << __LINE__ << ": TODO!!!" << endl;
            assert(false);
            break;
        }
    }
        break;
        ///////////////////////////////////////////
    case PorgyConstants::PPROPERTY:  //used by setPos(Prop(P)) or setBan(Prop(P))
        if(isBan)
            _ds->get<BooleanProperty *>(PorgyConstants::BAN, prop_new);
        else
            _ds->get<BooleanProperty *>(PorgyConstants::POSITION, prop_new);
        break;
        /////////////////////////////
    case PorgyConstants::FPROPERTY:
    {
        FromOp(check, isBan?ban_old:fold);
        StrategyElement::graph_functions func = StrategyElement::ROOT;
        _ds->get<StrategyElement::graph_functions>("Graph Operation", func);
        PortGraphModel pg(new_model);
        switch (func) {
        case StrategyElement::ROOT:
        {
            //check not taken into account correctly if check != crtgraph!!
            node n = new_model->getSource();
            if(check.getNodeValue(n)==true) {
                PortNode *p = pg.getPortNode(n);
                p->select(true, prop_new);
            }
            break;
        }
        case StrategyElement::LEAF:
        {
            PortNode *p;
            forEach(p, pg.getPortNodes()) {
                if(check.getNodeValue(p->getCenter())) {
                    if(p->outdeg()==0) {
                        p->select(true, prop_new);
                    }
                }
            }
            break;
        }

        }
        break;
    }
        ///////////////////////////////////////////////
    case PorgyConstants::NEXTNGB:
    {
        FromOp(check, isBan?ban_old:fold);

        //recup de l'ensemble des sommets de la position courante
        PortNode *pn;
        vector<PortNode *> crtpos;
        forEach(pn, pg.getPortNodes()) {
            if(check.getNodeValue(pn->getCenter())) {
                crtpos.push_back(pn) ;
            }
        }
        //je cherche les ports dont le state est next
        vector<Port::PortPtr> port;
        for(vector<PortNode *>::const_iterator it=crtpos.begin();it!=crtpos.end();++it) {
            Iterator<Port::PortPtr> *it2 = (*it)->getPorts();
            while(it2->hasNext()) {
                Port::PortPtr p=it2->next();
                if(p->getState()=="next")
                    port.push_back(p);
            }delete it2;
        }
        //je cherche les portnodes voisins de ceux marquÃ©s next et je les sÃ©lectionne dans le nouveau P
        for(vector<Port::PortPtr>::const_iterator it=port.begin();it!=port.end();++it) {
            Port::PortPtr p;
            forEach(p, (*it)->getConnectedPorts()) {
                p->getParentPortNode()->select(true,prop_new);
            }
        }
    }
        break;
        ////////////////////////////
    case PorgyConstants::ALLNGB:
    {
        FromOp(check, isBan?ban_old:fold);

        //get nodes of the current position
        PortNode *pn;
        set<const PortNode *> crtpos;
        forEach(pn, pg.getPortNodes()) {
            if(check.getNodeValue(pn->getCenter())) {
                crtpos.insert(pn);
            }
        }

        //find all neighbours
        for(set<const PortNode *>::const_iterator it=crtpos.begin();it!=crtpos.end();++it) {
            const PortNode *p;
            forEach(p, (*it)->getInOutConnectedPortNodes()) {
                if(crtpos.find(p)==crtpos.end())
                    p->select(true, prop_new);
            }
        }
        break;
    }

    case PorgyConstants::ONENGB:
        FromOp(check, isBan?ban_old:fold);

        //get nodes of the current position
        PortNode *pn;
        vector<PortNode *> crtpos;
        forEach(pn, pg.getPortNodes()) {
            if(check.getNodeValue(pn->getCenter())) {
                crtpos.push_back(pn);
            }
        }

        //find all neighbours
        vector<const PortNode *> v;
        for(vector<PortNode *>::const_iterator it=crtpos.begin();it!=crtpos.end();++it) {
            pn = *it;
            const PortNode *p;
            forEach(p, pn->getInOutConnectedPortNodes()) {
                v.push_back(p);
            }
        }

        //pick one
        random::uniform_int_distribution<> num(0,v.size());
        int index = num(gen);
        v.at(index)->select(true, prop_new);
        break;
    }

    return prop_new;
}


//////////////////
//to run a strategy
bool runStrategyElement::run_strat(const StrategyElement::vector_strategyelts &cond, const bool isCond) {
    assert(!cond.empty());
    BOOST_FOREACH(const StrategyElement* elt,cond) {
        if(_pp->state()==TLP_STOP) {
            return true;
        }
        if(_pp->state()==TLP_CANCEL) {
            _pp->setError("Cancel button pressed");
            return false;
        }
        if(elt->run(_pp, _g, _ds, isCond)) {
            _ds->get(PorgyConstants::newModel, _g);
        }
        else
            return false;
    }
    return true;
}
/////////////
runStrategyElement::runStrategyElement(PluginProgress *pp, Graph *g, DataSet *ds, const bool isCond):_pp(pp),_g(g),_ds(ds),_isCond(isCond){}
///////////////
class OneAllVisitor : public static_visitor<bool> {
    PluginProgress *_pp;
    Graph *_g;
    DataSet *_ds;
    bool _isCond;

public:
    OneAllVisitor(PluginProgress *pp, Graph *g, DataSet *ds, const bool isCond):_pp(pp),_g(g),_ds(ds),_isCond(isCond){}

    //for running a simple rule
    bool operator() (const string &rulename) {
        return run_rule(rulename, _g, _ds, _isCond, _pp);
    }

    //for running rules in //
    bool operator() (const StrategyElement::parallel_datastruct &ds) {
        bool ret=false;
        switch(fusion::at_c<2>(ds)) {
        case StrategyElement::PARALLEL:
            ret=run_parallel_rule(ds, _ds, _pp, _g, _isCond);
            break;
        case StrategyElement::WEAKPARALLEL:
            cerr << "____WEAK PARALLEL____: NOT YET IMPLEMENTED" << endl;
            ret=false;
            break;
        }
        return ret;
    }
};


bool runStrategyElement::operator()(const StrategyElement::oneall_datastruct &ds ) {
    switch(ds.second) {
    case StrategyElement::ALL:
#ifndef NDEBUG
        cerr << "__ALL__" << endl;
#endif
        _ds->set<int>(PorgyConstants::maximum_number_of_instances, -1);
        break;
    case StrategyElement::ONE:
#ifndef NDEBUG
        cerr << "__ONE__" << endl;
#endif
        _ds->set<bool>(PorgyConstants::isProba, true);
        _ds->set<int>(PorgyConstants::maximum_number_of_instances, 1);
        break;
    }
    OneAllVisitor visit(_pp, _g, _ds, _isCond);
    return ds.first.apply_visitor(visit);
}

///////////////
bool runStrategyElement::operator()(const StrategyElement::vector_strategyelts &v) {
#ifndef NDEBUG
    cerr << "____VECT____" << endl;
#endif
    //il faut exécuter chaque élément du vecteur.
    return run_strat(v, _isCond);
}
////////////
/// \brief runStrategyElement::operator ()
/// \param ifds
/// \return
bool runStrategyElement::operator()(const StrategyElement::if_datastruct &ifds) {
#ifndef NDEBUG
    cerr << "____IF____" << endl;
#endif
    //exécution de la condition du if
    DataSet ds = *_ds; //on sauve le dataset pour le restaurer ensuite. Le test de la condition ne doit rien modifier
    _g->push(false);
    Graph *g=_g;
#ifndef NDEBUG
    cerr << "__IF__: test condition" << endl;
#endif
    bool cond_status=run_strat(at_c<0>(ifds), true);
    _pp->setError("");
    _g->pop(false);
    *_ds=ds;
    _g=g;
    //la seule chose intéressante de la condition est le code de retour
    if(cond_status) { //test de la condition
#ifndef NDEBUG
        cerr << "__IF__: then" << endl;
#endif
        return run_strat(at_c<1>(ifds), _isCond);
    }
    else if(at_c<2>(ifds)) {  //exécution du else
#ifndef NDEBUG
        cerr << "__IF__: else" << endl;
#endif
        return run_strat(*at_c<2>(ifds), _isCond);
    }
    else {
#ifndef NDEBUG
        cerr << "__IF__: pas de else" << endl;
#endif
        _ds->set(PorgyConstants::newModel, _g);
        return true;
    }
}

bool runStrategyElement::operator()(const StrategyElement::orelse_while_datastruct &wds) {

    bool ret=false;
    switch(at_c<2>(wds)) {
    case StrategyElement::ORELSE:
    {
#ifndef NDEBUG
        cerr << "____ORELSE____" << endl;
#endif
        //on tente d'exécuter la première strategie
        _g->push(false);
        DataSet ds = *_ds; //on sauve le dataset pour le restaurer ensuite. Le test de la condition ne doit rien modifier
        Graph *g=_g;
        if(run_strat(at_c<0>(wds), _isCond))
            ret=true; //la première stratégie a fonctionné
        else {  //la première stratégie a terminé sur un Fail, on essaie l'autre
            _g->pop(false);
            _g=g;
            *_ds = ds; //restauration du dataset
            _pp->setError("");
            ret=run_strat(at_c<1>(wds), _isCond);
        }
        break;
    }
    case StrategyElement::WHILE:
    {
#ifndef NDEBUG
        cerr << "____WHILE____" << endl;
#endif
        bool exit_status=true;   //si la condition se réduit à Fail dès le début, while doit renvoyer Id
        bool cond_status=true;
        StrategyElement::vector_strategyelts cond = at_c<0>(wds);
        StrategyElement::vector_strategyelts code = at_c<1>(wds);
        _ds->set(PorgyConstants::newModel, _g); //in case the while stops immediately
        while(exit_status&&cond_status) {
            Graph *g=_g;
            DataSet ds = *_ds;  //on sauve le dataset pour le restaurer ensuite. Le test de la condition ne doit rien modifier
            _g->push(false);
            //la seule chose intéressante de la condition est le code de retour
#ifndef NDEBUG
            cerr << "__WHILE__: test condition" << endl;
#endif
            cond_status=run_strat(cond, true);
            _pp->setError("");
            _g->pop(false);
            _g=g;
            *_ds = ds;  //restauration du dataset
            if(cond_status) { //exécution du while
#ifndef NDEBUG
                cerr << "__WHILE__: condition ok" << endl;
#endif
                exit_status = run_strat(code, _isCond);
                if(_pp->state()==TLP_STOP) {
                    return true;
                }
            }
        }
        ret=exit_status;
        break;
    }
    }
    return ret;
}
//////////

mt19937 runStrategyElement::gen(static_cast<unsigned int>(time(NULL)));

unsigned runStrategyElement::random_number(const StrategyElement::ppick_datastruct &vect) const {
    vector<double> pr(vect.size());
    for(unsigned i=0;i<vect.size();++i) {
        pr[i] = vect[i].second;
    }
    vector<double> cumulative;
    partial_sum(pr.begin(), pr.end(), back_inserter(cumulative));
    uniform_real<> dist(0, cumulative.back());
    variate_generator<mt19937&, uniform_real<> > die(gen, dist);
    return (lower_bound(cumulative.begin(), cumulative.end(), die()) - cumulative.begin()) + 1;
}

bool runStrategyElement::operator()(const StrategyElement::ppick_datastruct &ds) {
#ifndef NDEBUG
    cerr << "__PPICK__" << endl;
#endif
    //calcul d'un nombre aléatoire en fonction des probas et exécution de la stratégie désignée
    //on ne conserve que la stratégie qui va bien et on l'exécute
    _ds->set<bool>(PorgyConstants::isProba, true);
    unsigned rnd = random_number(ds)-1;
    _ds->set<double>(PorgyConstants::Proba, ds[rnd].second);
    return run_strat(ds[rnd].first, _isCond);
}
////////////
bool runStrategyElement::operator()(const StrategyElement::repeat_ds &ds) {
#ifndef NDEBUG
    cerr << "____REPEAT____" << endl;
#endif
    while(run_strat(at_c<0>(ds), _isCond)) {
        if(_pp->state()==TLP_STOP) {
            return true;
        }
    }
    _g->pop(false);  //on supprime la dernière itération
    _pp->setError("");
    return true;
}
/////////////
//ID/FAIL
bool runStrategyElement::operator()(const StrategyElement::idfail &pu) {
    bool ret=false;
    switch(pu) {
    case StrategyElement::ID:
#ifndef NDEBUG
        cerr << "____ID____" << endl;
#endif
        ret=true;
        _ds->set(PorgyConstants::newModel, _g);
        break;
    case StrategyElement::FAIL:
    {
#ifndef NDEBUG
        cerr << "____FAIL____" << endl;
#endif
        node n = Trace(PorgyTlpGraphStructure::getMainTrace(_g)).addFailureNode(_g->getName(),-1, NULL,NULL);
        _ds->set(PorgyConstants::lastModel, n);
        _pp->setError(PorgyConstants::FAILURE_NODE_ADDED);
        ret=false;
        break;
    }
    }
    return ret;
}
////////////
bool runStrategyElement::operator()(const StrategyElement::focus_datastruct &fd) {  //IsEmpty, setPos, setBan
    BooleanProperty *prop_new = NULL;
    switch(fd.second) {
    case StrategyElement::SETPOS:
    {
#ifndef NDEBUG
        cerr << "____SETPOS____" << endl;
#endif
        prop_new = run_focus_operation(fd.first, false);
        break;
    }
    case StrategyElement::SETBAN:
    {
#ifndef NDEBUG
        cerr << "____SETBAN____" << endl;
#endif
        prop_new = run_focus_operation(fd.first, true);
        break;
    }
    case StrategyElement::ISEMPTY:
    {
#ifndef NDEBUG
        cerr << "___ISEMPTY___" << endl;
#endif
        _g->push(false);
        prop_new = run_focus_operation(fd.first, false);
        bool empty=true;
        node n;
        //prop_new should have all nodes set to false
        forEach(n, prop_new->getNodesEqualTo(true)) {
            empty=false;
            break;
        }
        _g->pop(false);
        if(empty)  {
            _ds->set(PorgyConstants::newModel, _g);
            break;
        }
        else {
            node n = Trace(PorgyTlpGraphStructure::getMainTrace(_g)).addFailureNode(_g->getName(),-1, NULL,NULL);
            _ds->set(PorgyConstants::lastModel, n);
            _pp->setError(PorgyConstants::FAILURE_NODE_ADDED);
            return false;
        }
    }
    }
    return true;
}

class PropertyVisitor : public static_visitor<> {
    DataSet * _ds;

public:
    PropertyVisitor(DataSet* d):_ds(d){}

    void operator()(const StrategyElement::check_prop_datastruct &ds) {
        _ds->set<PorgyConstants::focus_operation>("Operation", PorgyConstants::PROPERTY);
        _ds->set<string>("Property to check", at_c<0>(ds));
        _ds->set<StrategyElement::relational_operators>("Operator", at_c<1>(ds));
        _ds->set<std::string>("Value", at_c<2>(ds));
    }

    void operator()(const StrategyElement::graph_functions func) {
        _ds->set<PorgyConstants::focus_operation>("Operation", PorgyConstants::FPROPERTY);
        _ds->set<StrategyElement::graph_functions>("Graph Operation", func);
    }
};

class FocusOperationVisitor : public static_visitor<> {
    DataSet* _ds;
    Graph* _g;
    const bool _isBan;
public:
    FocusOperationVisitor(DataSet* d, Graph *g, const bool ban):_ds(d),_g(g), _isBan(ban){}

    void operator()(const PorgyConstants::focus_operation &op) {
        _ds->set<PorgyConstants::focus_operation>("Operation", op);
    }

    void operator()(const StrategyElement::property_datastruct &ppd) {
        _ds->set<PorgyConstants::focus_operation>("From", at_c<2>(ppd));
        _ds->set<StrategyElement::Graph_Element>("Element to check", at_c<0>(ppd));

        PropertyVisitor v(_ds);
        at_c<1>(ppd).apply_visitor(v);
    }

    void operator()(const std::string &s) {
        BooleanProperty* p = _g->getProperty<BooleanProperty>(s);
        if(!_isBan) {
            _ds->set<BooleanProperty *>(PorgyConstants::POSITION, p);
        }
        else
            _ds->set<BooleanProperty *>(PorgyConstants::BAN, p);
    }

    void operator()(const StrategyElement::vector2_focusop &pp) {
        _ds->set<PorgyConstants::focus_operation>("Operation", at_c<1>(pp));
        _ds->set<PorgyConstants::focus_operation>("From", at_c<0>(pp));
    }
};

//////////////////////////////
BooleanProperty* runStrategyElement::run_focus_operation(const StrategyElement::setPosBan_datastruct &ds, const bool isBan) {
    Graph *new_model=PorgyTlpGraphStructure::clone(_g, _ds);
    BooleanProperty *p=NULL, *ban=NULL, *prop_new=NULL;

    _ds->get(PorgyConstants::POSITION, p);
    _ds->get(PorgyConstants::BAN, ban);

#ifndef NDEBUG
    cerr << __PRETTY_FUNCTION__ << ":" << __LINE__ << ": new graph clone => " << new_model->getName() << endl;
    cerr << __PRETTY_FUNCTION__ << ":" << __LINE__ << ": property for Position: " << p->getName() << endl;
    cerr << __PRETTY_FUNCTION__ << ":" << __LINE__ << ": property for Ban: " << ((ban!=NULL)?ban->getName():"NULL") << endl;
#endif

    if(!at_c<1>(ds)) {  //no set theory operator given
        FocusOperationVisitor visit(_ds, _g, isBan);
        at_c<0>(ds).apply_visitor(visit);
        prop_new = FocusOperation(isBan, new_model);
        BooleanProperty* b;
        if(isBan) {
            string ban_name=(ban!=NULL)?ban->getName():"Ban";
            b = new_model->getLocalProperty<BooleanProperty>(ban_name);
            b->copy(prop_new);
            _ds->set<BooleanProperty*>(PorgyConstants::BAN, b);
        }
        else {
            b = new_model->getLocalProperty<BooleanProperty>(p->getName());
            b->copy(prop_new);
            _ds->set<BooleanProperty*>(PorgyConstants::POSITION, b);
        }
        prop_new = b;
    }
    else {  //set theory operator given. we need to make a copy of the P and Ban property first, then compute each member, then apply the set property operator
        StrategyElement::set_theory_datastruct set_ds(*at_c<1>(ds));
        BooleanProperty *p=NULL, *ban=NULL;
        _ds->get(PorgyConstants::POSITION, p);
        _ds->get(PorgyConstants::BAN, ban);

        BooleanProperty p_backup(_g), ban_backup(_g);
        bool pback=false, banback=false;
        //saving properties
        if(p!=NULL) {
              p_backup.copy(p);
              pback=true;
        }
        if(ban!=NULL) {
            ban_backup.copy(ban);
            banback=true;
        }
        //compute LHS of the operation
        FocusOperationVisitor visit(_ds, _g, isBan);
        at_c<0>(ds).apply_visitor(visit);
        BooleanProperty *lhs = FocusOperation(isBan, new_model);

        //restore properties original state
        if(pback) {
            p->copy(&p_backup);
        }
        else {
            delete p;
            p=NULL;
        }
        if(banback) {
            ban->copy(&ban_backup);
        }
        else {
            delete ban;
            ban=NULL;
        }
        //compute RHS
        _ds->set<BooleanProperty *>(PorgyConstants::POSITION, p);
        _ds->set<BooleanProperty *>(PorgyConstants::BAN, ban);
        FocusOperationVisitor visit2(_ds, _g, isBan);
        at_c<1>(set_ds).apply_visitor(visit2);
        BooleanProperty *rhs =  FocusOperation(isBan, new_model);

        //computing final results
        node n;
        edge e;
        switch(at_c<0>(set_ds)) {   //compute operation between p and p_backup or ban and ban_backup
        case StrategyElement::MINUS://set property to false in LHS for all RHS nodes set to true. Take LHS property for the final result
        {
            forEach(n, rhs->getNodesEqualTo(true)) {
                lhs->setNodeValue(n, false);
            }
            forEach(e, rhs->getEdgesEqualTo(true)) {
                lhs->setEdgeValue(e, false);
            }

            break;
        }
        case StrategyElement::UNION:
        {
            forEach(n, rhs->getNodesEqualTo(true)) {
                lhs->setNodeValue(n, true);
            }
            forEach(e, rhs->getEdgesEqualTo(true)) {
                lhs->setEdgeValue(e, true);
            }
            break;
        }
        case StrategyElement::INTER:
        {
            forEach(n, _g->getNodes()) {
                if(lhs->getNodeValue(n)&&rhs->getNodeValue(n))
                    continue;
                else
                    lhs->setNodeValue(n, false);
            }
            forEach(e, _g->getEdges()) {
                if(lhs->getEdgeValue(e)&&rhs->getEdgeValue(e))
                    continue;
                else
                    lhs->setEdgeValue(e,false);
            }
            break;
        }
        }

        if(pback) {
            p->copy(&p_backup);
        }
        else {
            delete p;
            p=NULL;
        }
        if(banback) {
            ban->copy(&ban_backup);
        }
        else {
            delete ban;
            ban=NULL;
        }
        BooleanProperty* b;
        if(isBan) {
            string ban_name=(ban!=NULL)?ban->getName():"Ban";

            b = new_model->getLocalProperty<BooleanProperty>(ban_name);
            b->copy(lhs);
            _ds->set<BooleanProperty*>(PorgyConstants::BAN, b);
            _ds->set<BooleanProperty*>(PorgyConstants::POSITION, p);
        }
        else {
            b = new_model->getLocalProperty<BooleanProperty>(p->getName());
            b->copy(lhs);
            _ds->set<BooleanProperty*>(PorgyConstants::POSITION, b);
            _ds->set<BooleanProperty*>(PorgyConstants::BAN, ban);
        }
        prop_new = b;
    }

    Trace traceobj(PorgyTlpGraphStructure::getMainTrace(new_model));
    node tgt = traceobj.addMetaNode(new_model);


    PorgyConstants::focus_operation operation=PorgyConstants::PROPERTY;
    _ds->get<PorgyConstants::focus_operation>("Operation", operation);
    traceobj.addFocusTransformationEdge(traceobj.findNode(_g->getName()), tgt, prop_new, _g, operation, isBan);
    _ds->set(PorgyConstants::newModel, new_model);

    return prop_new;
}
