/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <portgraph/PorgyTlpGraphStructure.h>
#include <portgraph/Trace.h>

#include <tulip/ForEach.h>
#include <tulip/PropertyAlgorithm.h>
#include <tulip/BooleanProperty.h>

#include "StrategyElement.h"
#include "StrategyParser.h"

#include <boost/lexical_cast.hpp>

using namespace std;
using namespace tlp;
using namespace boost;

//TODO:
// * an interface to handle M (resp. N) set is missing. Right now M=R for each rule application and N=NULL

namespace {
bool parse(const string &strat, string &errMsg, StrategyElement::vector_strategyelts &vect_elements, const vector<string> &rules, const vector<std::string> &propertiesname, const vector<std::string> &boolean_properties) {

    typedef string::const_iterator iter;
    comments<iter> comments_;    //le parser pour les commentaires
    strategy_grammar<iter> strat_parser(errMsg, rules, propertiesname, boolean_properties); //le parser du langage de stratégie

    iter begin = strat.begin();
    iter end = strat.end();

    bool r = spirit::qi::phrase_parse(begin, end, strat_parser, comments_, vect_elements);

    //parsing ok ?
    if(r && begin==end)
        return true;
    else if(errMsg.empty()){
        errMsg = " problem around "+string(begin, end);
    }
    return false;
}
}

class StrategySpirit:public tlp::Algorithm {

    std::string strategy;
    StrategyElement::vector_strategyelts strategy_elements;

    //find a local property in the while hierarchy
    bool findP(const string &property_name) const {
        Graph *models_root = PorgyTlpGraphStructure::getModelsRoot(graph);
        Graph *g;
        forEach(g, models_root->getSubGraphs()) {
            if(g->existLocalProperty(property_name))
                return true;
        }
        return false;
    }

public:
    PLUGININFORMATION(PorgyConstants::APPLY_STRATEGY_ALGORITHM, "Bruno Pinaud", "19/07/10", "Comments", "1.0", "Porgy")
    StrategySpirit(const PluginContext *context):Algorithm(context) {

        addDependency(PorgyConstants::CREATE_PARALLEL_RULE, "1.0.1");
        addDependency(PorgyConstants::CHECK_APPLY_RULE, "1.0");

        addInParameter<string>("Strategy", "", "ppick(rule_1;rule_1,0.5,rule_1||rule_1,0.5);rule_2");
        addInParameter<BooleanProperty>(PorgyConstants::POSITION, "", "", false);
        addInParameter<BooleanProperty>(PorgyConstants::BAN,"", "", false);

        addOutParameter<node>(PorgyConstants::lastModel, "", "", false);
    }

    ~StrategySpirit() {
        for(StrategyElement::vector_strategyelts::iterator it=strategy_elements.begin();it!=strategy_elements.end();++it)
            delete *it;
    }

    bool check(string &errorMsg) {
        //pas de preview
        pluginProgress->showPreview(false);

        if(!PorgyTlpGraphStructure::isModelGraph(graph)) {
            errorMsg = "A strategy must be applied on a model";
            return false;
        }

        if(dataSet!=NULL)
            dataSet->get("Strategy", strategy);
        else {
            errorMsg = "dataSet problem";
            return false;
        }
        if(strategy.empty()) {
            errorMsg = "Strategy is empty";
            return false;
        }

        //construit un vecteur avec l'ensemble des noms de rgles
        vector<string> rules, properties, boolean_property;
        Graph *rule;
        rules.reserve(PorgyTlpGraphStructure::getRulesRoot(graph)->numberOfSubGraphs());
        forEach(rule, PorgyTlpGraphStructure::getRulesRoot(graph)->getSubGraphs()) {
            rules.push_back(rule->getName());
        }

        //construit un vecteur avec l'ensemble des proprits du graphe et les proprits de type boolenne
        PropertyInterface *abstractp;
        forEach(abstractp, graph->getObjectProperties()) {
            properties.push_back(abstractp->getName());
            if(abstractp->getTypename()==BooleanProperty::propertyTypename) {
                boolean_property.push_back(abstractp->getName());
            }
        }

        //parse the strategy
        if(parse(strategy, errorMsg, strategy_elements, rules, properties, boolean_property)) {
            BooleanProperty *P=NULL;
            //position handling
            BooleanProperty *rest=graph->getProperty<BooleanProperty>("viewSelection");
            dataSet->get<BooleanProperty *>(PorgyConstants::POSITION, rest);
            if(rest->getName().compare("viewSelection")==0) {
                node n;
                bool sel=false;
                forEach(n, graph->getNodes()) {
                    if(rest->getNodeValue(n)==true) { //il y a au moins un sommet slectionn
                        sel=true;
                        break;
                    }
                }
                string property_name="P";
                int indice=0;
                while(1) {
                    if((!graph->existLocalProperty(property_name))&&(!findP(property_name))) {
                        P = graph->getLocalProperty<BooleanProperty>(property_name);
                        break;
                    }
                    ++indice;
                    property_name="P"+lexical_cast<string>(indice);
                }
                assert(P);
                if(!sel) //rien n'est slectionn = je slectionne tout le graphe
                    P->setAllNodeValue(true);
                else //je copie dans P tout ce qui est slectionn
                    P->copy(rest);
                dataSet->set<BooleanProperty *>(PorgyConstants::POSITION, P);
            }
            else
                dataSet->set<BooleanProperty *>(PorgyConstants::POSITION, rest);

            unsigned i=1;
            graph->getRoot()->push(false);
            Graph *g = graph;
            Trace traceobj(PorgyTlpGraphStructure::getMainTrace(graph));
            //running strategy
            BOOST_FOREACH(const StrategyElement *elt, strategy_elements) {
                pluginProgress->setComment(string("Running strategy step ").append(lexical_cast<string>(i)));
                ++i;
                if(elt->run(pluginProgress, g, dataSet,false)&&(pluginProgress->getError()!=PorgyConstants::FAILURE_NODE_ADDED)) {
                    assert(dataSet->exist(PorgyConstants::newModel));
                    dataSet->get(PorgyConstants::newModel, g);

                    if(g!=NULL) {
                        dataSet->set(PorgyConstants::lastModel, traceobj.findNode(g->getName()));
                    }

#ifndef NDEBUG
                    cerr << __PRETTY_FUNCTION__ << ":" << __LINE__ << ": new graph => " << g->getName() << endl;
                    BooleanProperty *p=NULL, *ban=NULL;
                    dataSet->get(PorgyConstants::POSITION, p);
                    dataSet->get(PorgyConstants::BAN, ban);
                    cerr << __PRETTY_FUNCTION__ << ":" << __LINE__ << ": property for Position: " << p->getName() << endl;
                    cerr << __PRETTY_FUNCTION__ << ":" << __LINE__ << ": property for Ban: " << ((ban!=NULL)?ban->getName():"NULL") << endl;
#endif
                }
                else
                    break;
                if(pluginProgress->state()!=TLP_CONTINUE)
                    break;
            }

            if(pluginProgress->state()==TLP_STOP) {
                return true;
            }
            if(pluginProgress->state()==TLP_CANCEL) {
                errorMsg = "Cancel button pressed";
                return false;
            }

            if(pluginProgress->getError()=="error") {
                errorMsg = "Error when running the strategy";
                return false;
            }

            node last;
            dataSet->get(PorgyConstants::lastModel, last);
            //Est-ce que cette stratgie a dj   t utilise avec le mme P et la mme target finale ?
            //        if(traceobj.existEdge(graph->getName(), last, strategy+P->getName())) {
            //            errorMsg = "Strategy already done before";
            //            return false;
            //        }

            //la stratgie a provoqu au moins une tape de rcriture ou un fail
            if(g!=NULL&&(graph->getName()!=g->getName())) {
#ifndef NDEBUG
                cerr << __PRETTY_FUNCTION__ << ": Ajout d'une arte verte au graphe de Trace entre " << graph->getName() << " et " << last << endl;
#endif
                BooleanProperty *ban=NULL;
                dataSet->get<BooleanProperty *>(PorgyConstants::BAN, ban);
                traceobj.addStrategyTransformationEdge(traceobj.findNode(graph->getName()),last,P,ban,strategy);
            }
            traceobj.redraw(errorMsg, pluginProgress);
            return true;
        }
        else
            return false;
    }

    bool run() {
        return true;
    }

};

PLUGIN(StrategySpirit)
