/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef STRATEGYELEMENT_H
#define STRATEGYELEMENT_H

#include <boost/variant.hpp>
#include <boost/fusion/include/vector.hpp>
#include <boost/spirit/include/qi_optional.hpp>

#include <portgraph/PorgyConstants.h>

namespace tlp{
class Graph;
class PluginProgress;
class DataSet;
}

class StrategyElement
{

public:
    enum strat_operators{RULE, VECT, FOR, IDFAIL, PPICK};
    enum rule_operators{ALL, ONE};
    enum syntactic_sugar{TRY, NOT, REPEAT};
    enum orelse_while{ORELSE, WHILE};
    enum parallel_operators{PARALLEL, WEAKPARALLEL};
    enum strat_focus_op{ISEMPTY, SETPOS, SETBAN};
    enum idfail{ID, FAIL};
    enum Graph_Element{NODE, PORT, EDGE, FUNCTION};
    enum relational_operators{EQ, NEQ, GT, GTE, LT, LTE};
    enum graph_functions{ROOT, LEAF};
    enum set_theory_op{MINUS, UNION, INTER};

    typedef std::vector<StrategyElement *> vector_strategyelts;

    //types utilisés dans l'implémentation uniquement
    typedef boost::fusion::vector2<PorgyConstants::focus_operation, PorgyConstants::focus_operation> vector2_focusop;

    //types utilisés uniquement dans le parser
    typedef boost::fusion::vector3<std::string, std::vector<std::string>, parallel_operators> parallel_datastruct;
    typedef boost::variant<std::string, parallel_datastruct> one_all_datastruct;
    typedef boost::fusion::vector3<std::string, StrategyElement::relational_operators, std::string> check_prop_datastruct;

    //types utilisés dans le parser et l'implémentation
    typedef boost::fusion::vector3<StrategyElement::Graph_Element, boost::variant<check_prop_datastruct, graph_functions>, PorgyConstants::focus_operation> property_datastruct;
    typedef boost::variant<PorgyConstants::focus_operation, property_datastruct, std::string, vector2_focusop> focusop_datastruct;
    typedef boost::fusion::vector2<set_theory_op, focusop_datastruct> set_theory_datastruct;
    typedef boost::fusion::vector2<focusop_datastruct, boost::optional<set_theory_datastruct> > setPosBan_datastruct;
    typedef boost::fusion::vector3<vector_strategyelts, vector_strategyelts, boost::optional<vector_strategyelts> > if_datastruct;
    typedef boost::fusion::vector3<vector_strategyelts, vector_strategyelts, orelse_while> orelse_while_datastruct;

    typedef std::pair<unsigned, vector_strategyelts> for_datastruct;
    typedef std::vector<std::pair<vector_strategyelts, double> > ppick_datastruct;
    typedef boost::fusion::vector2<vector_strategyelts, syntactic_sugar> repeat_ds;
    typedef std::pair<setPosBan_datastruct, strat_focus_op> focus_datastruct;
    typedef std::pair<one_all_datastruct, rule_operators> oneall_datastruct;

    /**
     * @brief StrategyElement Constructor for One() et All operators.
     * @brief by default, a rule R (or R||R or R|||R) alone is considered as One(R)
     * @param rules
     * @param op Operation type
     */
    StrategyElement(const one_all_datastruct &rules, const rule_operators op);
    //Id/Fail
    StrategyElement(const idfail op);
    //setPos, setBan, isEmpty
    StrategyElement(const setPosBan_datastruct &, const strat_focus_op op);
    //ppick
    StrategyElement(const ppick_datastruct &vect);
    //pfor
    StrategyElement(const for_datastruct &p);
    //while, orelse
    StrategyElement(const orelse_while_datastruct &v);
    //if
    StrategyElement(const if_datastruct &v);
    //try, not, repeat
    StrategyElement(const vector_strategyelts &v, const syntactic_sugar s);

    //exécute l'élement de la stratégie
    bool run(tlp::PluginProgress *pluginProgress, tlp::Graph *g, tlp::DataSet *dataSet, const bool isCond) const;

private:

    //variant pour stocker les différentes structures de données
    boost::variant<idfail, oneall_datastruct, vector_strategyelts, orelse_while_datastruct, if_datastruct, repeat_ds, ppick_datastruct, focus_datastruct> data;
};

#endif // STRATEGYELEMENT_H
