/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef STRATEGYPARSER_CXX
#define STRATEGYPARSER_CXX

#include <boost/spirit/include/qi.hpp>
#include <boost/spirit/include/phoenix_operator.hpp>
#include <boost/spirit/include/phoenix_object.hpp>
#include <boost/spirit/include/phoenix_bind.hpp>
#include <boost/spirit/include/phoenix_fusion.hpp>
#include <boost/spirit/include/support_info.hpp>
#include <boost/fusion/adapted/std_pair.hpp>

#include <portgraph/PorgyConstants.h>

#include "StrategyElement.h"

namespace qi = boost::spirit::qi;
namespace ascii = boost::spirit::ascii;
namespace phoenix = boost::phoenix;

//pour parser les stratégies qui se réduise à ID ou Fail
struct id_fail : qi::symbols<char, StrategyElement::idfail> {
    id_fail() {
        add     ("Id"    , StrategyElement::ID)
                ("Fail"  , StrategyElement::FAIL)
                ;}
} id_fail;

//available graph functions
struct graph_functions : qi::symbols<char, StrategyElement::graph_functions> {
    graph_functions() {
        add     ("Root"    , StrategyElement::ROOT)
                ("Leaf"  , StrategyElement::LEAF)
                ;}
} graph_functions;

//opérations pour le focus
struct focusing_op : qi::symbols<char, PorgyConstants::focus_operation> {
    focusing_op() {
        add     ("CrtGraph", PorgyConstants::CRTGRAPH)
                ("CrtPos", PorgyConstants::CRTPOS)
                ("CrtBan", PorgyConstants::CRTBAN)
                ;}
} focusing_op;

//opérations sur les éléments
struct graph_element : qi::symbols<char, StrategyElement::Graph_Element> {
    graph_element() {
        add     ("Node", StrategyElement::NODE)
                ("Port", StrategyElement::PORT)
                ("Edge", StrategyElement::EDGE)
                ("Function", StrategyElement::FUNCTION)
                ;}
} graph_element;

//relational operators
struct relop : qi::symbols<char, StrategyElement::relational_operators> {
    relop() {
        add     ("==", StrategyElement::EQ)
                ("!=", StrategyElement::NEQ)
                (">", StrategyElement::GT)
                (">=", StrategyElement::GTE)
                ("<", StrategyElement::LT)
                ("<=", StrategyElement::LTE)
                ;}
} relop;

//focus set theory operators
struct focusop : qi::symbols<char, StrategyElement::set_theory_op> {
    focusop() {
        add     ("\\", StrategyElement::MINUS)
                ("\\cup", StrategyElement::UNION)
                ("\\cap", StrategyElement::INTER)
                ;}
} focusop;

//parser pour des doubles utilisé avec ppick : nombre de la forme 0.???
template <typename T>
struct ts_ureal_policies : qi::strict_ureal_policies<T> {
    static bool const expect_dot = true;    //obligation d'avoir une virgule
    static bool const allow_leading_dot=false;
    static bool const allow_trailing_dot=false;
    //no exponent
    template <typename It>
    static bool parse_exp(It&, It const&){return false;}
    //  No exponent
    template <typename It, typename Attribute>
    static bool parse_exp_n(It&, It const&, Attribute&){return false;}
    //forcément 0 avant la virgule
    template <typename It, typename Attribute>
    static bool parse_n(It& first, It const& last, Attribute& ) {
        return qi::parse(first, last, qi::uint_(0)); //version 1.46.1 de boost minimum
    }
};

//grammaire pour définir des commentaires à la C++
template <typename Iterator>
struct comments : qi::grammar<Iterator> {

    comments():comments::base_type(start) {

        using ascii::char_;

        start = ascii::space  // tab/space/cr/lf
                // C++-style comments
                | "/*" >> *(char_ - "*/") >> "*/"
                          | "//" >> *(char_ - qi::eol)
                          ;
    }
    qi::rule<Iterator> start;
};

//le parser
template <typename Iterator>
struct strategy_grammar : qi::grammar<Iterator, StrategyElement::vector_strategyelts(), comments<Iterator> > {

    typedef comments<Iterator> comments_;

    strategy_grammar(std::string &errorMsg, const std::vector<std::string> &rules, const std::vector<std::string> &propertynames, const std::vector<std::string> &boolean_properties):strategy_grammar::base_type(start, "Porgy Grammar"),errMsg(errorMsg),sym_rules(rules, rules, "valid rule names"),sym_properties(propertynames, propertynames, "valid property names"), sym_boolean_property(boolean_properties, boolean_properties, "valid Boolean Properties") {
        using qi::lit;
        using qi::uint_;
        using qi::lexeme;
        using qi::char_;
        using phoenix::new_;
        using phoenix::construct;
        using phoenix::val;
        using phoenix::at_c;

        using namespace qi::labels;

        //parser pour des doubles de type 0.xx
        qi::real_parser<double, ts_ureal_policies<double> > doublep_;

        //top level structure of a strategy => s;s;....;s
        start %= seq % ';';

        //structure d'une instruction
        seq = parallel[_val = new_<StrategyElement>(_1, val(StrategyElement::ONE))]|
                pall[_val = new_<StrategyElement>(_1, val(StrategyElement::ALL))]|
                pone[_val = new_<StrategyElement>(_1, val(StrategyElement::ONE))]|
                weaker_parallel[_val = new_<StrategyElement>(_1, val(StrategyElement::ONE))]|
                ppick[_val = new_<StrategyElement>(_1)]|
                pwhile[_val = new_<StrategyElement>(_1)]|
                pif[_val = new_<StrategyElement>(_1)]|
                porelse[_val = new_<StrategyElement>(_1)]|
                ridfail[_val = new_<StrategyElement>(_1)]|
                pfor[_val = new_<StrategyElement>(_1)]|
                ptry[_val = new_<StrategyElement>(_1, val(StrategyElement::TRY))]|
                pnot[_val = new_<StrategyElement>(_1, val(StrategyElement::NOT))]|
                prepeat[_val = new_<StrategyElement>(_1, val(StrategyElement::REPEAT))]|
                setpos[_val = new_<StrategyElement>(_1, val(StrategyElement::SETPOS))]|
                setban[_val = new_<StrategyElement>(_1, val(StrategyElement::SETBAN))]|
                isEmpty[_val = new_<StrategyElement>(_1, val(StrategyElement::ISEMPTY))]|
                rulename[_val = new_<StrategyElement>(_1, val(StrategyElement::ONE))]; //the execution of a simple rule R is considered as One(R)

                //opérateur ppick
                ppick %= lexeme[lit("ppick")]
                > '('
                > ((start > lit(',') > doublep_) % ',')
                > lit(')');

                //boucle while
                pwhile %= lexeme[lit("while")]
                > lit('(')
                > start
                > lit(')')
                > lexeme[lit("do")]
                > lit('(')
                > start
                > lit(')')[at_c<2>(_val)=StrategyElement::WHILE];

                //opérateur orelse
                porelse %= lit('(')
                > start
                > lit(')')
                >lexeme[lit("orelse")]
                > lit('(')
                > start
                > lit(')')[at_c<2>(_val)=StrategyElement::ORELSE];

                //opérateur if
                pif %= lexeme[lit("if")]
                > lit('(')
                > start
                > lit(')')
                > lexeme[lit("then")]
                > lit('(')
                > start
                > lit(')')
                > -(lexeme[lit("else")]> lit('(')> start > lit(')'));

                //boucle for
                pfor %=  lexeme[lit("for")]
                >'('
                > uint_
                > lit(')')
                >lit('(')
                > start
                > lit(')');

                //opérateur try
                ptry %= lexeme[lit("try")]
                > lit('(')
                > start
                > lit(')');

                //opérateur not
                pnot %= lexeme[lit("not")]
                > lit('(')
                > start
                > lit(')');

                //opérateur repeat
                prepeat %= lexeme[lit("repeat")]
                > lit('(')
                > start
                > lit(')');

                //Id, Fail
                ridfail %= id_fail;

                //règle en parallèle A||A' (minimun 2, pas de maxi)
                parallel %= rulename
                >> +(lexeme[lit("||")] > rulename)[at_c<2>(_val)=StrategyElement::PARALLEL];

                //règle en parallèle A|||A'
                weaker_parallel %= rulename
                >> +(lexeme[lit("|||")] > rulename)[at_c<2>(_val)=StrategyElement::WEAKPARALLEL];

                application %= weaker_parallel | parallel | rulename;

                //codage du nom des règles
                rulename = sym_rules;

                //All() operator
                pall = lexeme[lit("All")]
                > lit('(')
                > application
                >lit(')');

                //One() operator
                pone = lexeme[lit("One")]
                > lit('(')
                > application
                >lit(')');

                focus_op = focusing_op | property | (lexeme[lit("Prop")] > L'(' > L'"' > prop > L'"' > L')') | nextngb | onengb | allngb;

                //pour coller à la stratégie, il faut remplacer focusing_op par focus et changer le code en conséquence
                nextngb = (lexeme[lit("NextNgb")]
                > L'('
                > focusing_op
                > L')')[_val=construct<StrategyElement::vector2_focusop>(_1, PorgyConstants::NEXTNGB)];

                //pour coller à la stratégie, il faut remplacer focusing_op par focus et changer le code en conséquence
                allngb = (lexeme[lit("AllNgb")]
                > L'('
                > focusing_op
                > L')')[_val=construct<StrategyElement::vector2_focusop>(_1, PorgyConstants::ALLNGB)];

                //pour coller à la stratégie, il faut remplacer focusing_op par focus et changer le code en conséquence
                onengb = (lexeme[lit("OneNgb")]
                > L'('
                > focusing_op
                > L')')[_val=construct<StrategyElement::vector2_focusop>(_1, PorgyConstants::ONENGB)];

                prop = sym_boolean_property;

                set_theory = focusop > focus_op;

                focus = focus_op >> -(set_theory);

                //MAJ des positions
                setpos %= lexeme[lit("setPos")]
                > L'('
                > focus
                > L')';

                //MAJ des positions : liste des positions interdites
                setban %= lexeme[lit("setBan")]
                > L'('
                > focus
                > L')';

                //isEmpty
                isEmpty %= lexeme[lit("isEmpty")]
                > L'('
                > focus
                > L')';

                check_prop %= L'"' > sym_properties > L'"' > relop > L'"' > +((char_) - L'"') > L'"' ;


                //Property
                //voir pour ajouter un test : si graph_element == FUNCTION, la suite doit être un graph_functions
                property %= lexeme[lit("Property")]
                > L'('
                > graph_element
                > ','
                > (graph_functions | check_prop)
                > ','
                > focusing_op   //should be focus_op to match the strategy
                > ')';

                start.name("start");
                seq.name("seq");
                ppick.name("ppick");
                ridfail.name("Id or Fail");
                rulename.name("valid rulename");
                pfor.name("for");
                pwhile.name("while");
                porelse.name("orelse");
                ptry.name("try");
                pnot.name("not");
                pif.name("if");
                prepeat.name("repeat");
                parallel.name("parallel");
                weaker_parallel.name("weaker parallel");
                setpos.name("setPos");
                setban.name("SetBan");
                focus_op.name("Focus Operation (property)");
                focus.name("Focus Operation");
                property.name("Property");
                prop.name("valid Tulip P Property");
                nextngb.name("NextNgb");
                allngb.name("AllNgb");
                onengb.name("OneNgb");
                pall.name("All");
                pone.name("One");
                set_theory.name("Set theory construct");
                application.name("Application");
                check_prop.name("Operation on a Tulip property");
                id_fail.name("Id or Fail");
                focusing_op.name("Position operation");
                graph_element.name("Graph Element");
                graph_functions.name("Graph functions (Root, Leaf)");
                focusop.name("Set theory operator");

//                debug(start);

                qi::on_error<qi::fail>
                (
                start
                , phoenix::ref(errMsg)
                = (construct<std::string>(val("Error! Expecting "))
                + phoenix::bind(&boost::spirit::info::tag, _4)      //what failed
                + construct<std::string>(val(" here: \"") )
                + construct<std::string>(_3, _2)
                + construct<std::string>( val("\"") )
                )
                );
    }

                std::string &errMsg;
                qi::symbols<char, std::string> sym_rules, sym_properties, sym_boolean_property;

                qi::rule<Iterator, StrategyElement::vector_strategyelts(), comments_> start, ptry, pnot, prepeat;
                qi::rule<Iterator, StrategyElement::ppick_datastruct(), comments_> ppick;
                qi::rule<Iterator, StrategyElement::for_datastruct(), comments_> pfor;
                qi::rule<Iterator, StrategyElement::one_all_datastruct(), comments_> pall, pone, application;
                qi::rule<Iterator, std::string()> rulename, prop;
                qi::rule<Iterator, StrategyElement::focusop_datastruct(), comments_> focus_op;
                qi::rule<Iterator, StrategyElement::setPosBan_datastruct(), comments_> setpos, setban, isEmpty, focus;
                qi::rule<Iterator, StrategyElement::set_theory_datastruct(), comments_> set_theory;
                qi::rule<Iterator, StrategyElement::idfail(), comments_> ridfail;
                qi::rule<Iterator, StrategyElement::vector2_focusop(), comments_> nextngb, onengb, allngb;
                qi::rule<Iterator, StrategyElement::orelse_while_datastruct(), comments_> pwhile, porelse;
                qi::rule<Iterator, StrategyElement::parallel_datastruct(), comments_> parallel, weaker_parallel;
                qi::rule<Iterator, StrategyElement *(), comments_> seq;
                qi::rule<Iterator, StrategyElement::if_datastruct(), comments_> pif;
                qi::rule<Iterator, StrategyElement::property_datastruct(), comments_> property;
                qi::rule<Iterator, StrategyElement::check_prop_datastruct(), comments_> check_prop;
    };


        #endif // STRATEGYPARSER_CXX
