/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "LayoutUtils.h"

#include <portgraph/PortGraphModel.h>
#include <portgraph/PortNode.h>
#include <portgraph/PorgyConstants.h>

#include <tulip/ForEach.h>
#include <tulip/ConnectedTest.h>
#include <tulip/StringCollection.h>
#include <tulip/LayoutProperty.h>
#include <tulip/ColorProperty.h>
#include <tulip/SizeProperty.h>

using namespace std;
using namespace tlp;

namespace {
const char* paramHelp[] = {
    //LayoutAlgorithm
    HTML_HELP_OPEN() \
    HTML_HELP_DEF("Type", "StringCollection") \
    HTML_HELP_DEF("Values", "GEM (Frick) <BR> Sugiyama (OGDF)") \
    HTML_HELP_DEF("Default", "GEM (Frick)") \
    HTML_HELP_BODY() \
    "Choose the layout algorithm to use" \
    HTML_HELP_CLOSE(),

    // node size
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "Size" ) \
    HTML_HELP_DEF( "values", "An existing size property" ) \
    HTML_HELP_DEF( "default", "viewSize" ) \
    HTML_HELP_BODY() \
    "This parameter defines the property used for node's sizes." \
    HTML_HELP_CLOSE(),

    // node color
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "Color" ) \
    HTML_HELP_DEF( "values", "An existing color property" ) \
    HTML_HELP_DEF( "default", "viewColor" ) \
    HTML_HELP_BODY() \
    "This parameter defines the property used for node/edge's color." \
    HTML_HELP_CLOSE(),

    // border color
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "Color" ) \
    HTML_HELP_DEF( "values", "An existing color property" ) \
    HTML_HELP_DEF( "default", "viewBorderColor" ) \
    HTML_HELP_BODY() \
    "This parameter defines the property used for node/edge border color." \
    HTML_HELP_CLOSE(),

    // use current layout
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "bool" ) \
    HTML_HELP_DEF( "values", "true/false" ) \
    HTML_HELP_DEF( "default", "true" ) \
    HTML_HELP_BODY() \
    "This parameter defines if the current graph layout has to be used as a starting point." \
    HTML_HELP_CLOSE(),

    //
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "LayoutProperty" ) \
    HTML_HELP_DEF( "values", "An existing layout property" ) \
    HTML_HELP_DEF( "default", "viewLayout" ) \
    HTML_HELP_BODY() \
    "This parameter defines the layout property to use when the current layout needs to be used as a starting point." \
    HTML_HELP_CLOSE(),
};
}

class RedrawModel:public LayoutAlgorithm {

    string layout_algo;

    Graph* createAbstractGraph(PortGraphModel &p) {
        Graph *sub= graph->addSubGraph();

        //    BooleanProperty *new_node = sub->getProperty<BooleanProperty>(PorgyConstants::NEW);
        //    BooleanProperty *temp = sub->getLocalProperty<BooleanProperty>("temp");
        //    temp->setAllNodeValue(false);
        //    LayoutProperty *layout = sub->getProperty<LayoutProperty>("viewLayout");

        //ajout des centres des portnodes
        PortNode *pn;
        forEach(pn, p.getPortNodes()) {
            sub->addNode(pn->getCenter());
            //si le sommet est nouveau, je mets un sommet identique considr comme non nouveau est je le relie   l'autre.
            //ce nouveau sommet permettra d'attirer l'autre dans GEM
            //        if((new_node->getNodeValue(n))&&(layout_algo==PorgyConstants::PORGY_GEM)) {
            //            node n2 = sub->addNode();
            //            new_node->setNodeValue(n2, false);
            //            layout->setNodeValue(n2, layout->getNodeValue(n));
            //            temp->setNodeValue(n2, true);
            //            sub->addEdge(n,n2);
            //        }
        }
        //ajout d'une arte si les centres sont connects par l'intermdiaire des ports
        edge e;
        stableForEach(e, p.getEdges()) {
            const pair<PortNode *, PortNode*> ends = p.ends(e);
            sub->addEdge(ends.first->getCenter(),ends.second->getCenter());
        }

        if(layout_algo==PorgyConstants::PORGY_GEM) {
            //make graph fully connected
            vector<edge> tmp;
            ConnectedTest::makeConnected(sub, tmp);
        }

        return sub;
    }


public:
    PLUGININFORMATION(PorgyConstants::REDRAW_GRAPH, "Bruno pinaud", "28/09/09", "Comments", "1.0", PorgyConstants::CATEGORY_NAME)
    RedrawModel(const tlp::PluginContext *context):LayoutAlgorithm(context) {

        addDependency(PorgyConstants::GEM, PorgyConstants::GEM_VERSION);
        addDependency(PorgyConstants::SUGIYAMA, "1.5");
        addDependency(PorgyConstants::PORGY_GEM, PorgyConstants::PORGY_GEM_VERSION);
        addDependency("Fast Overlap Removal", "1.3");

        addInParameter<StringCollection>("Layout", paramHelp[0], LAYOUT_ALGO);
        addInParameter<bool>("Use a given layout", paramHelp[4], "true");
        addInParameter<LayoutProperty>("Initial layout", paramHelp[5], "viewLayout",false);
        addInOutParameter<SizeProperty>("node size", paramHelp[1],"viewSize");
        addInOutParameter<ColorProperty>("Color", paramHelp[2],"viewColor");
        addInOutParameter<ColorProperty>("Bordercolor", paramHelp[3],"viewBorderColor");
    }

    bool check(string &errMsg) {
        if(!PorgyTlpGraphStructure::isPorgyGraph(graph, errMsg)) {
            errMsg = "Not a correct Porgy graph structure: "+errMsg;
            return false;
        }
        if(!PorgyTlpGraphStructure::isModelGraph(graph)) {
            errMsg = "Graph is not a model";
            return false;
        }

        return true;
    }

#define ELT_RANKINGLIST "LongestPathRanking;OptimalRanking"
    bool run() {
        StringCollection tmp;
        dataSet->get("Layout", tmp);
        layout_algo=tmp.getCurrentString();

        //parse graph into proper structure and classes
        PortGraphModel p(graph);
        string errMsg;
        Graph *sub = createAbstractGraph(p);
        if(p.numberOfEdges()>0) {
            if(layout_algo==PorgyConstants::SUGIYAMA) {
                StringCollection sc(ELT_RANKINGLIST);
                sc.setCurrent(1);
                dataSet->set("Ranking",sc);
            }
            else if((layout_algo==PorgyConstants::GEM)||(layout_algo==PorgyConstants::PORGY_GEM)) {
                bool use_current = true;
                dataSet->get("Use a given layout", use_current);
                if(use_current) {
                    LayoutProperty *p=NULL;
                    dataSet->get("Initial layout", p);
                    dataSet->set("initial layout", p);
                }

            }
            if(layout_algo!=PorgyConstants::NONE) {
                if(!sub->applyPropertyAlgorithm(layout_algo, result, errMsg, pluginProgress, dataSet)) {
                    cerr << __PRETTY_FUNCTION__ << ":" << __LINE__ << " problem with "<< layout_algo << ": " << errMsg << endl;
                    return false;
                }
                //            }
                //            else {
                //                LayoutProperty *l = sub->getProperty<LayoutProperty>("viewLayout");
                //                result->copy(l);
                //            }
                //    PorgyTlpGraphStructure::removeAbstractGraph(sub);
            }

        }

        SizeProperty *nodeSize = graph->getProperty<SizeProperty>("viewSize");
        ColorProperty *viewColor = graph->getProperty<ColorProperty>("viewColor");
        ColorProperty *bordercolor = graph->getProperty<ColorProperty>("viewBorderColor");
        dataSet->get("node size", nodeSize);
        dataSet->get("Color", viewColor);
        dataSet->get("Bordercolor", bordercolor);



        //fast overlap removal
        dataSet->set("layout", result);
        dataSet->set<double>("x border", 1.0f);
        dataSet->set<double>("y border", 1.0f);
        if(!sub->applyPropertyAlgorithm("Fast Overlap Removal", result, errMsg, pluginProgress, dataSet)) {
            cerr << __PRETTY_FUNCTION__ << ":" << __LINE__ << "Fast Overlap Removal: " << errMsg << endl;
        }
        removeAbstractGraph(sub);
        PortNode *pn;
        forEach(pn, p.getPortNodes()) {
            pn->draw(nodeSize, viewColor, result, bordercolor);
        }
        p.drawSameSizes(nodeSize, result);
        return true;
    }

};
PLUGIN(RedrawModel)


