/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "LayoutUtils.h"

#include <portgraph/PortGraphRule.h>
#include <portgraph/PortNode.h>
#include <portgraph/PorgyConstants.h>
#include <portgraph/Bridge.h>
#include <portgraph/BridgePort.h>

#include <tulip/ForEach.h>
#include <tulip/PropertyAlgorithm.h>
#include <tulip/BooleanProperty.h>
#include <tulip/LayoutProperty.h>
#include <tulip/SizeProperty.h>
#include <tulip/DoubleProperty.h>
#include <tulip/ColorProperty.h>

#include <boost/foreach.hpp>

using namespace std;
using namespace tlp;

namespace {
const char* paramHelp[] = {
    // node size
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "Size" ) \
    HTML_HELP_DEF( "values", "An existing size property" ) \
    HTML_HELP_DEF( "default", "viewSize" ) \
    HTML_HELP_BODY() \
    "This parameter defines the property used for node's sizes." \
    HTML_HELP_CLOSE(),

    // node color
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "Color" ) \
    HTML_HELP_DEF( "values", "An existing color property" ) \
    HTML_HELP_DEF( "default", "viewColor" ) \
    HTML_HELP_BODY() \
    "This parameter defines the property used for node/edge's color." \
    HTML_HELP_CLOSE(),

    // border color
    HTML_HELP_OPEN() \
    HTML_HELP_DEF( "type", "Color" ) \
    HTML_HELP_DEF( "values", "An existing color property" ) \
    HTML_HELP_DEF( "default", "viewBorderColor" ) \
    HTML_HELP_BODY() \
    "This parameter defines the property used for node/edge border color." \
    HTML_HELP_CLOSE(),
};
}

class RedrawRule:public LayoutAlgorithm {

    typedef pair<node, int> pairnodeint;
    typedef pair<node, map<node, int> >  pairnodemapnodeint;

    Graph* createAbstractGraph(PortGraphRule &p) const {
        Graph *sub= graph->addSubGraph();
        IntegerProperty *side = PortGraphRule::getSideProperty(graph);

        //add portnode centers
        PortNode *pn;
        forEach(pn, p.getPortNodes()) {
            sub->addNode(pn->getCenter());
        }
        sub->addNode(p.getBridge()->getCenter());
        //add an edge if centers are connected via their respective ports
        edge e;
        stableForEach(e, graph->getEdges()) {
            const pair<node, node> &ends = graph->ends(e);
            if((!PortGraphRule::isMember(e, graph, PorgyConstants::SIDE_BRIDGE_OPP))&&(!PortNodeBase::isCenter(ends.second, graph))&&(!PortNodeBase::isCenter(ends.first, graph))) {
                PortNodeBase *source, *target;
                if(Bridge::isBridge(ends.first, graph))
                    source = p.getBridge(); // p.getContainer<Bridge *>(ends.first);
                else
                    source = p.getPortNode(ends.first);
                if(Bridge::isBridge(ends.second, graph))
                    target = p.getBridge(); //p.getContainer<Bridge *>(ends.second);
                else
                    target = p.getPortNode(ends.second);
                if(source!=target) {
                    edge e1 = sub->addEdge(source->getCenter(),target->getCenter());
                    //copie de rule_side
                    side->setEdgeValue(e1, side->getEdgeValue(e));
                }
            }
        }
        return sub;
    }

    /*translation des centres vers leur place. Inutile de s'occuper
     * des ports.
     */
    void Translate(PortGraphRule &p, SizeProperty *size) const {
        const double translation_offset = 5;
        double extremeL = HUGE_VAL, extremeR = -HUGE_VAL;
        double x, width;

        node n;
        PortNode *pn;
        forEach(pn, p.getPortNodes()) {
            n = pn->getCenter();
            x = result->getNodeValue(n).getX();
            width = size->getNodeValue(n).getW();

            if (PortGraphRule::isMember(n, graph, PorgyConstants::SIDE_LEFT)){
                extremeR = max(extremeR, x+width/2);
            }
            else if (PortGraphRule::isMember(n, graph, PorgyConstants::SIDE_RIGHT)) {
                extremeL = min(extremeL, x-width/2);
            }
        }

        //calcul du dcalage
        double offsetLeftSide  = -translation_offset - extremeR;
        double offsetRightSide =  translation_offset - extremeL;

        //translation des sommets
        forEach(pn, p.getPortNodes()) {
            n = pn->getCenter();
            Coord posn = result->getNodeValue(n);
            if (PortGraphRule::isMember(n, graph, PorgyConstants::SIDE_LEFT))
                posn.setX(posn.getX()+offsetLeftSide);
            else
                if (PortGraphRule::isMember(n, graph, PorgyConstants::SIDE_RIGHT))
                    posn.setX(posn.getX()+offsetRightSide);
            result->setNodeValue(n, posn);
        }
    }

        // Once the right hand side is drawn, we can use this function to draw the left hand side by symetry
        void drawLeftHandSide(PortGraphRule &p) const {
            BooleanProperty init(graph); // tmp property
            float extremeLeft = 0., minY=0., maxY=0.;
            float currentX = extremeLeft;
            float currentY = minY;

            init.setAllNodeValue(false);
            // Map that takes a left node in param and returns a map that contains the number of edges in common with each right node
            map<node, map<node, int> > edgeCounter;
            BridgePort *bp;
            forEach(bp, p.getBridge()->getBridgePorts()) {
                node left = p.getPortNode(bp->getLeft())->getCenter();
                node right  = p.getPortNode(bp->getRight())->getCenter();
                ++edgeCounter[left][right];
            }

            // search for the most connected node in the right hand side of the rule
            BOOST_FOREACH(const pairnodemapnodeint &p, edgeCounter) { //left node
                int maxEdges = -1;
                node rightNode;
                BOOST_FOREACH(const pairnodeint &pni, p.second) { //right node
                    rightNode = pni.first;
                    if ((pni.second >= maxEdges) && (init.getNodeValue(rightNode) == false)){
                        maxEdges = pni.second;
                    }
                }

                assert((maxEdges > -1)||(init.getNodeValue(rightNode)==true));
                if(!init.getNodeValue(rightNode)) {
                    Coord rightCoord = result->getNodeValue(rightNode);
                    rightCoord.setX(-rightCoord[0]);
                    result->setNodeValue(p.first, rightCoord);

                    maxY = max(maxY, rightCoord.getY());
                    minY = min(minY, rightCoord.getY());
                    extremeLeft = max(extremeLeft, result->getNodeValue(p.first).getX());

                    init.setNodeValue(p.first, true);
                    init.setNodeValue(rightNode, true);
                }
            }
            PortNode *pn;
            forEach(pn, p.getMember(PorgyConstants::SIDE_LEFT)) {
                node left = pn->getCenter();
                node right;
                Iterator<PortNode *> *it2 = p.getMember(PorgyConstants::SIDE_RIGHT);
                while(it2->hasNext()) {
                    PortNode *pn2 = it2->next();
                    right = pn2->getCenter();
                    if (pn->getName()==pn2->getName() && init.getNodeValue(right) == false){
                        init.setNodeValue(left, true);
                        init.setNodeValue(right, true);
                        Coord rightCoord = result->getNodeValue(right);
                        rightCoord.setX(-rightCoord[0]);
                        result->setNodeValue(left, rightCoord);

                        maxY = max(maxY, result->getNodeValue(left).getY());
                        minY = min(minY, result->getNodeValue(left).getY());
                        extremeLeft = max(extremeLeft, result->getNodeValue(left).getX());// + leftDeg* PortNode::space + 2);
                        break;
                    }
                }
                delete it2;
            }

            if (maxY - minY == 0){
                maxY = 10;
                minY = -maxY;
            }
            currentX = extremeLeft;
            currentY = minY;

            forEach(pn, p.getMember(PorgyConstants::SIDE_LEFT)) {
                node left = pn->getCenter();
                if (init.getNodeValue(left) == true)
                    continue;
                result->setNodeValue(left, Coord(currentX, currentY, 0));
                currentY += 2*PorgyConstants::SPACE;
                extremeLeft = max(extremeLeft, result->getNodeValue(left).getX());
                if (currentY >= maxY){
                    currentY = minY;
                    currentX = extremeLeft+1;
                }
            }
        }

        bool drawSide(Graph *g) const {
            string errMsg;

            DoubleProperty elength(g);
            elength.setAllEdgeValue(5);
            dataSet->set("edge length", &elength);
            if(!g->applyPropertyAlgorithm(PorgyConstants::GEM, result, errMsg, pluginProgress, dataSet)) {
                cerr << __PRETTY_FUNCTION__ << ": " << __LINE__ << "=> Pb avec GEM : " << errMsg << endl;
                return false;
            }

            //fast overlap removal
            dataSet->set("layout", result);
            dataSet->set<double>("x border", 3.0f);
            dataSet->set<double>("y border", 3.0f);
            if(!g->applyPropertyAlgorithm("Fast Overlap Removal", result, errMsg, pluginProgress, dataSet)) {
                cerr << __PRETTY_FUNCTION__ << ":" << __LINE__ << "Pb avec Fast Overlap Removal: " << errMsg << endl;
                return false;
            }
            result->perfectAspectRatio();
            return true;
        }

    void drawSameSizes(PortGraphRule &) const;

public:
    PLUGININFORMATION(PorgyConstants::REDRAW_RULE, "Bruno Pinaud", "14/03/12", "Comments", "1.0", PorgyConstants::CATEGORY_NAME)

    RedrawRule(const tlp::PluginContext *context):LayoutAlgorithm(context) {
        addDependency(PorgyConstants::GEM, PorgyConstants::GEM_VERSION);
        addDependency("Fast Overlap Removal", "1.3");

        addInOutParameter<SizeProperty>("node size", paramHelp[0],"viewSize");
        addInOutParameter<ColorProperty>("Color", paramHelp[1],"viewColor");
        addInOutParameter<ColorProperty>("Bordercolor", paramHelp[2],"viewBorderColor");
    }


    bool run() {
        PortGraphRule p(graph);
        Graph *sub = createAbstractGraph(p);

        //split graph in LHS, RHS and bridge
        pair<Graph *, Graph *> lhs_rhs(PortGraphRule::extractLHSRHS(sub,pluginProgress));

        //start by drawing RHS
        if(lhs_rhs.second!=NULL) {
            drawSide(lhs_rhs.second);
            //draw LHS if it exists (possible when editing a rule)
            if(lhs_rhs.first!=NULL)
                drawLeftHandSide(p);
        }
        //no RHS. Do a layout of LHS
        else if(lhs_rhs.first!=NULL) {
            drawSide(lhs_rhs.first);
        }
        else //nothing to do
            return true;

        PortGraphRule::cleanLHSRHS(sub);
        removeAbstractGraph(sub);

        SizeProperty *nodeSize = graph->getProperty<SizeProperty>("viewSize");
        ColorProperty *viewColor = graph->getProperty<ColorProperty>("viewColor");
        ColorProperty *bordercolor = graph->getProperty<ColorProperty>("viewBorderColor");

        if(dataSet!=NULL) {
            dataSet->get("node size", nodeSize);
            dataSet->get("Color", viewColor);
            dataSet->get("Bordercolor", bordercolor);
        }

        Translate(p, nodeSize);

    //    PortNode *pn;
    //    forEach(pn, p.getPortNodes()) {
    //        if(!Bridge::isBridge(pn->getCenter(), graph)) {
    //            pn->draw(nodeSize, viewColor, result, bordercolor);
    //        }
    //    }
        p.getBridge()->draw(nodeSize, viewColor, result, bordercolor);
        p.getBridge()->setPositions(result);
        p.drawSameSizes(nodeSize, result);

        return true;
    }

    bool check(string &errMsg) {

        if(!PorgyTlpGraphStructure::isPorgyGraph(graph, errMsg)) {
            errMsg = "Not a correct Porgy graph structure: "+errMsg;
            return false;
        }
        if(!PorgyTlpGraphStructure::isRuleGraph(graph)) {
            errMsg = "Graph is not a rule";
            return false;
        }

        return true;
    }

};

PLUGIN(RedrawRule)
