#ifndef LOG_H
#define LOG_H

#include <string>
#include <list>
#include <limits.h>
#include <stdio.h>
#include <iostream>
#include<sstream>
#include <stdarg.h>
#include <stdlib.h>

#include <QAbstractTableModel>
#include <QTime>

namespace tlp{
class Controller;
}

class QStatusBar;

class Log {
public:
    enum Level{INFO_LEVEL=0x1,WARN_LEVEL=0x2,ERROR_LEVEL=0x4,FATAL_LEVEL=0x8,DEBUG_LEVEL=0x10,TRACE_LEVEL=0x20};
    virtual ~Log(){}
    virtual bool writeLog(Level level,const std::string& msg) = 0;
};




/**
 * @brief The LogManager class handle all the Log objects stack. If you want to register a new logger just use :
 *
 * LogManager::getInstance()->addLogToList();
 *
 * To send a message to all logger use the writeInLogs static function or use sinfo, serror, swarning stream. See LogStream for more information.
 */
class LogManager : public Log
{
public :
    virtual ~LogManager();

    static LogManager* getInstance();
    static LogManager* instance;

    static bool writeInLogs(Level level,const std::string& msg);
    virtual bool writeLog(Level level,const std::string& msg);

    void addLogToList(Log* log);
    void removeLogFromList(Log* log);

private:
    QList<Log*> logList;
    LogManager();
    LogManager(const LogManager& logManager);



};

/**
 * @brief The LogStream class allows to handle log message like stream.
 */
class LogStream {
public:
    LogStream(Log::Level level):level(level){
        out.str("");
    }

    LogStream(const LogStream& copy):level(copy.level){
        out.str("");
    }

    virtual ~LogStream(){
        if(!out.str().empty()){
            flush();
        }
    }
    template<typename T>
    LogStream& operator<<(const T& object){
        out << object;
        return *this;
    }

    void flush(){
        LogManager::writeInLogs(level,out.str());
    }

    static inline LogStream sfatal(){
        return LogStream(Log::FATAL_LEVEL);
    }

    static inline LogStream serror(){
        return LogStream(Log::ERROR_LEVEL);
    }

    static inline LogStream swarn(){
        return LogStream(Log::WARN_LEVEL);
    }

    static inline LogStream sinfo(){
        return LogStream(Log::INFO_LEVEL);
    }

    static inline LogStream sdebug(){
        return LogStream(Log::DEBUG_LEVEL);
    }

    static inline LogStream strace(){
        return LogStream(Log::TRACE_LEVEL);
    }

private:
    Log::Level level;
    std::ostringstream out;
};

#define sinfo LogStream::sinfo()
#define serror LogStream::serror()
#define swarning LogStream::swarn()
#define sdebugwithline LogStream::sdebug()<<__PRETTY_FUNCTION__<<":"<<__LINE__<<" "


class STDLog : public Log {
    std::ostream& out;
    Level _levelFilter;
public:
    STDLog(std::ostream& out,Level levelFilter);
    virtual ~STDLog();
    virtual bool writeLog(Level level,const std::string& msg);
};


//Declare metatype for Log::Level to use it with QVariant
Q_DECLARE_METATYPE(Log::Level)
/**
 * @brief The LogModel class store the log message in a Qt Model.
 */
class LogModel : public QAbstractTableModel, public Log{

    struct LogMessage{
        LogMessage(Level level,const QString& message,QTime time):level(level),message(message),time(time),seen(false){}
        Level level;
        QString message;
        QTime time;
        bool seen;
    };

    struct LevelMessageSorter{
        Qt::SortOrder _order;
        LevelMessageSorter(Qt::SortOrder order):_order(order){}

        bool operator()(const LogMessage& a, const LogMessage& b) const
          {
            return _order==Qt::AscendingOrder? (a.level < b.level) : (a.level > b.level);
          }
    };

    struct TimeMessageSorter{
        Qt::SortOrder _order;
        TimeMessageSorter(Qt::SortOrder order):_order(order){}
        bool operator()(const LogMessage& a, const LogMessage& b) const
          {
            return _order==Qt::AscendingOrder? (a.time < b.time) : (a.time > b.time);
          }
    };

    QList<LogMessage> _messages;

public:

    enum LogModelRole{LevelRole=Qt::UserRole+1};
    LogModel(QObject* parent=NULL);

    bool writeLog(Level level, const std::string &msg){
        beginInsertRows(QModelIndex(),_messages.size(),_messages.size());
        _messages.push_back(LogMessage(level,QString::fromUtf8(msg.data()),QTime::currentTime()));
        endInsertRows();
        return true;
    }


    void clearMessages(){
        beginResetModel();
        _messages.clear();
        endResetModel();
    }

    int unseenMessageCount()const;
    void markAllSeen();

    int columnCount(const QModelIndex &) const{
        //Severity, Time, Message
        return 3;
    }
    int rowCount(const QModelIndex &parent) const;
    QVariant data(const QModelIndex &index, int role) const;
    QVariant headerData(int section, Qt::Orientation orientation, int role) const;

    void sort(int column, Qt::SortOrder order);

private:
    QColor color(Level)const;
    QIcon icon(Level)const;    
    bool compareLevelAsc(const LogMessage&l1, const LogMessage&l2){
        return l1.level < l2.level;
    }

    bool compareDateAsc(const LogModel::LogMessage&l1, const LogModel::LogMessage&l2){
        return l1.time < l2.time;
    }

    bool compareMessageAsc(const LogModel::LogMessage&l1, const LogModel::LogMessage&l2){
        return l1.message< l2.message;
    }
};


class LogStringBuf : public std::stringbuf{
    Log::Level _level;
public:
    LogStringBuf(Log::Level level):std::stringbuf(std::ios_base::out),_level(level){}
    int sync(){

        std::string s = str();
        std::string::size_type pos = str().find( '\n' );
        std::string::size_type initialPos = 0;

        // Decompose statement
        while( pos != std::string::npos ) {
            std::string ss = str().substr( initialPos, pos - initialPos );
            if(!ss.empty()){
                LogManager::writeInLogs(_level,ss);

            }
            initialPos = pos + 1;
            pos = str().find( '\n', initialPos );
        }
        // Add the last one
        std::string ss = str().substr( initialPos, std::min( pos, str().size() ) - initialPos );
        if(!ss.empty()){
            LogManager::writeInLogs(_level,ss);
        }
        str("");//Clear the buffer
        return 0;
    }
};

extern LogStringBuf OutputStream;
extern LogStringBuf errorStream;

extern std::streambuf* coutStream;
extern std::streambuf* cerrStream;

extern std::ostream consoleOut;
extern std::ostream consoleErrOut;

extern void redirectConsoleStreams();
extern void restoreConsoleStreams();


#endif // LOG_H
