/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef PORGYPERSPECTIVE_H
#define PORGYPERSPECTIVE_H

#include <tulip/Perspective.h>
#include <tulip/Coord.h>

#include <porgy/porgysettings.h>

#include <portgraph/PorgyConstants.h>

namespace tlp{
class Graph;
class GraphNeedsSavingObserver;
class View;
}

class PortNodeConfigurationWidget;
class PortNodeQt;

namespace Ui{
class PorgyPerspective;
}

class PorgyPerspective : public tlp::Perspective
{
    Q_OBJECT

    enum ViewManagerImplementation{MDI_AREA_IMPL,WORKSPACE_IMPL};

    /**
     * @brief Build all the GUI of the controllers, ie widgets, menu ect...
     **/
    void buildGUI();
    /**
     * @brief Build the menu bar of the Porgy controller.
     **/
    void initMenus();

    /**
     * @brief Reset all the internal data.
     **/
    void initData();

    void zoomOnTraceElementAnimation(const std::set<tlp::node>&elements,const QObject* object=NULL);

    bool eventFilter(QObject *obj, QEvent *event);    

    bool setData(tlp::Graph *graph,tlp::DataSet dataSet=tlp::DataSet());

    /**
     * @brief save saves the current state in the given file.
     * @param fileName
     */
    void save(const QString& fileName);

    tlp::Graph *fileOpen(const QString& file, tlp::PluginProgress *prg);

    /**
     * @brief getCurrentGraph Returns the graph of the current view or NULL if there is no view.
     * @return
     */
    tlp::Graph* getCurrentGraph();

    tlp::View* currentView();

    bool applyAlgorithm(const std::string& algorithm,tlp::Graph* graph,tlp::DataSet& parameters,std::string& errorMsg);
    bool changeProperty(const QString& name,tlp::Graph* graph,std::string& err);

    /**
     * @brief graphIsCompatibleWithView checks if the view with the given name is compatible with the graph.
     * @param graph The graph to test.
     * @param viewName The name of the view.
     * @return true if the view and the graph are compatible.
     */
    bool graphIsCompatibleWithView(tlp::Graph* graph,const std::string& viewName ) const;
    void open(tlp::PluginProgress *prg);

    Ui::PorgyPerspective* _ui;

    //Graph
    tlp::Graph* rootGraph;
    tlp::Graph* rulesGraph;
    tlp::Graph* modelsGraph;
    tlp::Graph* traceGraph;

    PortNodeConfigurationWidget* portNodeConfigurationWidget;
    tlp::GraphNeedsSavingObserver* savingObserver;

    PorgySettings settings;  

public:
    PLUGININFORMATION("Porgy","Porgy Team", "8/11/2010", "Porgy", "2","Porgy")
    PorgyPerspective(const tlp::PluginContext* context);
    virtual ~PorgyPerspective();

    void redrawPanels(bool center=false);
    void start(tlp::PluginProgress *progress);

    void getData(tlp::Graph **graph,tlp::DataSet *data);

    tlp::PluginProgress *progress(ProgressOptions options = ProgressOptions(IsPreviewable | IsStoppable | IsCancellable));



private slots:

    /**
     * @brief viewActivated is called when a view is activated in the view manager.
     * @param view
     */
    void viewActivated(tlp::View* view);
    void fileNew();

    /**
         * @brief fileOpen open a file and load data in the perspective.
         */
    void fileOpen();

    /**
         * @brief fileSave saves the modifications in the current edited file or in a new file if there is the first time the data are saved.
         **/
    void fileSave();
    /**
         * @brief fileSaveAs saves the modifications in a new file.
         */
    void fileSaveAs();

    void applyRuleOnModel(tlp::Graph* model,tlp::Graph* rule);
    void applyStrategy(QString code);
    void applyStrategyOnModel(tlp::Graph* model,QString strategy);
    void createNewRule();

    void showModel(tlp::Graph* graph);
    void showRule(tlp::Graph* graph);
    void showTrace(tlp::Graph* graph,bool useDefaultView=false);
    tlp::View* showPythonView(tlp::Graph* graph);
    tlp::View* showTableView(tlp::Graph* graph);

    void editDelSelection();
    void editReverseSelection();
    void editSelectAll();
    void editDeselectAll();
    void editCut();
    void ImportGraph();
    void editCopy();
    void editPaste(PorgyConstants::RuleSide side=PorgyConstants::NO_SIDE);
    void send_other_side(const PorgyConstants::RuleSide);

    void pastePortNodes(tlp::Coord &center,tlp::Graph*,QList<PortNodeQt*> portNodes);
    //rules handling
    void deleteRule(tlp::Graph* rule);
    void copyRule(tlp::Graph *);

    void changeMetric();
    bool applyAlgorithm();

    /**
     * @brief deleteTrace delete the trace represented by the given graph.
     * @param traceGraph the trace graph to delete
     */
    void deleteTrace(tlp::Graph* traceGraph);
    /**
     * @brief snapshot take a snapshot of the current view.
     */
    void snapshot();
    /**
     * @brief showDocumentation display the documentation dock widget.
     */
    void showDocumentation();

    void toggleViewManagerImplementation();

    void addPanel(tlp::Graph* graph);

    /**
     * @brief undo undo the last modifications.
     */
    void undo();

    /**
     * @brief redo redo last undo modifications.
     */
    void redo();

};

#endif // PORGYPERSPECTIVE_H
