/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "NewTrace.h"
#include "ui_NewTrace.h"

#include <portgraph/PorgyTlpGraphStructure.h>
#include <portgraph/Trace.h>
#include <portgraph/PortGraphModel.h>
#include <portgraph/PortNode.h>
#include <portgraph/Bridge.h>

#include <tulip/Graph.h>
#include <tulip/ConnectedTest.h>
#include <tulip/BooleanProperty.h>
#include <tulip/DoubleProperty.h>
#include <tulip/TlpQtTools.h>
#include <tulip/ColorProperty.h>

#include <QMessageBox>

using namespace tlp;
using namespace std;

Graph* NewTrace::createNewTrace(tlp::Graph* parentGraph,QWidget *parent){
    NewTrace trace(parent, parentGraph);
    Graph* graph = NULL;
    if(trace.exec()){
        Observable::holdObservers();
        graph = trace.buildNewTrace();
        Observable::unholdObservers();
    }
    return graph;
}

NewTrace::NewTrace(QWidget *parent, Graph *g) :QDialog(parent),graph(g), ui(new Ui::NewTrace)
{
    ui->setupUi(this);
    connect(ui->HistogramTree,SIGNAL(pressed()), this, SLOT(histogram_radio()));
    connect(ui->StandardTree,SIGNAL(pressed()), this, SLOT(standard_radio()));
    ui->StandardTree->setEnabled(true);
    ui->portSelectionComboBox->setEnabled(false);
    fillPortNodesCombobox();
}

NewTrace::~NewTrace()
{
    delete ui;
}

void NewTrace::histogram_radio() {
    ui->portSelectionComboBox->setEnabled(true);
}

void NewTrace::standard_radio() {
    ui->portSelectionComboBox->setEnabled(false);
}

Graph* NewTrace::buildNewTrace(){
    Graph *Traceroot = PorgyTlpGraphStructure::getTraceRoot(graph);
    Graph *TraceMain = PorgyTlpGraphStructure::getMainTrace(graph);
    Graph* newTraceGraph = Traceroot->addSubGraph(TraceMain->getProperty<BooleanProperty>("viewSelection"));
    newTraceGraph->setName(tlp::QStringToTlpString(ui->graphname->text()));
    if(ui->HistogramTree->isChecked()) {
        Trace trace(newTraceGraph);
        Observable::holdObservers();
        trace.followPortNode(QStringToTlpString(ui->portSelectionComboBox->currentText()));
        newTraceGraph->getLocalProperty<ColorProperty>("viewColor")->setAllNodeValue(Color(0,0,0));
        newTraceGraph->getLocalProperty<IntegerProperty>("viewShape")->setAllNodeValue(NodeShape::Circle);
        Observable::unholdObservers();
    }
    return newTraceGraph;
}

void NewTrace::accept() {
    Graph *TraceMain = PorgyTlpGraphStructure::getMainTrace(graph);
    Graph* g = tlp::newGraph();
    Observable::holdObservers();
    copyToGraph(g,TraceMain,TraceMain->getProperty<BooleanProperty>("viewSelection"));
    Observable::unholdObservers();
    //Check number of nodes
    if(g->numberOfNodes()==0) {
        QMessageBox::critical(this,"Create a new derivation tree", "Cannot create an empty graph.");
        delete g;
        QDialog::reject();
        return;
    }
    //Check connectivity
    if(!ConnectedTest::isConnected(g)){
        QMessageBox::critical(this,"Create a new derivation tree", "Error during the creation of the subgraph. Is it connected?");
        QDialog::reject();
        return;
    }
    QDialog::accept();
    return;
}

void NewTrace::fillPortNodesCombobox(){
    QSet<QString> portNodes;
    //Search in models
    fillPortNodesForGraph(PorgyTlpGraphStructure::getModelsRoot(graph),portNodes);
    //Search in rules
    fillPortNodesForGraph(PorgyTlpGraphStructure::getRulesRoot(graph),portNodes);

    ui->portSelectionComboBox->clear();
    QStringList nodesNames = QStringList::fromSet(portNodes);
    nodesNames.sort();
    ui->portSelectionComboBox->addItems(nodesNames);

}


void NewTrace::fillPortNodesForGraph(tlp::Graph* g,QSet<QString>& list){
    Graph *sub;
    forEach (sub, g->getSubGraphs()) {
        PortGraphModel portGraph(sub);
        PortNode *pn;
        forEach(pn, portGraph.getPortNodes()) {
            QString name = tlp::tlpStringToQString(pn->getName());
            if(!list.contains(name) && !Bridge::isBridge(pn->getCenter(),portGraph.getGraph()))
                list.insert(name);
        }
    }
}
