#include "helpwidget.h"
#include <ui_helpwidget.h>
#include <QHelpEngine>
#include <QHelpContentWidget>
#include <QHelpIndexWidget>

#include <QLineEdit>

#include <QDebug>
#include <tulip/TlpTools.h>

#include <cassert>


namespace {

const QString WindowTitleKey(QLatin1String("WindowTitle"));
const QString DefaultHomePageKey(QLatin1String("defaultHomepage"));
const QString widgetNoFound =QString("<html><head><meta http-equiv=\""
                                     "content-type\" content=\"text/html; charset=UTF-8\"><title>Error 404...</title>"
                                     "</head><body><div align=\"center\"><br><br><h1>No help found for the pointed object.</h1>"
                                     "<br><h3>There is no help for the given object. Try to search it in the index.</h3></div></body>");

} // anonymous namespace

//No more needed as not in a static lib
//struct HelpWidgetResourceInitializer
//{
//    HelpWidgetResourceInitializer()
//    {
//         Q_INIT_RESOURCE(HelpWidget);
//    }
//};

//extern "C"{
//HelpWidgetResourceInitializer staticLibInitializer;
//}



QtHelpTextBrowser::QtHelpTextBrowser(QWidget* parent):QTextBrowser(parent),_helpEngine(NULL)
{
}

QVariant QtHelpTextBrowser::loadResource ( int type, const QUrl & url ){
    if ( _helpEngine!= NULL && url.scheme() == "qthelp")
        return QVariant(_helpEngine->fileData(url));
    else
        return QTextBrowser::loadResource(type, url);
}

void QtHelpTextBrowser::setHelpEngineCore(QHelpEngineCore* helpEngine){
    _helpEngine = helpEngine;
}




HelpWidget::HelpWidget(QWidget *parent) :
    QWidget(parent),
    _ui(new Ui::HelpWidget),_helpEngine(NULL),_indexWidget(NULL),_indexFilter(NULL)
{
    _ui->setupUi(this);

    _ui->backHistoryPushButton->setIcon(QIcon::fromTheme("go-previous",QIcon(":/HelpWidget/go-previous.png")));
    _ui->nextHistoryPushButton->setIcon(QIcon::fromTheme("go-next",QIcon(":/HelpWidget/go-next.png")));
    _ui->homePushButton->setIcon(QIcon::fromTheme("go-home",QIcon(":/HelpWidget/go-home.png")));
    _ui->searchPushButton->setIcon(QIcon::fromTheme("system-search",QIcon(":/HelpWidget/system-search.png")));

    //Force the combobox to have the same size than the navigation control widget.
    _ui->comboBox->setMinimumHeight(_ui->backHistoryPushButton->height());
    _ui->comboBox->setMaximumHeight(_ui->backHistoryPushButton->height());

    connect(_ui->homePushButton,SIGNAL(clicked(bool)),SLOT(goHomePage()));
    connect(_ui->backHistoryPushButton,SIGNAL(clicked(bool)),SLOT(goPreviousPage()));
    connect(_ui->nextHistoryPushButton,SIGNAL(clicked(bool)),SLOT(goNextPage()));

    connect(_ui->comboBox,SIGNAL(currentIndexChanged(int)),_ui->stackedWidget,SLOT(setCurrentIndex(int)));

    connect(_ui->tabWidget,SIGNAL(currentChanged(int)),SLOT(updateBackwardForwardButtonState()));
    connect(_ui->tabWidget,SIGNAL(tabCloseRequested(int)),SLOT(tabCloseRequested(int)));
    connect(_ui->searchPushButton,SIGNAL(toggled(bool)),this,SLOT(showHideSearchPanel(bool)));

    showHideSearchPanel(_ui->searchPushButton->isChecked());
    setEnabled(false);
}

HelpWidget::~HelpWidget()
{
    delete _ui;
    delete _helpEngine;
}


bool HelpWidget::setCollectionFile(const QString& collectionFile){

    if(_helpEngine){
        _ui->comboBox->clear();
        //Clear the stacked widget.
        while(_ui->stackedWidget->count()!=0){
            _ui->stackedWidget->removeWidget(_ui->stackedWidget->widget(0));
        }
        //Delete old widgets
        _helpEngine->deleteLater();
        _helpEngine=NULL;
        _indexFilter->deleteLater();
        _indexWidget = NULL;
        _indexWidget->deleteLater();
        _indexFilter= NULL;
    }

    _helpEngine = new QHelpEngine(collectionFile,this);
    if(_helpEngine->setupData()){
        //Init documentation navigation widgets
        _ui->comboBox->addItem(QString("Index"));
        //Add a filter filed to the default index widget.
        _indexWidget = new QWidget(_ui->stackedWidget);
        _indexWidget->setObjectName("IndexWidget");
        QVBoxLayout * vboxLayout = new QVBoxLayout(_indexWidget);
        vboxLayout->setContentsMargins(0,0,0,0);
        vboxLayout->setSpacing(0);
        _indexWidget->setLayout(vboxLayout);
        _indexFilter = new QLineEdit(_indexWidget);
        _indexFilter->setObjectName("IndexFilter");
        _indexFilter->setPlaceholderText(tr("Search in index"));
        _indexWidget->layout()->addWidget(_indexFilter);
        _indexWidget->layout()->addWidget(_helpEngine->indexWidget());
        _ui->stackedWidget->addWidget(_indexWidget);
        connect(_indexFilter,SIGNAL(textChanged(QString)),_helpEngine->indexWidget(),SLOT(filterIndices(QString)));
        connect(_indexFilter,SIGNAL(returnPressed()),_helpEngine->indexWidget(),SLOT(activateCurrentItem()));

        connect(_helpEngine->indexWidget(),SIGNAL(linkActivated(QUrl,QString)),this,SLOT(linkActivated(QUrl,QString)));
        connect(_helpEngine->indexWidget(),SIGNAL(linksActivated(QMap<QString,QUrl>,QString)),this,SLOT(linksActivated(QMap<QString,QUrl>,QString)));
        _ui->comboBox->addItem(QString("Content"));
        _ui->stackedWidget->addWidget(_helpEngine->contentWidget());
        connect(_helpEngine->contentWidget(),SIGNAL(linkActivated(QUrl)),this,SLOT(linkActivated(QUrl)));
        _ui->stackedWidget->setCurrentIndex(0);

        setEnabled(true);
        return true;
    }else{
        _helpEngine->deleteLater();
        _helpEngine=NULL;
        _indexFilter->deleteLater();
        _indexWidget = NULL;
        _indexWidget->deleteLater();
        _indexFilter= NULL;
        setEnabled(false);
        return false;
    }
}

void HelpWidget::linkActivated ( const QUrl & link, const QString &  ){
    showLink(link,NewTab);
}

void HelpWidget::linksActivated ( const QMap<QString, QUrl> & , const QString &  ){

}

void HelpWidget::linkActivated ( const QUrl &  link){
    showLink(link,NewTab);
}
void HelpWidget::showLink(const QUrl & link,Target targetTab){
    QtHelpTextBrowser *browser = NULL;
    int index = -1;
    //If the user want to display in the current index and if the current index is valid.
    if(targetTab == CurrentTab && _ui->tabWidget->currentIndex() != -1){
        browser = qobject_cast<QtHelpTextBrowser*>(_ui->tabWidget->currentWidget());
        index = _ui->tabWidget->currentIndex();
    }else{
        browser = new QtHelpTextBrowser(_ui->tabWidget);
        connect(browser,SIGNAL(historyChanged()),SLOT(historyChanged()));
        browser->setHelpEngineCore(_helpEngine);
        index = _ui->tabWidget->addTab(browser,link.toString());
    }

    browser->setSource(link);
    _ui->tabWidget->setTabText(index,browser->documentTitle());
    _ui->tabWidget->setCurrentIndex(index);
}

void HelpWidget::goHomePage(){
    showLink(_helpEngine->customValue(DefaultHomePageKey, QLatin1String("help")).toUrl(),CurrentTab);
}

QString HelpWidget::collectionFileTitle()const{
    return _helpEngine->customValue(WindowTitleKey,QLatin1String("help")).toString();
}

void HelpWidget::goPreviousPage(){
    QtHelpTextBrowser *browser = qobject_cast<QtHelpTextBrowser*>(_ui->tabWidget->currentWidget());
    if(browser ){
        browser->backward();
    }

}

void HelpWidget::goNextPage(){
    QtHelpTextBrowser *browser = qobject_cast<QtHelpTextBrowser*>(_ui->tabWidget->currentWidget());
    if(browser ){
        browser->forward();
    }
}

void HelpWidget::updateBackwardForwardButtonState(){
    QtHelpTextBrowser *browser = qobject_cast<QtHelpTextBrowser*>(_ui->tabWidget->currentWidget());
    _ui->backHistoryPushButton->setEnabled(browser != NULL && browser->isBackwardAvailable());
    _ui->nextHistoryPushButton->setEnabled(browser != NULL && browser->isForwardAvailable());
    _ui->homePushButton->setEnabled(browser != NULL);
}

void HelpWidget::tabCloseRequested ( int index ){
    QtHelpTextBrowser *browser =  qobject_cast<QtHelpTextBrowser*>(_ui->tabWidget->widget(index));
    _ui->tabWidget->removeTab(index);
    browser->deleteLater();

}

void HelpWidget::historyChanged(){
    if(sender() == _ui->tabWidget->currentWidget()){
        updateBackwardForwardButtonState();
    }
}

void HelpWidget::showHideSearchPanel(bool isVisible){
    _ui->stackedWidget->setVisible(isVisible);
    _ui->comboBox->setVisible(isVisible);

}
