/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "modelswidget.h"
#include "ui_modelswidget.h"
#include "modelgraphtreemodel.h"
#include "sortfilterproxymodel.h"

#include <tulip/Graph.h>
#include <tulip/LayoutProperty.h>

#include <QHelpEvent>

#include <qtimagetooltip.h>

using namespace tlp;
using namespace std;

ModelsWidget::ModelsWidget(QWidget *parent) : QWidget(parent), ui(new Ui::ModelsWidget),_model(NULL)
{
    ui->setupUi(this);
    setContextMenuPolicy(Qt::ActionsContextMenu);

    connect(ui->modelsTreeView,SIGNAL(doubleClicked(QModelIndex)),SLOT(indexDoubleClicked(QModelIndex)));
    ui->modelsTreeView->installEventFilter(this);
    ui->modelsTreeView->setContextMenuPolicy(Qt::ActionsContextMenu);
    //Intialize proxy
    _proxyModel = new SortFilterProxyModel(this);
    _proxyModel->setDynamicSortFilter(true);
    _proxyModel->setFilterOrder(SortFilterProxyModel::BottomToTop);

    //Initialize model
    _model = new ModelGraphTreeModel(this);
    GlGraphRenderingParameters parameters;
    parameters.setEdgeColorInterpolate(false);
    parameters.setLabelsDensity(0);
    parameters.setLabelScaled(true);
    _model->snapshotManager().setRenderingParameters(parameters);
    _model->snapshotManager().setPreviewSize(QSize(32,32));
    _model->setIconColumn(0);

    connect(ui->comboBox,SIGNAL(currentIndexChanged(int)),this,SLOT(updateIconSize()));

    _proxyModel->setSourceModel(_model);
#ifndef NDEBUG
    //If in debug mode the name column is set to 1
    _proxyModel->setFilterKeyColumn(1);
#endif

    connect(ui->filterModelLineEdit,SIGNAL(textEdited(QString)),this,SLOT(filterModels(QString)));
    ui->modelsTreeView->setModel(_proxyModel);
}

ModelsWidget::~ModelsWidget()
{
    delete ui;
}

void ModelsWidget::setGraph(Graph* graph){
    _graph = graph;
    _model->setRootGraph(graph);
}


QList<Graph*> ModelsWidget::selectedModels()const{
    QModelIndexList indexes = ui->modelsTreeView->selectionModel()->selectedRows();
    QModelIndexList sourcesIndexes;
    sourcesIndexes.reserve(indexes.size());
    foreach(const QModelIndex m, indexes) {
        sourcesIndexes.push_back(_proxyModel->mapToSource(m));
    }
    return _model->indexesGraphs(sourcesIndexes);
}

void ModelsWidget::indexDoubleClicked(const QModelIndex& index){
    emit showModel(_model->indexGraph(_proxyModel->mapToSource(index)));
}

bool ModelsWidget::eventFilter(QObject *obj, QEvent *evt){
    switch(evt->type()){
    case QEvent::ToolTip:
    {

        if(obj == ui->modelsTreeView){
            //Need to use viewport <idget to compute correct coordinate tranformation.
            //Return wrong QModelIndex if we use he->Pos();
            QHelpEvent *he = static_cast<QHelpEvent *>(evt);
            Graph *graph = _model->indexGraph(_proxyModel->mapToSource(ui->modelsTreeView->indexAt(ui->modelsTreeView->viewport()->mapFromGlobal(he->globalPos()))));
            if(graph){
                GlGraphRenderingParameters parameters;
                parameters.setEdgeColorInterpolate(false);
                parameters.setLabelsDensity(0);
                parameters.setLabelScaled(true);

                QtImageToolTip::showGraph(he->globalPos(), graph,QSize(512,512),Color(255,255,255),parameters,NULL,this,QRect());
                return true;
            }
        }
    }
        break;
    default:
        break;
    }

    return QObject::eventFilter(obj,evt);
}

void ModelsWidget::filterModels(const QString& pattern){
    _proxyModel->setFilterWildcard(pattern);
}

void ModelsWidget::updateIconSize(){
    int iconsSize = ui->comboBox->currentIndex();
    if(iconsSize > 0){
        QSize iconSize;
        if(iconsSize == 1 ){
            iconSize = QSize(32,32);
        }else if(iconsSize == 2){
            iconSize = QSize(64,64);
        }else{
            iconSize = QSize(128,128);
        }
        _model->snapshotManager().setPreviewSize(iconSize);
        _model->setIconColumn(0);
    }else{
        _model->setIconColumn(-1);
    }
}
