/**
 *
 *
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <tulip/TlpQtTools.h>
#include <tulip/DrawingTools.h>
#include <tulip/GlMainView.h>
#include <tulip/SnapshotDialog.h>
#include <tulip/GraphNeedsSavingObserver.h>
#include <tulip/TulipProject.h>
#include <tulip/View.h>
#include <tulip/Graph.h>
#include <tulip/TlpTools.h>
#include <tulip/ImportModule.h>

#include <QParallelAnimationGroup>
#include <QMessageBox>
#include <QInputDialog>
#include <QClipboard>
#include <QFileDialog>
#include <QHelpEngine>
#include <QFileInfo>
#include <QKeyEvent>

#include "ui_porgyperspective.h"

#include "porgyperspective.h"
#include "log.h"
#include "graphmodificationlistener.h"
#include "special_paste.h"
#include "statusbarpluginprogress.h"
#include "GlMainWidgetZoomAndPanAnimation.h"

#include <porgy/pluginselectionwizard.h>
#include <porgy/pluginparametersconfigurationdialog.h>
#include <porgy/PortNodeQt.h>

#include <portgraph/PorgyTlpGraphStructure.h>
#include <portgraph/PortGraphRule.h>
#include <portgraph/Bridge.h>
#include <portgraph/PortGraphModel.h>
#include <portgraph/Trace.h>

#include <boost/foreach.hpp>

using namespace tlp;
using namespace std;

Log::Level convertQtMsgTypeToLogLevel(QtMsgType type){
    switch(type){
    case QtDebugMsg:
        return Log::DEBUG_LEVEL;
        break;
    case QtWarningMsg:
        return Log::WARN_LEVEL;
        break;
    case QtCriticalMsg:
        return Log::ERROR_LEVEL;
        break;
    case QtFatalMsg:
        return Log::FATAL_LEVEL;
        break;
    }
    return Log::INFO_LEVEL;
}

void perspectiveLogger(QtMsgType type, const char* msg) {
    LogManager::getInstance()->writeInLogs(convertQtMsgTypeToLogLevel(type),msg);
}

static const QString porgyHelpCollectionFile(QString::fromStdString(TulipShareDir)+"Porgy.qhc");
static const QString porgyModelParameter("model");
static const QString GRAPHS_PATH("/graphs/");
static const QString strategies_path("/strategies/");

PorgyPerspective::PorgyPerspective(const tlp::PluginContext* context):Perspective(context),_ui(NULL),rootGraph(NULL),rulesGraph(NULL),modelsGraph(NULL),traceGraph(NULL),portNodeConfigurationWidget(NULL), savingObserver(NULL) {
    addDependency(PorgyConstants::GRAPH_VIEW_NAME, "1.0");
    addDependency(PorgyConstants::RULE_VIEW_NAME, "1.0");
    addDependency(PorgyConstants::TRACE_VIEW_NAME, "1.0");
    addDependency(PorgyConstants::SMALL_MULTIPLE_VIEW_NAME, "1.0");
    addDependency(PorgyConstants::APPLY_STRATEGY_ALGORITHM, "1.0");
    addDependency(PorgyConstants::SIMPLECHECKAPPLY, "1.0");
}

void PorgyPerspective::redrawPanels(bool){
    _ui->viewManager->drawAllViews();
}

void PorgyPerspective::start(tlp::PluginProgress *progress){
    _ui = new Ui::PorgyPerspective();
    _ui->setupUi(_mainWindow);
    //    redirectConsoleStreams();//Redirect streams
    qInstallMsgHandler(perspectiveLogger);
    //Initialize log system
    LogModel* logModel = new LogModel(this);
    LogManager::getInstance()->addLogToList(logModel);//Register log
    _ui->logWidget->setLogModel(logModel);//Set widget data


    initMenus();
    buildGUI();

    _mainWindow->installEventFilter(this);

    //Settings
    settings.setTraceViewZoomAnimationTime(1000);

    _ui->viewManager->setImplementation(new WorkspacePanelImplementation());
    _ui->actionToggleViewManagerImplementation->setData(WORKSPACE_IMPL);

    if(!_externalFile.isEmpty()) {
        if(QFileInfo(_externalFile).exists()) {
            fileOpen(_externalFile,progress);
        }
        else{
            serror<<"The file "<<_externalFile.toUtf8().data()<<" does not exists";
        }
    }

    if(!_project->projectFile().isEmpty()) {
        open(progress);
    }
    showTrayMessage("Porgy Perspective started");
}

void PorgyPerspective::initMenus(){
    connect(_ui->actionNew,SIGNAL(triggered()),SLOT(fileNew()));
    connect(_ui->actionOpen,SIGNAL(triggered()),SLOT(fileOpen()));
    connect(_ui->actionSave,SIGNAL(triggered()),SLOT(fileSave()));
    connect(_ui->actionSave_as,SIGNAL(triggered()),SLOT(fileSaveAs()));

    connect(_ui->actionUndo,SIGNAL(triggered()),SLOT(undo()));
    connect(_ui->actionRedo,SIGNAL(triggered()),SLOT(redo()));
    connect(_ui->actionCut,SIGNAL(triggered()),SLOT(editCut()));
    connect(_ui->actionCopy,SIGNAL(triggered()),SLOT(editCopy()));
    connect(_ui->actionPaste,SIGNAL(triggered()),SLOT(editPaste()));
    connect(_ui->actionSelect_all,SIGNAL(triggered()),SLOT(editSelectAll()));
    connect(_ui->actionDeselect_all,SIGNAL(triggered()),SLOT(editDeselectAll()));
    connect(_ui->actionInvert_selection,SIGNAL(triggered()),SLOT(editReverseSelection()));
    connect(_ui->actionDel_selection,SIGNAL(triggered()),SLOT(editDelSelection()));
    connect(_ui->actionCreate_new_rule,SIGNAL(triggered()),SLOT(createNewRule()));

    connect(_ui->actionDocumentation,SIGNAL(triggered()),SLOT(showDocumentation()));
    connect(_ui->actionFull_Screen, SIGNAL(triggered(bool)), this, SLOT(showFullScreen(bool)));
    connect(_ui->actionQuit,SIGNAL(triggered()),_mainWindow,SLOT(close()));
    connect(_ui->action_Close_All, SIGNAL(triggered()), _ui->viewManager, SLOT(closeAllView()));

    connect(_ui->actionToggleViewManagerImplementation,SIGNAL(triggered()),SLOT(toggleViewManagerImplementation()));

    QMenu *algomenu = _ui->menuAlgorithms->addMenu(tlp::tlpStringToQString(tlp::ALGORITHM_CATEGORY));
    QMenu *algoDebug = algomenu->addMenu("Debug");
    QMenu *layoutmenu = _ui->menuAlgorithms->addMenu(tlp::tlpStringToQString(tlp::LAYOUT_ALGORITHM_CATEGORY));
    QMenu *boolmenu = _ui->menuAlgorithms->addMenu(tlp::tlpStringToQString(tlp::BOOLEAN_ALGORITHM_CATEGORY));
    QMenu *importMenu = _ui->menuFile->addMenu(tlp::tlpStringToQString(tlp::IMPORT_CATEGORY));
    BOOST_FOREACH(const string &pName, PluginLister::instance()->availablePlugins<Plugin>()) {
        tlp::Plugin* plugin = PluginLister::instance()->getPluginObject(pName,NULL);
        string className = plugin->category();
        if(plugin->group()=="Porgy"){
            if (className == tlp::LAYOUT_ALGORITHM_CATEGORY)
                layoutmenu->addAction(plugin->name().c_str(), this, SLOT(changeMetric()));
            else if (className == tlp::ALGORITHM_CATEGORY)
                algomenu->addAction(plugin->name().c_str(), this, SLOT(applyAlgorithm()));
            else if (className == tlp::BOOLEAN_ALGORITHM_CATEGORY)
                boolmenu->addAction(plugin->name().c_str(), this, SLOT(changeMetric()));
            else if (className == tlp::IMPORT_CATEGORY)
                importMenu->addAction(plugin->name().c_str(), this, SLOT(ImportGraph()));
        }
        else if(plugin->group()=="Porgy_internal") {
            if (className == tlp::ALGORITHM_CATEGORY)
                algoDebug->addAction(plugin->name().c_str(), this, SLOT(applyAlgorithm()));
        }
        delete plugin;
    }

    connect(_ui->actionSnapshot,SIGNAL(triggered()),this,SLOT(snapshot()));
    //Docks menu
    foreach(QDockWidget* child , _mainWindow->findChildren<QDockWidget*>()){
        _ui->menuDocks->addAction(child->toggleViewAction());
    }
}

void PorgyPerspective::buildGUI(){
    //Build code management widget
    connect(_ui->strategiesManagementWidget,SIGNAL(executeStrategy(QString)),this,SLOT(applyStrategy(QString)));

    //Building element properties management widget
    connect(_ui->viewManager,SIGNAL(showGraphRequest(tlp::Graph*)),SLOT(addPanel(tlp::Graph*)));
    connect(_ui->viewManager,SIGNAL(currentViewChanged(tlp::View*)),SLOT(viewActivated(tlp::View*)));
    connect(_ui->rulesWidget,SIGNAL(showRule(tlp::Graph*)),this,SLOT(showRule(tlp::Graph*)));
    connect(_ui->rulesWidget,SIGNAL(addNewRule()),this,SLOT(createNewRule()));
    connect(_ui->rulesWidget,SIGNAL(deleteRule(tlp::Graph*)),this,SLOT(deleteRule(tlp::Graph*)));
    connect(_ui->rulesWidget,SIGNAL(copyRule(tlp::Graph *)),this,SLOT(copyRule(tlp::Graph *)));

    connect(_ui->graphsWidget,SIGNAL(showModel(tlp::Graph*)),this,SLOT(showModel(tlp::Graph*)));

    connect(_ui->tracesWidget,SIGNAL(showTrace(tlp::Graph*,bool)),this,SLOT(showTrace(tlp::Graph*)));
    connect(_ui->tracesWidget,SIGNAL(deleteTrace(tlp::Graph*)),this,SLOT(deleteTrace(tlp::Graph*)));

    _mainWindow->tabifyDockWidget(_ui->rulesDockWidget,_ui->graphsDockWidget);//Group graphs, rules and traces dock widget
    _mainWindow->tabifyDockWidget(_ui->rulesDockWidget,_ui->tracesDockWidget);
    _ui->graphsDockWidget->raise();//Display graph dock widget

    //Load help widget data
    if(!_ui->helpWidget->setCollectionFile(porgyHelpCollectionFile)){//Porgy documentation was not loaded correctly
        serror<<" Cannot find porgy help collection cannot use help";
        _ui->menuDocks->removeAction(_ui->helpDockWidget->toggleViewAction());//Avoid user to enable it
    }

    //Hide these widgets
    _ui->logDockWidget->setVisible(false);
    _ui->helpDockWidget->setVisible(false);

}


PorgyPerspective::~PorgyPerspective() {
    if(_ui != NULL){
        LogManager::getInstance()->removeLogFromList(_ui->logWidget->logModel());
        delete _ui;
    }
    delete savingObserver;
}

void PorgyPerspective::fileNew(){
    setData(PorgyTlpGraphStructure::newPorgyStructure());
}

void PorgyPerspective::open(PluginProgress *prg) {
    QStringList entries = _project->entryList(GRAPHS_PATH,QDir::Dirs | QDir::NoDot | QDir::NoDotDot, QDir::Name);
    QString file = GRAPHS_PATH + entries.first() + "/graph.tlp";
    if (!_project->exists(file))
        return;
    QString absolutePath = _project->toAbsolutePath(file);
    Graph* g = fileOpen(absolutePath,prg);
    if(g!=NULL) {
        QMap<QString,tlp::Graph*> rootIds;
        rootIds[entries.first()]= g;
        _ui->viewManager->restoreFromProject(_project,rootIds,prg);
        _ui->strategiesManagementWidget->importStrategies(_project->toAbsolutePath(strategies_path));
        //reconnect signals
        foreach(View *v, _ui->viewManager->views()) {
            if(v->name()==PorgyConstants::TRACE_VIEW_NAME||v->name()==PorgyConstants::GRAPH_VIEW_NAME) {
                connect(v,SIGNAL(applyRuleOnModel(tlp::Graph*,tlp::Graph*)),SLOT(applyRuleOnModel(tlp::Graph*,tlp::Graph*)));
                connect(v,SIGNAL(applyStrategyOnModel(tlp::Graph*,QString)),SLOT(applyStrategyOnModel(tlp::Graph*,QString)));
            }
            if(v->name()==PorgyConstants::TRACE_VIEW_NAME) {
                connect(v,SIGNAL(openModelView(tlp::Graph *)),this,SLOT(showModel(tlp::Graph *)));
            }
            if(v->name()==PorgyConstants::RULE_VIEW_NAME) {
                connect(v,SIGNAL(copy_paste_other_side(PorgyConstants::RuleSide)),SLOT(send_other_side(PorgyConstants::RuleSide)));
            }
            if((v->name()==PorgyConstants::RULE_VIEW_NAME)||(v->name()==PorgyConstants::GRAPH_VIEW_NAME)) {
                connect(v,SIGNAL(pastePortNodes(tlp::Coord &,tlp::Graph*,QList<PortNodeQt*>)),SLOT(pastePortNodes(tlp::Coord&,tlp::Graph*,QList<PortNodeQt*>)));
            }
        }
    }
}

void PorgyPerspective::fileOpen(){
    QString filename = QFileDialog::getOpenFileName(_mainWindow,tr("Open a file"),QString(),"All supported formats(*.tlpx *.tlp *.tlp.gz);;Tulip project (*.tlpx);;Tulip graph (*.tlp);;Gzipped Tulip graph(*.tlp.gz)");
    if(QFile::exists(filename)) {
        PluginProgress* prg = progress();
        if(rootGraph==NULL) {

            if(!filename.endsWith(".tlpx")){//load old porgy files
                fileOpen(filename,prg);
            }else{
                delete _project;
                _project = TulipProject::openProject(filename,prg);
                open(prg);
            }

        }
        else if(!filename.endsWith(".tlpx")){//load old porgy files
            delete _project;
            _project=NULL;
           // _ui->viewManager->closeAllView();
            fileOpen(filename,prg);
        }
        else
            Perspective::openProjectFile(filename);
        delete prg;
    }
}

Graph* PorgyPerspective::fileOpen(const QString& filename,PluginProgress *prg){
    tlp::DataSet dataSet;
    dataSet.set<std::string>("file::filename", std::string(filename.toUtf8().data()));
    Graph *graph = tlp::importGraph("TLP Import", dataSet,prg);
    if(graph){
        tlp::DataSet controllerData;
        if(dataSet.exist("controller")) {//Old style file
            DataSet controller;
            dataSet.get<DataSet>("controller", controller);
            Iterator< std::pair<std::string, DataType*> > *it=controller.getValues();
            assert(it->hasNext());
            pair<string, DataType*> p;
            p = it->next();
            delete it;
            controllerData=*((DataSet*)p.second->value);
        }
        if(setData(graph,controllerData)){
            return graph;
        }
    }
    return NULL;
}

void PorgyPerspective::fileSave(){
    if((_project!=NULL)&&QFile::exists(_project->projectFile())){
        save(_project->projectFile());
    }else{
        fileSaveAs();
    }
}

void PorgyPerspective::fileSaveAs(){
    QString outFile = QFileDialog::getSaveFileName(_mainWindow,tr("Choose the destination file"),QString(),"Tulip project (*.tlpx);;Tulip graph (*.tlp);;Tulip gzipped graph (*.tlp.gz)");
    if(!outFile.isEmpty()){
        save(outFile);
    }
}

/**
 * @brief PorgyPerspective::save the code of these function is extracted from the graph perspective to try to have a little compatibility.
 * @param fileName
 */
void PorgyPerspective::save(const QString& fileName){
    tlp::PluginProgress* p=progress();
    if(fileName.endsWith(".tlpx")){//Tulip project

        if(_project==NULL)
            _project = TulipProject::newProject();
        _project->removeAllDir(GRAPHS_PATH);//Clear old data
        _project->mkpath(GRAPHS_PATH);//Create a new graph structure
        QMap<Graph*,QString> rootIds;

        rootIds[rootGraph] = QString::number(0);//Save the root graph
        QString folder = GRAPHS_PATH + "/" + QString::number(0) + "/";
        _project->mkpath(folder);
        tlp::saveGraph(rootGraph,_project->toAbsolutePath(folder + "graph.tlp").toStdString(),p);

        _project->removeAllDir(strategies_path);//Clear old data
        _project->mkpath(strategies_path);//Create a new strategies structure
        _ui->strategiesManagementWidget->exportStrategies(_project->toAbsolutePath(strategies_path));//Export strategies

        _ui->viewManager->saveToProject(_project,rootIds,p);
        _project->write(fileName,p);
        savingObserver->saved();

    }else{//Simple tlp file
       QMessageBox::StandardButton button = (QMessageBox::StandardButton)QMessageBox::warning(_mainWindow,"Old Tulip file format","Warning: you are going to save your project with the old Tulip graph format. Strategies will be lost",QMessageBox::Save,QMessageBox::Abort);
        if(button==QMessageBox::Save) {
            tlp::saveGraph(rootGraph,fileName.toStdString(),p);
            savingObserver->saved();
        }
    }
    delete p;
}


bool PorgyPerspective::setData(Graph *graph,DataSet dataSet){
    if(rootGraph != NULL){//If there is some previous data
        if(savingObserver->needsSaving()) { //do we need to save the previous graph
          QMessageBox::StandardButton answer = QMessageBox::question(_mainWindow, trUtf8("Save"), trUtf8("The project has been modified, do you want to save your changes ?"),QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel | QMessageBox::Escape);
          if(answer == QMessageBox::Yes)
              fileSave();
          if(answer==QMessageBox::Cancel) {
              delete graph;
              return false;
          }
        }
    delete savingObserver;
    }
    _ui->viewManager->closeAllView();
    tlp::Graph* oldGraph = rootGraph;//Keep old graph to delete it later


    std::string errorMsg;
    if(!PorgyTlpGraphStructure::isPorgyGraph(graph,errorMsg)){
        serror<<"This graph is not a porgy structure: "<<errorMsg;
        return false;
    }

    //The structure is valid load data
    rootGraph = graph;
    savingObserver = new GraphNeedsSavingObserver(rootGraph);

    //Retrieve graphs.
    rulesGraph = PorgyTlpGraphStructure::getRulesRoot(graph);
    modelsGraph =PorgyTlpGraphStructure::getModelsRoot(graph);
    traceGraph = PorgyTlpGraphStructure::getTraceRoot(graph);

    _ui->graphsWidget->setGraph(modelsGraph);
    _ui->graphsWidget->update();

    _ui->rulesWidget->setGraph(rulesGraph);
    _ui->rulesWidget->update();

    _ui->tracesWidget->setGraph(traceGraph);
    _ui->tracesWidget->update();

    _ui->strategiesManagementWidget->setGraph(rulesGraph);
    //Restore strategies
    if(dataSet.exist("strategies")){
        DataSet strategiesDataSet;
        dataSet.get("strategies",strategiesDataSet);
        _ui->strategiesManagementWidget->setData(strategiesDataSet);
    }
    else {
        //no strategy. remove data in strategiesManagementWidget
        _ui->strategiesManagementWidget->setData(DataSet());
    }

    sinfo<<"Graph "<<graph->getName()<<" loaded";

    _ui->actionUndo->setEnabled(true);
    _ui->actionRedo->setEnabled(true);
    _ui->actionCopy->setEnabled(true);
    _ui->actionPaste->setEnabled(true);
    _ui->actionCut->setEnabled(true);
    _ui->actionSave->setEnabled(true);
    _ui->actionSave_as->setEnabled(true);
    _ui->actionSnapshot->setEnabled(true);
    _ui->menuEdit->setEnabled(true);
    _ui->menuAlgorithms->setEnabled(true);

    _ui->graphsWidget->setEnabled(true);
    _ui->rulesWidget->setEnabled(true);
    _ui->tracesWidget->setEnabled(true);
    _ui->strategiesManagementWidget->setEnabled(true);

    delete oldGraph;//Destruct old data.
    return true;
}


void PorgyPerspective::deleteTrace(Graph* trace) {
    QMessageBox::StandardButton reply;
    reply = QMessageBox::question(_mainWindow, tr("Delete a derivation tree"),
                                  QString("Do you really want to delete \"").append(tlpStringToQString(trace->getName())).append("\"?"),
                                  QMessageBox::Yes | QMessageBox::No);
    if(reply==QMessageBox::Yes) {
        if(PorgyTlpGraphStructure::getMainTrace(trace)==trace)
            QMessageBox::critical(_mainWindow,tr("Cannot delete derivation tree"),tr("cannot delete the main derivation tree"));
        else{
            Graph* traceRoot = PorgyTlpGraphStructure::getTraceRoot(trace);
            traceRoot->push();
            traceRoot->delSubGraph(trace);
        }
    }
}

void PorgyPerspective::deleteRule(Graph* rule) {
    QMessageBox::StandardButton reply;
    reply = QMessageBox::question(_mainWindow, tr("Delete a rule"),
                                  QString("Do you really want to delete \"").append(tlpStringToQString(rule->getName())).append("\"?"),
                                  QMessageBox::Yes | QMessageBox::No);
    if(reply==QMessageBox::Yes) {
#ifndef NDEBUG
        cerr << "deleting rule " << rule->getName() << endl;
#endif
        Graph *rules_root = PorgyTlpGraphStructure::getRulesRoot(rulesGraph);
        string rule_name = rule->getName();
        rules_root->push();
        Observable::holdObservers();
        Iterator<node> *it = rule->getNodes();
        rule->delNodes(it, true);
        delete it;
        rules_root->delAllSubGraphs(rule);
        Observable::unholdObservers();
        sinfo<<"Rules "<<rule_name<<" deleted";
    }
}

void PorgyPerspective::createNewRule() {
    bool ok;
    QString text = QInputDialog::getText(_mainWindow, tr("Create a new empty rule"),tr("Rule name:"), QLineEdit::Normal, "newrule", &ok);
    if (ok && !text.isEmpty()) {
        rulesGraph->push();
        Observable::holdObservers();
        Graph *new_rule =  rulesGraph->addSubGraph(QStringToTlpString(text));
        Bridge::createEmptyBridge(new_rule, QStringToTlpString(text));
        Observable::unholdObservers();
        sinfo<<"New rule "<<new_rule->getName()<<" created";
    }else{
        if(text.isEmpty()){
            serror<<"Cannot create a rule with an empty name";
        }
    }
}

void PorgyPerspective::copyRule(tlp::Graph *rule) {

    bool ok;
    QString text = QInputDialog::getText(_mainWindow, tr("Copy rule"),tr("Rule name:"), QLineEdit::Normal, tlp::tlpStringToQString(rule->getName())+"_2", &ok);
    if (ok && !text.isEmpty()) {
        rulesGraph->push();
        Observable::holdObservers();
        Graph *new_rule = rulesGraph->addSubGraph(QStringToTlpString(text));
        tlp::copyToGraph( new_rule, rule);
        Observable::unholdObservers();
        _ui->rulesWidget->update();
    }
}


void PorgyPerspective::showRule(Graph* graph){
    if(graph != NULL){
        tlp::View* view = _ui->viewManager->createView(PorgyConstants::RULE_VIEW_NAME,graph);
        if(view != NULL){
            connect(view,SIGNAL(copy_paste_other_side(PorgyConstants::RuleSide)),SLOT(send_other_side(PorgyConstants::RuleSide)));
            connect(view,SIGNAL(pastePortNodes(tlp::Coord &,tlp::Graph*,QList<PortNodeQt*>)),SLOT(pastePortNodes(tlp::Coord&,tlp::Graph*,QList<PortNodeQt*>)));
        }
    }
}
void PorgyPerspective::showModel(Graph* graph){
    if(graph != NULL){
        tlp::View* view = _ui->viewManager->createView(PorgyConstants::GRAPH_VIEW_NAME,graph);
        if(view != NULL){
            connect(view,SIGNAL(applyRuleOnModel(tlp::Graph*,tlp::Graph*)),SLOT(applyRuleOnModel(tlp::Graph*,tlp::Graph*)));
            connect(view,SIGNAL(applyStrategyOnModel(tlp::Graph*,QString)),SLOT(applyStrategyOnModel(tlp::Graph*,QString)));
            connect(view,SIGNAL(pastePortNodes(tlp::Coord &,tlp::Graph*,QList<PortNodeQt*>)),SLOT(pastePortNodes(tlp::Coord&,tlp::Graph*,QList<PortNodeQt*>)));
        }
    }
}
void PorgyPerspective::showTrace(Graph* graph,bool useDefaultView){
    if(graph != NULL){
        DataSet pluginParameters;
        string plugin;
        Trace trace(graph);
        //User want to show directly the graph with it's default view.
        if(useDefaultView){
            //This graph follows a portnode, shows it in histogram
            if(trace.type() == Trace::Histogram){
                plugin = PorgyConstants::SCATTER_PLOT_NAME;
            }else{
                if(graphIsCompatibleWithView(graph,PorgyConstants::SMALL_MULTIPLE_VIEW_NAME)){
                    plugin = PorgyConstants::SMALL_MULTIPLE_VIEW_NAME;
                }else{
                    plugin = PorgyConstants::TRACE_VIEW_NAME;
                }
            }
        }
        else{
            //User does not know the kind of view to use, ask for it.
            plugin = PorgyConstants::TRACE_VIEW_NAME;
            if(graphIsCompatibleWithView(graph,PorgyConstants::SMALL_MULTIPLE_VIEW_NAME)){
                list<string> plugins;
                plugins.push_back(PorgyConstants::TRACE_VIEW_NAME);
                plugins.push_back(PorgyConstants::HISTOGRAM_NAME);
                plugins.push_back(PorgyConstants::SCATTER_PLOT_NAME);
                plugins.push_back(PorgyConstants::SMALL_MULTIPLE_VIEW_NAME);

                PluginSelectionWizard wizard(_mainWindow);
                wizard.setPluginList(plugins);
                wizard.page(0)->setTitle(tr("Visualize ")+tlpStringToQString(graph->getName()));

                if(wizard.exec()!= QDialog::Accepted){
                    return;
                }
                plugin = wizard.selectedPlugin();
            }
        }
        //If a plugin is set
        if(!plugin.empty()){
            //Configure plugin parameters.
            //Scatter plot
            if(plugin == PorgyConstants::SCATTER_PLOT_NAME){
                pluginParameters.set("background color",Color(255,255,255));
                pluginParameters.set("min Size Mapping",20);
                pluginParameters.set("max Size Mapping",30);
                pluginParameters.set("display graph edges",true);
                if(trace.type() == Trace::Histogram){
                    DataSet selectedDataSetProperties;
                    selectedDataSetProperties.set<string>("0",trace.depthProperty()->getName());
                    vector<IntegerProperty*> portNodeProperties = trace.followedPortnodesProperties();
                    ostringstream oss;
                    for(unsigned int i= 0 ; i < portNodeProperties.size() ; ++i){
                        oss.str("");
                        oss<<i+1;
                        selectedDataSetProperties.set<string>(oss.str(),portNodeProperties[i]->getName());
                    }
                    pluginParameters.set("selected graph properties",selectedDataSetProperties);
                }
            }
            View* v=NULL;
            v = _ui->viewManager->createView(plugin,graph,pluginParameters);
            if(v->name()==PorgyConstants::TRACE_VIEW_NAME) {
                connect(v,SIGNAL(applyRuleOnModel(tlp::Graph*,tlp::Graph*)),SLOT(applyRuleOnModel(tlp::Graph*,tlp::Graph*)));
                connect(v,SIGNAL(applyStrategyOnModel(tlp::Graph*,QString)),SLOT(applyStrategyOnModel(tlp::Graph*,QString)));
                connect(v,SIGNAL(openModelView(tlp::Graph *)),this,SLOT(showModel(tlp::Graph *)));
            }
        }
    }
}

bool PorgyPerspective::graphIsCompatibleWithView(Graph* graph,const string& viewName ) const{
    if(PorgyTlpGraphStructure::isModelGraph(graph)){
        return viewName.compare(PorgyConstants::GRAPH_VIEW_NAME)==0;
    }else if(PorgyTlpGraphStructure::isRuleGraph(graph)){
        return viewName.compare(PorgyConstants::RULE_VIEW_NAME)==0;
    }else if(PorgyTlpGraphStructure::isTraceGraph(graph)){
        if(viewName.compare(PorgyConstants::TRACE_VIEW_NAME)==0){
            return true;
        }else if(viewName.compare(PorgyConstants::SMALL_MULTIPLE_VIEW_NAME)==0 ){
            bool containsErrorNode=false;
            node n;
            Trace trace(graph);
            forEach(n,graph->getNodes()){
                if(trace.getTypeProperty()->getNodeValue(n)!=Trace::Model_Node){
                    containsErrorNode = true;
                    break;
                }
            }
            return !containsErrorNode && (graph->numberOfNodes()>1) && PorgyTlpGraphStructure::graphIsASimplePath(graph);
        }
    }
    return false;
}

typedef std::vector<node> NodeA;
typedef std::vector<edge> EdgeA;

void GetSelection(NodeA & outNodeA, EdgeA & outEdgeA,
                  Graph *inG, BooleanProperty * inSel ) {
    assert( inSel );
    assert( inG );
    outNodeA.clear();
    outEdgeA.clear();
    // Get edges
    Iterator<edge> * edgeIt = inG->getEdges();

    while( edgeIt->hasNext() ) {
        edge e = edgeIt->next();

        if( inSel->getEdgeValue(e) )
            outEdgeA.push_back( e );
    }

    delete edgeIt;
    // Get nodes
    Iterator<node> * nodeIt = inG->getNodes();

    while( nodeIt->hasNext() ) {
        node n = nodeIt->next();

        if( inSel->getNodeValue(n) )
            outNodeA.push_back( n );
    }

    delete nodeIt;
}

void SetSelection(BooleanProperty * outSel, NodeA & inNodeA,
                  EdgeA & inEdgeA, Graph * inG) {
    assert( outSel );
    assert( inG );
    (void) inG; //to remove unused param warning
    outSel->setAllNodeValue( false );
    outSel->setAllEdgeValue( false );

    // Set edges
    for( unsigned int e = 0 ; e < inEdgeA.size() ; e++ )
        outSel->setEdgeValue( inEdgeA[e], true );

    // Set nodes
    for( unsigned int n = 0 ; n < inNodeA.size() ; n++ )
        outSel->setNodeValue( inNodeA[n], true );
}

//==============================================================
void PorgyPerspective::editCut() {
    //no copy action on the trace view
    assert(currentView()->name() != PorgyConstants::TRACE_VIEW_NAME);

    Graph *graph=getCurrentGraph();

    if(!graph)
        return;
    Observable::holdObservers();

    PortGraph *pg=NULL;
    if(PorgyTlpGraphStructure::isModelGraph(graph))
        pg = new PortGraphModel(graph);
    else
        pg = new PortGraphRule(graph);

    BooleanProperty *selP = pg->getSelectionProperty();

    if( selP ){
        edge e;
        forEach(e, selP->getEdgesEqualTo(true,graph)){
            const pair<PortNode*, PortNode*> ends = pg->ends(e);
            ends.first->select(true);
            ends.second->select(true);
        }
        delete pg;
        // Save selection
        NodeA nodeA;
        EdgeA edgeA;
        GetSelection( nodeA, edgeA, graph, selP );
        Graph* newGraph = tlp::newGraph();
        tlp::copyToGraph( newGraph, graph, selP );
        stringstream tmpss;
        DataSet dataSet;
        tlp::exportGraph(newGraph, tmpss, "TLP Export", dataSet, NULL);
        delete newGraph;
        QClipboard *clipboard = QApplication::clipboard();
        clipboard->setText(tmpss.str().c_str());
        graph->push();
        // Restore selection
        SetSelection( selP, nodeA, edgeA, graph );
        tlp::removeFromGraph( graph, selP );
        redrawPanels();
    }
    Observable::unholdObservers();
}
//==============================================================
void PorgyPerspective::editCopy() {

    //no copy action on the trace view
    assert(currentView()->name() != PorgyConstants::TRACE_VIEW_NAME);

    Graph *graph=getCurrentGraph();

    if(!graph)
        return;

    Observable::holdObservers();
    edge e;

    PortGraph *pg=NULL;
    if(PorgyTlpGraphStructure::isModelGraph(graph))
        pg = new PortGraphModel(graph);
    else
        pg = new PortGraphRule(graph);

    BooleanProperty *selP = pg->getSelectionProperty();

    if( !selP ) return;

    forEach(e, selP->getEdgesEqualTo(true,graph)){
        const pair<PortNode*, PortNode*> ends = pg->ends(e);
        ends.first->select(true);
        ends.second->select(true);
    }
    Graph* newGraph = tlp::newGraph();
    tlp::copyToGraph( newGraph, graph, selP );
    stringstream tmpss;
    DataSet dataSet;
    tlp::exportGraph(newGraph, tmpss, "TLP Export", dataSet, NULL);
    QClipboard *clipboard = QApplication::clipboard();
    clipboard->setText(QString::fromUtf8(tmpss.str().c_str()));
    delete newGraph;
    Observable::unholdObservers();
    delete pg;
}
//==============================================================
void PorgyPerspective::editPaste(PorgyConstants::RuleSide newside) {
    Graph *graph=getCurrentGraph();

    if(!graph)
        return;

    //Get data in the clipboard.
    QClipboard *clipboard = QApplication::clipboard();
    QString data = clipboard->text();

    //If there is data
    if(!data.isEmpty()){

        Graph *newGraph=tlp::newGraph();
        DataSet dataSet;
        dataSet.set<string>("file::data", data.toUtf8().data());
        tlp::importGraph("TLP Import", dataSet, NULL ,newGraph);
        //If there is nodes in the imported graph.
        if(newGraph->numberOfNodes() != 0){
            Observable::holdObservers();
            BooleanProperty * selP = graph->getProperty<BooleanProperty>("viewSelection");
            graph->push();
            IntegerProperty *side = PortGraphRule::getSideProperty(newGraph);
            //If the destination graph is a rule graph.
            if(PorgyTlpGraphStructure::isRuleGraph(graph)){
                //If at least one node as no side treat the whole graph as a model graph.
                bool sidePropertyExists = true;
                node n;
                forEach(n,newGraph->getNodes()){
                    if(side->getNodeValue(n)==PorgyConstants::NO_SIDE){
                        sidePropertyExists = false;
                        break;
                    }
                }
                //If the source graph is a model graph and if the side was not specified ask user.
                if(!sidePropertyExists && newside==PorgyConstants::NO_SIDE){
                    SpecialPasteDialog dialog(_mainWindow);
                    if(dialog.exec()) {
                        newside = dialog.getNewRuleSideName();
                    }
                    else {
                        delete newGraph;
                        Observable::unholdObservers();
                        return;
                    }
                }
                //If the side was defined overwrite existing or create new one
                if(newside!=PorgyConstants::NO_SIDE) {
                    side->setAllNodeValue(newside);
                    side->setAllEdgeValue(newside);
                }
            }else{
                //If the destination graph is model erase all side values.
                if(graph->existProperty(PorgyConstants::RULE_SIDE)) {
                    graph->delLocalProperty(PorgyConstants::RULE_SIDE);
                }
            }

            Observable::holdObservers();
            tlp::copyToGraph( graph, newGraph, 0, selP);
            Observable::unholdObservers();
            //Try to find the first empty space
            LayoutProperty* positions = graph->getProperty<LayoutProperty>("viewLayout");
            SizeProperty* sizes = graph->getProperty<SizeProperty>("viewSize");
            //Compute the size of the selection
            BoundingBox b = computeBoundingBox(graph,positions,sizes,graph->getProperty<DoubleProperty>("viewRotation"),selP);
            Coord shift;
            bool foundEmptySpace = false;
            int layer = 0;
            while(!foundEmptySpace){
                //Increase the layer
                ++layer;
                //0: north, 1: north-east, 2: east, 3:south-east, 4:south, 5:south-west, 6:west, 7:north-west
                for(int direction = 0 ; direction < 8 ; ++direction){
                    //Reset shift
                    shift=Coord();
                    //east
                    if(direction>0 && direction <4){
                        shift[0]=1;
                    }else if(direction >4){
                        //west
                        shift[0]=-1;
                    }
                    //North
                    if(direction ==0 || direction ==1 || direction == 7){
                        shift[1] = -1;
                    }else if(direction != 2 ){
                        //South
                        shift[1] = 1;
                    }
                    shift[0]*= b.width()*layer*1.1;
                    shift[1]*= b.height()*layer*1.1;
                    BoundingBox newBox(b);
                    newBox.translate(shift);

                    foundEmptySpace = true;
                    node n;
                    forEach(n,graph->getNodes()){
                        BoundingBox nodeBox;
                        nodeBox.expand(positions->getNodeValue(n) + sizes->getNodeValue(n)/2.f);
                        nodeBox.expand(positions->getNodeValue(n) - sizes->getNodeValue(n)/2.f);
                        if(newBox.intersect(nodeBox)){
                            foundEmptySpace = false;
                            break;
                        }
                    }
                    if(foundEmptySpace){
                        edge e;
                        forEach(e,graph->getEdges()){
                            const vector<Coord>& bends = positions->getEdgeValue(e);
                            for(vector<Coord>::const_iterator it = bends.begin() ; it != bends.end() ; ++it){
                                if(newBox.contains(*it)){
                                    foundEmptySpace = false;
                                    break;
                                }
                            }
                        }
                        if(foundEmptySpace){
                            break;
                        }
                    }
                }
            }
            //Shift nodes center
            node n;
            forEach(n,selP->getNodesEqualTo(true)){
                positions->setNodeValue(n,positions->getNodeValue(n)+shift);
            }
            //Shift edges bends
            edge e;
            forEach(e,selP->getEdgesEqualTo(true)){
                vector<Coord> bends = positions->getEdgeValue(e);
                for(vector<Coord>::iterator it = bends.begin() ; it != bends.end() ; ++it){
                    (*it) = (*it)+shift;
                }
                positions->setEdgeValue(e,bends);
            }


            Observable::unholdObservers();
            redrawPanels(false);
        }
        //Delete new graph
        delete newGraph;
    }
}


//==============================================================
void PorgyPerspective::editDelSelection() {
    Graph *graph= getCurrentGraph();
    if(!graph)
        return;

    //no delete action on the trace view
    assert(currentView()->name() != PorgyConstants::TRACE_VIEW_NAME);
    BooleanProperty *selection = NULL;

    PortGraph *pgm = NULL;
    if(PorgyTlpGraphStructure::isModelGraph(graph)) {
        pgm = new PortGraphModel(graph);
    }
    else if(PorgyTlpGraphStructure::isRuleGraph(graph)) {
        pgm = new PortGraphRule (graph);
    }
    else {
        Trace t(graph);
        selection = t.getSelectionProperty();
    }

    if(pgm) {
        selection = pgm->getSelectionProperty();
        delete pgm;
    }

    graph->push();
    Observable::holdObservers();
    node n;
    edge e;
    stableForEach(n, graph->getNodes()) {
        if (selection->getNodeValue(n)==true)
            graph->delNode(n);
    }
    stableForEach(e, graph->getEdges()) {
        if (selection->getEdgeValue(e)==true)
            graph->delEdge(e);
    }
    Observable::unholdObservers();
}
//==============================================================
void PorgyPerspective::editReverseSelection() {
    Graph *graph= getCurrentGraph();

    if(!graph)
        return;

    graph->push();
    PortGraph *pgm=NULL;
    BooleanProperty *selectionProperty=NULL;
    edge e;
    Observable::holdObservers();
    if(PorgyTlpGraphStructure::isModelGraph(graph)) {
        pgm = new PortGraphModel(graph);
    }
    else if(PorgyTlpGraphStructure::isRuleGraph(graph)) {
        pgm = new PortGraphRule(graph);
    }
    if(pgm!=NULL) {
        selectionProperty = pgm->getSelectionProperty();
        stableForEach(e, pgm->getEdges()) {
            selectionProperty->setEdgeValue(e,!selectionProperty->getEdgeValue(e));
        }
        PortNode *p;
        stableForEach(p, pgm->getPortNodes()) {
            p->select(!selectionProperty->getNodeValue(p->getCenter()));
        }
        delete pgm;
    }

    if (PorgyTlpGraphStructure::isTraceGraph(graph)) {
        Trace t(graph);
        selectionProperty = t.getSelectionProperty();
        stableForEach(e, graph->getEdges()) {
            selectionProperty->setEdgeValue(e,!selectionProperty->getEdgeValue(e));
        }
        node n;
        stableForEach(n, graph->getNodes()) {
            selectionProperty->setNodeValue(n,!selectionProperty->getNodeValue(n));
        }
    }
    Observable::unholdObservers();
}
//==============================================================

Graph* PorgyPerspective::getCurrentGraph(){
    View* currentView = _ui->viewManager->currentView();
    if(currentView != NULL){
        return currentView->graph();
    }else{
        return NULL;
    }
}

tlp::View* PorgyPerspective::currentView(){
    return _ui->viewManager->currentView();
}

//==============================================================
void PorgyPerspective::editSelectAll() {
    Graph *graph= getCurrentGraph();
    if(!graph)
        return;
    graph->push();
    PortGraph *pgm=NULL;
    edge e;

    BooleanProperty *selectionProperty = NULL;
    Observable::holdObservers();
    if(PorgyTlpGraphStructure::isModelGraph(graph)) {
        pgm = new PortGraphModel(graph);
    }
    else if(PorgyTlpGraphStructure::isRuleGraph(graph)) {
        pgm = new PortGraphRule(graph);
    }

    if(pgm) {
        selectionProperty =  pgm->getSelectionProperty();
        forEach(e, pgm->getEdges()) {
            selectionProperty->setEdgeValue(e, true);
        }

        PortNode *p=NULL;
        forEach(p, pgm->getPortNodes()) {
            p->select(true);
        }
        delete pgm;
    }

    if (PorgyTlpGraphStructure::isTraceGraph(graph)) {
        Trace t(graph);
        selectionProperty = t.getSelectionProperty();
        forEach(e, graph->getEdges()) {
            selectionProperty->setEdgeValue(e, true);
        }
        node n;
        forEach(n, graph->getNodes()) {
            selectionProperty->setNodeValue(n, true);
        }
    }

    Observable::unholdObservers();
}
//==============================================================
void PorgyPerspective::editDeselectAll() {
    Graph *graph= getCurrentGraph();
    if(!graph)
        return;
    PortGraph *pgm=NULL;
    BooleanProperty *selectionProperty=NULL;
    Observable::holdObservers();
    if(PorgyTlpGraphStructure::isModelGraph(graph)) {
        pgm = new PortGraphModel(graph);
    }
    else if(PorgyTlpGraphStructure::isRuleGraph(graph)) {
        pgm = new PortGraphRule(graph);
    }

    if(pgm) {
        selectionProperty = pgm->getSelectionProperty();
        PortNode *p;
        forEach(p, pgm->getPortNodes()) {
            p->select(false);
        }
        delete pgm;
    }
    if(PorgyTlpGraphStructure::isTraceGraph(graph)) {
        Trace t(graph);
        selectionProperty = t.getSelectionProperty();
        node n;
        forEach(n, graph->getNodes()) {
            selectionProperty->setNodeValue(n, false);
        }
    }

    edge e;
    forEach(e, graph->getEdges()) {
        selectionProperty->setEdgeValue(e, false);
    }

    Observable::unholdObservers();
}

void PorgyPerspective::viewActivated(View* view) {
    //disabled copy/cut/paste for the trace view
    if(view->name()==PorgyConstants::TRACE_VIEW_NAME) {
        _ui->actionCut->setEnabled(false);
        _ui->actionCopy->setEnabled(false);
        _ui->actionPaste->setEnabled(false);
        _ui->actionDel_selection->setEnabled(false);
    }
    else {
        _ui->actionCut->setEnabled(true);
        _ui->actionCopy->setEnabled(true);
        _ui->actionPaste->setEnabled(true);
        _ui->actionDel_selection->setEnabled(true);
    }
}



void PorgyPerspective::applyStrategy(QString strategy){
    QList<Graph*> selectedGraphs = _ui->graphsWidget->selectedModels();
    Graph* selectedGraph = selectedGraphs.empty()?NULL:selectedGraphs.front();
    if(selectedGraph == NULL){
        serror<<QStringToTlpString(tr("No model selected, cannot apply strategy."));
    }else{
        applyStrategyOnModel(selectedGraph,strategy);
    }
}
void PorgyPerspective::applyStrategyOnModel(Graph* model,QString strategy){
    assert(model != NULL);
    DataSet dataSet;
    dataSet.set<string>("Strategy",QStringToTlpString(strategy));
    Observable::holdObservers();
    GraphModficationListener listener;
    Graph* mainTrace = PorgyTlpGraphStructure::getMainTrace(rootGraph);
    listener.startRecording(mainTrace);

    std::string err;
    if(applyAlgorithm(PorgyConstants::APPLY_STRATEGY_ALGORITHM,model,dataSet,err)){
        sinfo<<QStringToTlpString(tr("Strategy applied."));
        listener.stopRecording(mainTrace);
        set<node> toVisualize = listener.getCreatedNodes();
        //Add parent node
        Trace trace(mainTrace);
        node modelNode = trace.nodeForModel(model);
        if(modelNode.isValid()){
            toVisualize.insert(modelNode);
        }
        zoomOnTraceElementAnimation(toVisualize,sender());
    }else{
        listener.stopRecording(mainTrace);
        serror<<"Error applying strategy "<<err;
    }
    Observable::unholdObservers();
}

void PorgyPerspective::zoomOnTraceElementAnimation(const std::set<tlp::node>&elements,const QObject* object){
    unsigned int animationTime =settings.traceViewZoomAnimationTime();
    Graph* mainTrace = PorgyTlpGraphStructure::getMainTrace(rootGraph);
    //If we created nodes
    if(!elements.empty()){
        BooleanProperty createdNodesProperty(mainTrace);
        createdNodesProperty.setAllNodeValue(false);
        createdNodesProperty.setAllEdgeValue(false);
        for(set<node>::const_iterator it = elements.begin(); it != elements.end();++it){
            createdNodesProperty.setNodeValue(*it,true);
        }
        QList<View*> views = _ui->viewManager->getViewsForGraph(PorgyTlpGraphStructure::getMainTrace(rootGraph));
        Graph* mainTrace = PorgyTlpGraphStructure::getMainTrace(rootGraph);
        if(!views.empty()){
            //Don't animate views if user don't want it.
            if(animationTime > 0){

                QParallelAnimationGroup *group = new QParallelAnimationGroup();
                foreach(View* view,views){
                    BoundingBox bbox;

                    if(view == object){
                        //If the current view is the view where the elements where applied zoom on modification.
                        bbox = computeBoundingBox(mainTrace,mainTrace->getProperty<LayoutProperty>("viewLayout"),mainTrace->getProperty<SizeProperty>("viewSize"),mainTrace->getProperty<DoubleProperty>("viewRotation"),&createdNodesProperty);
                    }else{
                        //Center the view.
                        bbox = computeBoundingBox(mainTrace,mainTrace->getProperty<LayoutProperty>("viewLayout"),mainTrace->getProperty<SizeProperty>("viewSize"),mainTrace->getProperty<DoubleProperty>("viewRotation"),NULL);
                    }

                    if(view->name().compare(PorgyConstants::TRACE_VIEW_NAME)==0){
                        GlMainView *mainView = dynamic_cast<GlMainView*>(view);
                        GlMainWidgetZoomAndPanAnimation* animation = new GlMainWidgetZoomAndPanAnimation(mainView->getGlMainWidget(),bbox,50,group);
                        animation->setDuration(animationTime);
                        group->addAnimation(animation);
                    }
                }
                QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
                group->start();
                while(group->state() == QAbstractAnimation::Running){
                    QApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
                }
                QApplication::restoreOverrideCursor();
                group->deleteLater();
            }else{
                foreach(View* view,views){
                    if(view->name().compare(PorgyConstants::TRACE_VIEW_NAME)==0){
                        view->centerView();
                    }
                }
            }
        }
    }
}


void PorgyPerspective::applyRuleOnModel(Graph* model,Graph* rule){
    assert(PorgyTlpGraphStructure::isModelGraph(model));
    assert(PorgyTlpGraphStructure::isRuleGraph(rule));
    QApplication::setOverrideCursor(QCursor(Qt::ArrowCursor));
    DataSet dataSet;
    ParameterDescriptionList params = PluginLister::instance()->getPluginParameters(PorgyConstants::SIMPLECHECKAPPLY);
    params.buildDefaultDataSet(dataSet, model );
    dataSet.set<std::string>(PorgyConstants::RuleName,rule->getName());
    QString title = QString("Apply rule ")+ QString::fromUtf8(rule->getName().c_str())+QString(" on model ")+QString::fromUtf8(model->getName().c_str());
    bool ok = true;
    dataSet = PluginParametersConfigurationDialog::getParameters(_mainWindow,title,params, ok, &dataSet,model);

    GraphModficationListener listener;
    Graph* mainTrace = PorgyTlpGraphStructure::getMainTrace(rootGraph);
    if (ok) {
        Observable::holdObservers();
        //Register modifications
        listener.startRecording(mainTrace);
        dataSet.set<std::string>(PorgyConstants::RuleName,rule->getName());//Set last parameter
        string error;
        //applyAlgorithm(PorgyConstants::SIMPLECHECKAPPLY,model,dataSet,error);
        model->push();
        PluginProgress* p = progress();
       bool ok = model->applyAlgorithm(PorgyConstants::SIMPLECHECKAPPLY, error, &dataSet,p);
        delete p;
        listener.stopRecording(mainTrace);
        Observable::unholdObservers();
        if(ok){
            sinfo<<"Rule "<<rule->getName()<<" correctly applied on model "<<model->getName();
            set<node> toVisualize = listener.getCreatedNodes();
            //Add parent node
            Trace trace(mainTrace);
            node modelNode = trace.nodeForModel(model);
            if(modelNode.isValid()){
                toVisualize.insert(modelNode);
            }
            zoomOnTraceElementAnimation(toVisualize,sender());
        }else{
            serror<<"Error when applying rule "<<rule->getName()<<" on model "<<model->getName()<<" : "<<error;
        }
    }

    QApplication::restoreOverrideCursor();
}

//**********************************************************************
bool PorgyPerspective::changeProperty(const QString& name, tlp::Graph* graph, std::string& err) {
    DataSet parameters;
    // plugin parameters dialog
    ParameterDescriptionList params = PluginLister::instance()->getPluginParameters(tlp::QStringToTlpString(name));
    DataSet dataSet;
    params.buildDefaultDataSet(dataSet, graph );
    QString title = QString("Tulip Parameter Editor: ") + name;
    bool   resultBool=true;
    parameters = PluginParametersConfigurationDialog::getParameters(_mainWindow,title,params,resultBool,&dataSet,graph);

    if(resultBool) {
        Observable::holdObservers();
        graph->push();//Record current state.
        PluginProgress* p = progress();
        resultBool = graph->applyAlgorithm(tlp::QStringToTlpString(name), err, &parameters, p);
        delete p;
    }
    return resultBool;
}

//**********************************************************************
void PorgyPerspective::changeMetric() {
    QAction *action=static_cast<QAction*>(sender());
    std::string err;
    if(!changeProperty(action->text(),getCurrentGraph(),err)){
        serror<<"Error when applying metric algorithm "<<action->text().toStdString();
    }
}

void PorgyPerspective::ImportGraph() {
    QAction *action=static_cast<QAction*>(sender());
    Graph *graph=getCurrentGraph();
    if(graph!=NULL) {
        QString algorithm=action->text();
        DataSet dataSet;
        ParameterDescriptionList params = PluginLister::instance()->getPluginParameters(tlp::QStringToTlpString(algorithm));
        if(!params.empty()) {
            params.buildDefaultDataSet(dataSet, graph );
            bool ok = true;
            dataSet = PluginParametersConfigurationDialog::getParameters(_mainWindow,algorithm+" plugin parameters",params,ok, &dataSet,graph);
        }
        tlp::importGraph(algorithm.toStdString(), dataSet, NULL, graph);
    }
}

//**********************************************************************
/// Apply a general algorithm
bool PorgyPerspective::applyAlgorithm() {
    QAction *action=static_cast<QAction*>(sender());
    Graph *graph=getCurrentGraph();
    if(!graph)
        return false;

    std::string errorMsg;
    QString algorithm=action->text();
    DataSet dataSet;
    ParameterDescriptionList params = PluginLister::instance()->getPluginParameters(tlp::QStringToTlpString(algorithm));
    if(!params.empty()) {
        params.buildDefaultDataSet(dataSet, graph );
        bool ok = true;
        dataSet = PluginParametersConfigurationDialog::getParameters(_mainWindow,algorithm+" plugin parameters",params,ok, &dataSet,graph);
    }
    return applyAlgorithm(tlp::QStringToTlpString(algorithm),graph,dataSet,errorMsg);
}

bool PorgyPerspective::applyAlgorithm(const std::string& algorithm,tlp::Graph* graph,tlp::DataSet& parameters,std::string& errorMsg) {
    Observable::holdObservers();
    graph->push();

    PluginProgress* p = progress();
    bool ok=true;
    if (!graph->applyAlgorithm(algorithm, errorMsg, &parameters,p)) {
        // no possible redo
        graph->pop(false);
        QMessageBox::warning(_mainWindow, "Cannot execute algorithm", "Cannot run the plugin: "+tlp::tlpStringToQString(errorMsg));
        ok=false;
    }
    delete p;
    Observable::unholdObservers();
    return ok;
}


//**********************************************************************
void PorgyPerspective::snapshot(){
    View* view = currentView();
    if(view != NULL) {
        QPixmap image=view->snapshot(QSize(16,16));
        if(image.isNull()){
            serror<<"Sorry but you cannot create a snapshot with this view";
        }else{
            SnapshotDialog snapshotDialog(view, _mainWindow);
            snapshotDialog.exec();
        }
    }
}

bool PorgyPerspective::eventFilter(QObject *obj, QEvent *event){
    if( event->type() == QEvent::ShortcutOverride){
        QKeyEvent *keyEvent = static_cast<QKeyEvent*>(event);
        if(keyEvent->key() == Qt::Key_F1 || keyEvent->key() == Qt::Key_Help || (keyEvent->modifiers().testFlag(Qt::ControlModifier) &&  keyEvent->modifiers().testFlag(Qt::ShiftModifier) && keyEvent->key()== Qt::Key_Question)){
            QWidget *widget =QApplication::widgetAt(QCursor::pos());
            QObject *obj = widget;
            if(qobject_cast<QMenu*>(widget)!= NULL){
                QMenu *menu = qobject_cast<QMenu*>(widget);
                obj = menu->activeAction();
            }
            if(obj){
                //Search for object ref in the help file
                QMap<QString, QUrl> links ;
                while(obj != NULL && links.empty()){
                    QString className =obj->metaObject()->className();
                    //Remove namespace as it's not supported by sphinx
                    if(className.contains("::")){
                        className = className.split("::").at(1);
                    }
                    QString id = QString(className+QString(".")+obj->objectName());
                    links = _ui->helpWidget->helpEngine()->linksForIdentifier(id);
                    obj = obj->parent();
                }
                if(!links.empty()){
                    _ui->helpWidget->showLink(links.begin().value(),HelpWidget::CurrentTab);
                }
                _ui->helpDockWidget->setVisible(true);
                _ui->helpDockWidget->raise();
            }
            event->accept();
            return true;
        }
        else if(keyEvent->modifiers()&Qt::ControlModifier &&  keyEvent->key() == Qt::Key_P ){
            showPythonView(rootGraph);
            event->accept();
            return true;
        }
        else if(keyEvent->modifiers()&Qt::ControlModifier &&  keyEvent->key() == Qt::Key_T ){
            showTableView(rootGraph);
            event->accept();
            return true;
        }
    }
    else if(event->type() == QEvent::Close) {
      if(savingObserver!=NULL&&savingObserver->needsSaving()) {
        QMessageBox::StandardButton answer = QMessageBox::question(_mainWindow, trUtf8("Save"), trUtf8("The project has been modified, do you want to save your changes ?"),QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel | QMessageBox::Escape);

        if(answer == QMessageBox::Yes)
            fileSave();
        else if(answer == QMessageBox::Cancel) {
          event->ignore();
          return true;
        }
      }
    }

    return Perspective::eventFilter(obj,event);
}

void PorgyPerspective::send_other_side(const PorgyConstants::RuleSide side) {
    View* view = currentView();
    if(view != NULL){
        assert(view->name()==PorgyConstants::RULE_VIEW_NAME);
        Observable::holdObservers();
        editCopy();
        editPaste(side);
        Observable::unholdObservers();
    }
}

void PorgyPerspective::pastePortNodes(Coord &center,Graph *graph, QList<PortNodeQt *> portNodes){
    graph->push();
    Observable::holdObservers();
    foreach(const PortNodeQt *pn,portNodes) {
        pn->addToTlpGraphUsingPosition(graph,center);
    }
    Observable::unholdObservers();
}



void PorgyPerspective::showDocumentation(){
    if(_ui->helpWidget->helpEngine() != NULL){
        _ui->helpDockWidget->setVisible(true);
        _ui->helpDockWidget->raise();
    }
}

void PorgyPerspective::toggleViewManagerImplementation(){

    ViewManagerImplementation currentImpl = static_cast<ViewManagerImplementation>(_ui->actionToggleViewManagerImplementation->data().toInt());
    if(currentImpl == MDI_AREA_IMPL){
        _ui->viewManager->setImplementation(new WorkspacePanelImplementation());
        _ui->actionToggleViewManagerImplementation->setData(WORKSPACE_IMPL);
    }else{
        _ui->viewManager->setImplementation(new MDIAreaImplementation());
        _ui->actionToggleViewManagerImplementation->setData(MDI_AREA_IMPL);
    }
}

void PorgyPerspective::addPanel(Graph* graph){
    if(graph != NULL){
        if(PorgyTlpGraphStructure::isModelGraph(graph)){
            showModel(graph);
        }else if(PorgyTlpGraphStructure::isRuleGraph(graph)){
            showRule(graph);
        }else if(PorgyTlpGraphStructure::isTraceGraph(graph)){
            showTrace(graph);
        }
    }else{

    }
}

tlp::View* PorgyPerspective::showPythonView(Graph *g) {
    return _ui->viewManager->createView("Python Script view",g);
}

tlp::View* PorgyPerspective::showTableView(Graph *g) {
    return _ui->viewManager->createView("Spreadsheet view",g);
}

void PorgyPerspective::undo() {
    if(rootGraph->canPop()){
        Observable::holdObservers();
        rootGraph->pop();
        Observable::unholdObservers();
        foreach(View* v, _ui->viewManager->views()) {
            v->undoCallback();
        }
    }
}

void PorgyPerspective::redo() {
    if(rootGraph->canUnpop()){
        Observable::holdObservers();
        rootGraph->unpop();
        Observable::unholdObservers();

        foreach(View* v, _ui->viewManager->views()) {
            v->undoCallback();
        }
    }
}

tlp::PluginProgress *PorgyPerspective::progress(ProgressOptions){
    StatusBarPluginProgress* PluginProgress = new StatusBarPluginProgress();
    _ui->statusbar->addPermanentWidget(PluginProgress);
    PluginProgress->setVisible(true);
    return PluginProgress;
}


PLUGIN(PorgyPerspective)







