/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "ruleswidget.h"
#include "ui_ruleswidget.h"

#include <QAbstractItemModel>
#include <QPushButton>
#include <QMenu>
#include <QHelpEvent>
#include <QItemSelectionModel>
#include <QAbstractItemView>

#include <tulip/Graph.h>
#include <tulip/TlpQtTools.h>

#include "qtimagetooltip.h"
#include "rulesgraphitemmodel.h"
#include "sortfilterproxymodel.h"

#include <portgraph/PorgyTlpGraphStructure.h>

using namespace tlp;
using namespace std;


RulesWidget::RulesWidget(QWidget *parent) :QWidget(parent),_ui(new Ui::RulesWidget),_graph(NULL), _model(NULL)
{
    _ui->setupUi(this);
    connect(_ui->ruleTreeView, SIGNAL(customContextMenuRequested(const QPoint&)),this, SLOT(ShowTreeContextMenu(const QPoint&)));
    connect(_ui->ruleTreeView,SIGNAL(doubleClicked(QModelIndex)),this,SLOT(showRule(QModelIndex)));
    connect(_ui->listView,SIGNAL(doubleClicked(QModelIndex)),this,SLOT(showRule(QModelIndex)));

    _ui->listView->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(_ui->listView, SIGNAL(customContextMenuRequested(const QPoint&)),this, SLOT(ShowTreeContextMenu(const QPoint&)));  

    connect(_ui->comboBox,SIGNAL(currentIndexChanged(int)),this,SLOT(updateIconSize()));



    _ui->listView->installEventFilter(this);
    _ui->ruleTreeView->installEventFilter(this);

    _proxyModel = new SortFilterProxyModel(this);
    _proxyModel->setDynamicSortFilter(true);
    _ui->listView->setModel(_proxyModel);
    _ui->ruleTreeView->setModel(_proxyModel);

    connect(_ui->lineEdit,SIGNAL(textEdited(QString)),SLOT(filterRules(QString)));


    _model = new RulesGraphItemModel(this);
    GlGraphRenderingParameters parameters;
    parameters.setEdgeColorInterpolate(false);
    parameters.setLabelsDensity(0);
    parameters.setLabelScaled(true);
    _model->snapshotManager().setRenderingParameters(parameters);
    _model->snapshotManager().setPreviewSize(QSize(32,32));
    _proxyModel->setSourceModel(_model);

    _selectionModel = new QItemSelectionModel(_proxyModel,this);
    _ui->listView->setSelectionModel(_selectionModel);
    _ui->ruleTreeView->setSelectionModel(_selectionModel);


#ifndef NDEBUG
    //If in debug mode the name column is set to 1
    _proxyModel->setFilterKeyColumn(1);
#endif

}

RulesWidget::~RulesWidget()
{
    delete _ui;
}

void RulesWidget::setGraph(Graph* graph){
    this->_graph = graph;
    _model->setRootGraph(graph);
    //Force previews to update to the right size to update
    updateIconSize();
}

void RulesWidget::ShowTreeContextMenu(const QPoint& point)
{
    QAbstractItemView *view = qobject_cast<QAbstractItemView*>(sender());
    assert(view);
    AbstractGraphHierachyItemModel *itemModel = qobject_cast<AbstractGraphHierachyItemModel*>(view->model());
    assert(_model);
    QMenu menu(this);
    Graph * selectedGraph = itemModel->indexGraph(_proxyModel->mapToSource(view->indexAt(point)));

    QAction *visualizeRule = menu.addAction(tr("&Show"), this,SLOT(showSelectedRule()));
    visualizeRule->setEnabled( selectedGraph != NULL);

    menu.addAction(tr("&Create new"), this,SIGNAL(addNewRule()));

    QAction *copyrule = menu.addAction(tr("&Clone"), this,SLOT(copy_one_rule()));
    copyrule->setEnabled(selectedGraph != NULL);

    QAction *delrule = menu.addAction(tr("&Delete"), this,SLOT(deleteSelectedRule()));

    bool canDeleteRule=selectedGraph != NULL;
    //check if rule is already used in a model
    if(canDeleteRule){
        Graph *models_root = PorgyTlpGraphStructure::getModelsRoot(selectedGraph);
        Iterator<Graph *> *itg = models_root->getSubGraphs();
        while(itg->hasNext()) {
            Graph *g=itg->next();
            Graph *inst;
            forEach(inst, g->getSubGraphs()) {
                QRegExp re("^"+QRegExp::escape("sg_"+tlp::tlpStringToQString(selectedGraph->getName())+"_")+".*$");
                if(re.exactMatch(tlp::tlpStringToQString(inst->getName()))) {
                    canDeleteRule=false;
                    break;
                }
            }
            if(!canDeleteRule)
                break;
        }
        delete itg;
    }
    delrule->setEnabled( canDeleteRule );
    menu.exec(_ui->ruleTreeView->viewport()->mapToGlobal(point));
}

void RulesWidget::copy_one_rule() {
    emit copyRule(selectedRule());
}

void RulesWidget::deleteSelectedRule(){
    emit deleteRule(selectedRule());
}

void RulesWidget::showRule(const QModelIndex& index){        
    if(index.isValid()){
        emit showRule(_model->indexGraph(_proxyModel->mapToSource(index)));
    }

}

void RulesWidget::showSelectedRule(){
    emit showRule(selectedRule());
}

Graph* RulesWidget::selectedRule()const{    
    QModelIndexList selectedIndexes = _selectionModel->selectedRows();
    if(selectedIndexes.empty()){
        return NULL;
    }else{
        return _model->indexGraph(_proxyModel->mapToSource(selectedIndexes.front()));
    }
}

void RulesWidget::updateIconSize(){
    //    int slideValue = _ui->horizontalSlider->value();
    int slideValue = _ui->comboBox->currentIndex();
    if(slideValue > 0){
        QSize iconSize;
        if(slideValue == 1 ){
            iconSize = QSize(32,32);
        }else if(slideValue == 2){
            iconSize = QSize(64,64);
        }else{
            iconSize = QSize(128,128);
        }
        _model->snapshotManager().setPreviewSize(iconSize);
        _ui->stackedWidget->setCurrentIndex(1);        
#ifndef NDEBUG
        //Only display Name column
        _model->setIconColumn(1);
        _ui->listView->setModelColumn(1);
#else
        _model->setIconColumn(0);
        _ui->listView->setModelColumn(0);
#endif
    }else{
        _ui->stackedWidget->setCurrentIndex(0);
        _model->setIconColumn(-1);
    }
}


bool RulesWidget::eventFilter(QObject * obj, QEvent * evt){

    switch(evt->type()){
    case QEvent::ToolTip:
    {
        QHelpEvent *he = static_cast<QHelpEvent *>(evt);
        Graph *graph = NULL;
        if(obj == _ui->ruleTreeView){
            //Need to use viewport <idget to compute correct coordinate tranformation.
            //Return wrong QModelIndex if we use he->Pos();
            graph = _model->indexGraph(_proxyModel->mapToSource(_ui->ruleTreeView->indexAt(_ui->ruleTreeView->viewport()->mapFromGlobal(he->globalPos()))));
        }else if(obj == _ui->listView){
            graph = _model->indexGraph(_proxyModel->mapToSource(_ui->listView->indexAt(he->pos())));
        }
        if(graph){
            GlGraphRenderingParameters parameters;
            parameters.setEdgeColorInterpolate(false);
            parameters.setLabelsDensity(0);
            parameters.setAntialiasing(true);
            parameters.setLabelScaled(true);
            QtImageToolTip::showGraph(he->globalPos(),graph,QSize(512,512),Color(255,255,255),parameters,NULL,this);
            return true;
        }
    }
        break;        
    default:
        break;
    }

    return QObject::eventFilter(obj,evt);
}

void RulesWidget::filterRules(const QString& pattern){
    _proxyModel->setFilterWildcard(pattern);
}
