#include "viewmanager.h"
#include "log.h"

#include "ui_viewmanager.h"
#include "ui_mdiareaworkspace.h"
#include "ui_workspacepanelimplementation.h"

#include <QMdiSubWindow>
#include <QLabel>
#include <QGraphicsView>

#include <tulip/View.h>
#include <tulip/GraphHierarchiesModel.h>
#include <tulip/TulipProject.h>
#include <tulip/Graph.h>

using namespace tlp;

ViewManager::ViewManager(QWidget* parent):QWidget(parent),_ui(new Ui::ViewManager()),_implementation(NULL),_parentWidget(NULL),_currentView(NULL)
{
    _ui->setupUi(this);
}

ViewManager::~ViewManager(){
    delete _ui;
}

void ViewManager::setImplementation(ViewManagerImplementation* implementation){
    ViewManagerImplementation* oldImpl = _implementation;
    QWidget* oldParentWidget = _parentWidget;
    if(_parentWidget != NULL){
        _parentWidget->setVisible(false);
        layout()->removeWidget(_parentWidget);
    }

    _implementation = implementation;
    _parentWidget = NULL;
    if(_implementation != NULL){
        _parentWidget = new QWidget(this);
        _ui->horizontalLayout->addWidget(_parentWidget);
        _implementation->setupUi(_parentWidget);
        _parentWidget->setVisible(true);
        connect(_implementation,SIGNAL(showGraphRequest(tlp::Graph*)),SIGNAL(showGraphRequest(tlp::Graph*)));
        connect(_implementation,SIGNAL(currentViewActivated(tlp::View*)),SLOT(currentViewChangedInternal(tlp::View*)));
    }

    delete oldImpl;
    delete oldParentWidget;
}

tlp::View* ViewManager::createView(const std::string& viewName, tlp::Graph* graph, const tlp::DataSet& parameters){

    tlp::View* view = tlp::PluginLister::instance()->getPluginObject<tlp::View>(viewName,NULL);
    if(view != NULL){
        view->setupUi();
        view->setGraph(graph);
        view->setState(parameters);
        _implementation->showView(view);
        //std::cerr << "Opening " << viewName << " on graph " << graph->getName() << std::endl;
    }else{
        swarning<<"Cannot create the view: \""<<viewName<<"\"";
    }
    return view;

}
tlp::View* ViewManager::currentView()const{
    return _currentView;
}

void ViewManager::currentViewChangedInternal(tlp::View* view){
    _currentView = view;
    emit currentViewChanged(_currentView);
}

void ViewManager::drawAllViews(){
    _implementation->drawAllViews();
}

QList<tlp::View*> ViewManager::getViewsForGraph(tlp::Graph* graph)const{
    QList<tlp::View*> views = _implementation->getAllViews();
    QList<tlp::View*> result;
    foreach(tlp::View* view,views){
        if(view->graph() ==  graph){
            result.push_back(view);
        }
    }
    return result;
}

QList<tlp::View*> ViewManager::views() const{
    return _implementation->getAllViews();
}

void ViewManager::saveToProject(tlp::TulipProject* project, QMap<tlp::Graph *, QString> rootIds, tlp::PluginProgress* progress){
    _implementation->saveToProject(project, rootIds, progress);
}

void ViewManager::restoreFromProject(tlp::TulipProject* project, QMap<QString, Graph *> rootIds, tlp::PluginProgress* progress){
    _implementation->restoreFromProject(project, rootIds, progress);
}

void ViewManager::closeView(tlp::View* view){
    _implementation->closeView(view);
}

void ViewManager::closeAllView(){
    _implementation->closeAll();
}

MDIAreaImplementation::MDIAreaImplementation():_ui(new Ui::MDIAreaWorkspace()){

}

MDIAreaImplementation::~MDIAreaImplementation(){
    delete _ui;
}

void MDIAreaImplementation::setupUi(QWidget* parent){
    _ui->setupUi(parent);
}

void MDIAreaImplementation::showView(tlp::View *view){
    QMdiSubWindow *subWindow1 = new QMdiSubWindow(_ui->mdiArea,Qt::WindowFlags(Qt::WA_NativeWindow));
    view->graphicsView()->setSizePolicy(QSizePolicy::Expanding,QSizePolicy::Expanding);
    view->graphicsView()->setCacheMode(QGraphicsView::CacheBackground);
    subWindow1->setWidget(view->graphicsView());
    subWindow1->setAttribute(Qt::WA_DeleteOnClose);
    subWindow1->setOption(QMdiSubWindow::RubberBandResize,true);
    subWindow1->setOption(QMdiSubWindow::RubberBandMove,true);
    _ui->mdiArea->addSubWindow(subWindow1);
    subWindow1->setVisible(true);
}

void MDIAreaImplementation::drawAllViews(){

}

WorkspacePanelImplementation::WorkspacePanelImplementation():_ui(new Ui::WorkspacePanelImplementation()),_model(new tlp::GraphHierarchiesModel(this)){

}
WorkspacePanelImplementation::~WorkspacePanelImplementation(){
    delete _ui;
    delete _model;
}

void WorkspacePanelImplementation::setupUi(QWidget* parent){
    _ui->setupUi(parent);
    _ui->widget->setModel(_model);
    _ui->widget->setAutoCenterPanelsOnDraw(false);
    connect(_ui->widget,SIGNAL(addPanelRequest(tlp::Graph*)),SIGNAL(showGraphRequest(tlp::Graph*)));
    connect(_ui->widget,SIGNAL(panelFocused(tlp::View*)),SIGNAL(currentViewActivated(tlp::View*)));
}
void WorkspacePanelImplementation::showView(tlp::View *view){
    if(!_model->indexOf(view->graph()).isValid())
        _model->addGraph(view->graph());
    _ui->widget->addPanel(view);
}


void WorkspacePanelImplementation::drawAllViews(){
    _ui->widget->redrawPanels();
}

QList<tlp::View*> WorkspacePanelImplementation::getAllViews()const{
    return _ui->widget->panels();
}

void WorkspacePanelImplementation::saveToProject(tlp::TulipProject* p, QMap<Graph *, QString> rootIds,tlp::PluginProgress*pp){
    _ui->widget->writeProject(p,rootIds,pp);
}

void WorkspacePanelImplementation::restoreFromProject(tlp::TulipProject* p,QMap<QString, Graph *> rootIds,tlp::PluginProgress*pp){
    _ui->widget->readProject(p,rootIds,pp);
}

void WorkspacePanelImplementation::closeView(tlp::View* view){
    _ui->widget->delView(view);
}

void WorkspacePanelImplementation::closeAll() {
    _ui->widget->closeAll();
}
