/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "porgyrulegraphview.h"
#include "RuleAttributesWidget.h"
#include "../../StandardInteractorPriority.h"

#include <tulip/GlMainWidget.h>
#include <tulip/Graph.h>
#include <tulip/Interactor.h>
#include <tulip/Algorithm.h>
#include <tulip/TlpQtTools.h>
#include <tulip/DrawingTools.h>
#include <tulip/GlGraphComposite.h>
#include <tulip/TulipMimes.h>
#include <tulip/Camera.h>

#include <portgraph/PorgyTlpGraphStructure.h>
#include <portgraph/PortGraphRule.h>
#include <portgraph/PortNode.h>
#include <portgraph/Bridge.h>
#include <portgraph/Port.h>

#include <porgy/porgymimedata.h>

#include <QMenu>
#include <QDragEnterEvent>
#include <QDropEvent>
#include <QToolTip>
#include <QMessageBox>
#include <QGraphicsView>
#include <QGraphicsSceneDragDropEvent>

using namespace tlp;
using namespace std;

PorgyRuleGraphView::PorgyRuleGraphView(const tlp::PluginContext*):AbstractPortGraphView(),_RuleAttributes(NULL) {
    addDependency(PorgyConstants::REDRAW_RULE, "1.0");
}

PorgyRuleGraphView::~PorgyRuleGraphView() {    
    delete _RuleAttributes;
}

QList<QWidget*> PorgyRuleGraphView::configurationWidgets()const{
    return AbstractPortGraphView::configurationWidgets()<<_RuleAttributes;
}

void PorgyRuleGraphView::graphChanged(tlp::Graph *g) {
    AbstractPortGraphView::graphChanged(g);
    _RuleAttributes->setRuleAttributes(g);
}

void PorgyRuleGraphView::setupWidget(){
    AbstractPortGraphView::setupWidget();
    if(_RuleAttributes==NULL)
        _RuleAttributes = new RuleAttributesWidget();
    graphicsView()->setAcceptDrops(true);
    getGlMainWidget()->setAcceptDrops(false);
    graphicsView()->scene()->installEventFilter(this);
    graphicsView()->installEventFilter(this);//Handle tooltip events
}

PortGraph* PorgyRuleGraphView::buildPortGraph(tlp::Graph* graph)const{
    return new PortGraphRule(graph);
}

void PorgyRuleGraphView::initGlParameters(tlp::GlGraphRenderingParameters& parameters)const{
    parameters.setViewArrow(false);
    parameters.setEdgeColorInterpolate(false);
    parameters.setEdgeSizeInterpolate(true);
    parameters.setLabelsDensity(0);
}

void PorgyRuleGraphView::fillContextMenu(QMenu *contextMenu, const QPointF & cursorPosition) {
    contextMenu->addAction(tr("&Center view"),this,SLOT(centerWithAnimation()));
    contextMenu->addAction(tr("&Redraw"),this,SLOT(redrawGraph()));
    contextMenu->addAction(trUtf8("&Take snapshot"),this,SLOT(openSnapshotDialog()));
    QMenu *menu = new QMenu(tr("Selection"), contextMenu);
    menu->addAction("Select the left hand side",this,SLOT(selectLHS()));
    menu->addAction("Select the right hand side",this,SLOT(selectRHS()));
    menu->addSeparator();
    menu->addAction("Copy/Paste selection to the left hand side",this,SLOT(pasteSelectionToLHS()));
    menu->addAction("Copy/Paste selection to the right hand side",this,SLOT(pasteSelectionToRHS()));
    contextMenu->addMenu(menu);

    bool result;
    // look if the mouse pointer is over a node or edge
    result = getGlMainWidget()->pickNodesEdges(cursorPosition.x(), cursorPosition.y(), selectedEntity);
    node n(selectedEntity.getNode());
    if (!result||((n.isValid())&&(Bridge::isBridge(n, graph()))))
        return;

    contextMenu->addSeparator();
    // Display a context menu
    QString str = n.isValid() ? "Node " : "Edge ";
    contextMenu->addAction(str)->setEnabled(false);
    contextMenu->addSeparator();
    contextMenu->addAction(tr("Add to/Remove from selection"),this,SLOT(invertSelection()));
}

void PorgyRuleGraphView::invertSelection(){
    AbstractPortGraphView::invertSelection(selectedEntity);
}

void PorgyRuleGraphView::pasteSelectionToLHS(){
    emit copy_paste_other_side(PorgyConstants::SIDE_LEFT);
}
void PorgyRuleGraphView::pasteSelectionToRHS(){
    emit copy_paste_other_side(PorgyConstants::SIDE_RIGHT);
}

void PorgyRuleGraphView::selectRHS(){
    selectSide(PorgyConstants::SIDE_RIGHT);
}
void PorgyRuleGraphView::selectLHS(){
    selectSide(PorgyConstants::SIDE_LEFT);
}

void PorgyRuleGraphView::selectSide(PorgyConstants::RuleSide side){
    PortGraphRule pg(graph());
    BooleanProperty *selP = pg.getSelectionProperty();
    PortNode *pn=NULL;
    forEach(pn, pg.getMember(side)) {
        pn->select(true);
    }
    edge e;
    forEach(e, pg.getMemberEdges(side)) {
        selP->setEdgeValue(e, true);
    }
}





bool PorgyRuleGraphView::eventFilter(QObject *object, QEvent *event){    

    if(event->type() == QEvent::GraphicsSceneDragEnter){
        return treatGraphicsDragMoveEvent(static_cast<QGraphicsSceneDragDropEvent*>(event));
    }else if(event->type() == QEvent::GraphicsSceneDragMove){
        return treatGraphicsDragMoveEvent(static_cast<QGraphicsSceneDragDropEvent*>(event));
    }else if(event->type() == QEvent::GraphicsSceneDrop){
        QGraphicsSceneDragDropEvent *dropEvent = static_cast<QGraphicsSceneDragDropEvent*>(event);
        const QMimeData* mimeData = dropEvent->mimeData();
        if(mimeData->hasFormat(GRAPH_MIME_TYPE)){
            const GraphMimeType* gMimeData = static_cast<const GraphMimeType*>(mimeData);
            Graph *dragedGraph = gMimeData->graph();
            if(dragedGraph != graph() && PorgyTlpGraphStructure::isRuleGraph(dragedGraph)){
                setGraph(dragedGraph);//Update the current view graph
            }
            event->accept();
            return true;
        }else if(mimeData->hasFormat(PorgyMimeData::portNodeQtPtrListMimeTypes)){
            const PorgyMimeData* pMimeData = static_cast<const PorgyMimeData*>(mimeData);
            Coord posCenter(static_cast<double>(getGlMainWidget()->width()) - static_cast<double>(dropEvent->pos().x()), static_cast<double>(dropEvent->pos().y()), 0);
            posCenter = getGlMainWidget()->getScene()->getGraphCamera().screenTo3DWorld(posCenter);
            emit pastePortNodes(posCenter,graph(),pMimeData->getPortNodeQtPtrList());
            dropEvent->accept();
            return true;
        }
    }
    else if (event->type() == QEvent::ToolTip) {
        SelectedEntity selectedEntity;
        QHelpEvent *he = static_cast<QHelpEvent *>(event);
        PortGraphRule pg(getGlMainWidget()->getScene()->getGlGraphComposite()->getInputData()->getGraph());
        if (getGlMainWidget()->pickNodesEdges(he->x(), he->y(), selectedEntity, NULL, true, false)) {
            std::string label;
            node n(selectedEntity.getNode());
            if(n.isValid()) {
                try
                {
                    if(pg.isCenter(n)) {
                        PortNode *p = pg.getPortNode(n);
                        label = "<b>Name: </b>"+p->getName();
                    }else if(pg.isPort(n)) {
                        Port::PortPtr port = pg.getPortContainer(n);
                        label = "<b>Name: </b>"+port->getName()+"<br><b>State: </b>"+port->getState();
                    }
                    QToolTip::showText(he->globalPos(), tlpStringToQString(label),getGlMainWidget());
                }
                catch (...)
                {
                    //An exception raised by boost ignore the event.
                }
            }

        }
    }
    return GlMainView::eventFilter(object,event);;
}

bool PorgyRuleGraphView::treatDragMoveEvent(QDragMoveEvent* evt){
    const QMimeData* mimeData = evt->mimeData();
    if(mimeData->hasFormat(GRAPH_MIME_TYPE)){
        const GraphMimeType* gMimeData = static_cast<const GraphMimeType*>(mimeData);
        Graph *dragedGraph = gMimeData->graph();
        if(dragedGraph != graph() && PorgyTlpGraphStructure::isRuleGraph(dragedGraph)){
            evt->setDropAction(Qt::MoveAction);
            evt->accept();
            return true;
        }else{
            //Block event
            evt->setDropAction(Qt::IgnoreAction);
            evt->accept();
            return true;
        }
    }else if(mimeData->hasFormat(PorgyMimeData::portNodeQtPtrListMimeTypes)){
        evt->setDropAction(Qt::MoveAction);
        evt->accept();
        return true;
    }
    return false;
}

bool PorgyRuleGraphView::treatGraphicsDragMoveEvent(QGraphicsSceneDragDropEvent* evt){
    const QMimeData* mimeData = evt->mimeData();
    if(mimeData->hasFormat(GRAPH_MIME_TYPE)){
        const GraphMimeType* gMimeData = static_cast<const GraphMimeType*>(mimeData);
        Graph *dragedGraph = gMimeData->graph();
        if(dragedGraph != graph() && PorgyTlpGraphStructure::isRuleGraph(dragedGraph)){
            evt->setDropAction(Qt::MoveAction);
            evt->accept();
            return true;
        }else{
            //Block event
            evt->setDropAction(Qt::IgnoreAction);
            evt->accept();
            return true;
        }
    }else if(mimeData->hasFormat(PorgyMimeData::portNodeQtPtrListMimeTypes)){
        evt->setDropAction(Qt::MoveAction);
        evt->accept();
        return true;
    }
    return false;
}


PLUGIN(PorgyRuleGraphView)

INTERACTORPLUGINVIEWEXTENSIONWITHPRIORITY(RuleViewNavigation,"RuleViewNavigation","InteractorNavigation",PorgyConstants::RULE_VIEW_NAME,"Tulip Team", "21/11/2008", "Navigate", "1.0",StandardInteractorPriority::Navigation)
INTERACTORPLUGINVIEWEXTENSIONWITHPRIORITY(RuleViewMagnifyingGlass,"RuleViewMagnifyingGlass","MouseMagnifyingGlassInteractor",PorgyConstants::RULE_VIEW_NAME,"Tulip Team", "21/11/2008", "Magnyfiying glass", "1.0",StandardInteractorPriority::MagnifyingGlass)
INTERACTORPLUGINVIEWEXTENSIONWITHPRIORITY(RuleViewFishEye,"RuleViewFishEye","FishEyeInteractor",PorgyConstants::RULE_VIEW_NAME,"Tulip Team", "21/11/2008", "Fisheye", "1.0",StandardInteractorPriority::FishEye)
INTERACTORPLUGINVIEWEXTENSIONWITHPRIORITY(RuleViewRectangleZoom,"RuleViewRectangleZoom","InteractorRectangleZoom",PorgyConstants::RULE_VIEW_NAME,"Tulip Team", "21/11/2008", "Rectangle zoom", "1.0",StandardInteractorPriority::ZoomOnRectangle)
