#include <porgy/PortNodeQtModel.h>
#include <porgy/porgymimedata.h>

#include <portgraph/PortGraphModel.h>
#include <portgraph/PortGraphRule.h>

PortNodeQtModel::PortNodeQtModel(QObject *parent) : QAbstractListModel(parent),_rules_root(NULL),_models_root(NULL),_needToReloadData(false)
{
}

PortNodeQtModel::~PortNodeQtModel(){
    clear();
}

void PortNodeQtModel::clear(){
    qDeleteAll(_portNodesList);
    _portNodesList.clear();
}

void PortNodeQtModel::initPortNodesList(tlp::Graph* graph){
    //Stop listening old graph
    if(_rules_root != NULL){
        _rules_root->removeListener(this);
        _rules_root->removeObserver(this);
    }
    if(_models_root != NULL){
        _models_root->removeListener(this);
        _models_root->removeObserver(this);
    }
    //Clear data
    beginResetModel();
    clear();
    endResetModel();

    _rules_root=PorgyTlpGraphStructure::getRulesRoot(graph);
    _models_root=PorgyTlpGraphStructure::getModelsRoot(graph);
    if(_rules_root != NULL && _models_root != NULL){
        updateData();
        _rules_root->addListener(this);
        _rules_root->addObserver(this);

        _models_root->addListener(this);
        _models_root->addObserver(this);
    }


}

void PortNodeQtModel::addPortNode(const PortNodeQt& prtNode){
    //Check if the port node already exist in the list.
    PortNodeQt* port;
    bool add = true;
    foreach(port, _portNodesList){
        //Check if the port node is not already in the list.
        if(port->isSimilarTo(prtNode)){
            //The port was found no need to add it again
            add = false;
            //Treat the next port
            break;
        }
    }
    if(add){
        //The port was not found add it in the list
        addPortNode(new PortNodeQt(prtNode));
    }
}

void PortNodeQtModel::addPortNode(PortNodeQt *portNode){
    beginInsertRows(QModelIndex(),_portNodesList.size(),_portNodesList.size());
    _portNodesList.push_back(portNode);
    endInsertRows();
}

int PortNodeQtModel::rowCount(const QModelIndex &) const{
    return _portNodesList.size();
}
QVariant PortNodeQtModel::data(const QModelIndex &index, int role) const{
    if(hasIndex(index.row(),index.column())){
        switch(role){
        case Qt::DisplayRole:
            return QVariant(_portNodesList[index.row()]->getName());
            break;
        case Qt::BackgroundRole:
            return QVariant(_portNodesList[index.row()]->getColor());
            break;
        }
    }
    return QVariant();
}
Qt::ItemFlags PortNodeQtModel::flags(const QModelIndex &) const{
    return Qt::ItemFlags(Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsDragEnabled);
}

PortNodeQt* PortNodeQtModel::portNodeFromIndex(const QModelIndex &index)const{
    if(!hasIndex(index.row(),index.column())){
        return NULL;
    }
    return _portNodesList[index.row()];
}

QStringList PortNodeQtModel::mimeTypes(){
    QStringList list;
    list << PorgyMimeData::portNodeQtPtrListMimeTypes;
    return list;
}

QMimeData* PortNodeQtModel::mimeData(const QModelIndexList &indexes) const{
    PorgyMimeData *data = NULL;
    if(!indexes.empty()){
        data = new PorgyMimeData();
        data->setPortNodeQtPtrList(portNodesFromIndexes(indexes));
    }
    return data;
}

QList<PortNodeQt *> PortNodeQtModel::portNodesFromIndexes(const QModelIndexList& indexes)const{
    QList<PortNodeQt *> portNodes;
    foreach(const QModelIndex p, indexes) {
        portNodes << portNodeFromIndex(p);
    }
    return portNodes;
}

void PortNodeQtModel::treatEvent(const tlp::Event & event){

    if(event.type() == tlp::Event::TLP_DELETE){
        //One of a graph is deleted avoid segfault
        if(_rules_root != NULL){
            _rules_root->removeListener(this);
            _rules_root->removeObserver(this);
        }
        _rules_root= NULL;
        if(_models_root != NULL){
            _models_root->removeListener(this);
            _models_root->removeObserver(this);
        }
        _models_root = NULL;

        //Clear data
        beginResetModel();
        clear();
        endResetModel();

    }else if(event.type() == tlp::Event::TLP_MODIFICATION){
        const tlp::GraphEvent* gEvt = dynamic_cast<const tlp::GraphEvent*>(&event);
        if(gEvt != NULL){
            switch(gEvt->getType()){
            case tlp::GraphEvent::TLP_ADD_NODE:
            case tlp::GraphEvent::TLP_ADD_NODES:
                _needToReloadData = true;
                break;

            default:
                break;
            }
        }
    }
}

void PortNodeQtModel::treatEvents(const std::vector<tlp::Event> &){
    if(_needToReloadData){
        updateData();
        _needToReloadData = false;
    }
}

void PortNodeQtModel::updateData(){    

    //Search for new elements
    tlp::Graph *sub;
    forEach (sub, _models_root->getSubGraphs()) {
        loadOneGraph<PortGraphModel>(PortGraphModel(sub));
    }
    forEach (sub, _rules_root->getSubGraphs()){
        loadOneGraph<PortGraphRule>(PortGraphRule(sub));
    }
}
