/**
 *
 * This file is part of Tulip (www.tulip-software.org)
 *
 * Authors: David Auber and the Tulip development Team
 * from LaBRI, University of Bordeaux 1 and Inria Bordeaux - Sud Ouest
 *
 * Tulip is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Tulip is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 */
#ifndef GRAPHVISUALPROPERTIES_H
#define GRAPHVISUALPROPERTIES_H

#include <QMap>

namespace tlp{
    class Graph;
    class PropertyInterface;
    class GlGraphInputData;
    class ColorProperty;
    class LayoutProperty;
    class SizeProperty;
    class DoubleProperty;
}

enum VisualProperties{NoVisualProperties=0,ElementColor=0x1,ElementBorderColor=0x2,ElementLabelColor=0x4,AllColorProperties=0x7,ElementSize=0x8,ElementLayout=0x10,ElementBorderWidth=0x20,AllVisualProperties=0xffffffff};

Q_DECLARE_FLAGS ( VisualPropertiesFlags, VisualProperties)

class GraphVisualProperties{
public:
    /**
      * @brief Build an object with no visual properties.
      **/
    GraphVisualProperties();
    /**
      * @brief Delete all the initialized properties.
      **/
    void deleteProperties();
    /**
      * @brief Copy contructor.
      **/
    GraphVisualProperties(const GraphVisualProperties& toCopy);
    /**
      * @brief Build an object from a GlGraphInputData visual properties.
      **/
    GraphVisualProperties(tlp::GlGraphInputData* inputData,VisualPropertiesFlags propertiesToGet);
    /**
      * @brief Build an object from a graph.
      **/
    GraphVisualProperties(tlp::Graph* graph,VisualPropertiesFlags propertiesToGet);
    virtual ~GraphVisualProperties();

    void initFromInputData(tlp::GlGraphInputData* inputData,VisualPropertiesFlags propertiesToGet);
    void initFromGraph(tlp::Graph* graph,VisualPropertiesFlags propertiesToGet);


    /**
     * @brief property return the property corresponding to the given property type.
     * @param property
     * @return
     */
    tlp::PropertyInterface* property(VisualProperties property)const{
        return _properties.contains(property)?_properties[property]:NULL;
    }

    void setProperty(VisualProperties type, tlp::PropertyInterface* property){
        _properties[type]=property;
    }


    /**
      * @brief Overload visual properties stored in a GlGraphInputData by visual properties in this object.
      **/
    void copyToInputData(tlp::GlGraphInputData* inputData)const;

    //Color properties
    tlp::ColorProperty* getElementColor()const;
    void setElementColor(tlp::ColorProperty* property);

    /**
     * @brief getElementBorderColor
     * @return
     */
    tlp::ColorProperty* getElementBorderColor()const;

    /**
     * @brief setElementBorderColor
     * @param property
     */
    void setElementBorderColor(tlp::ColorProperty* property);
    tlp::ColorProperty* getElementLabelColor()const;
    void setElementLabelColor(tlp::ColorProperty* property);

    //Layout prop
    tlp::LayoutProperty* getElementLayout()const;

    void setElementLayout(tlp::LayoutProperty* property);

    //Size property
    tlp::SizeProperty* getElementSize()const;
    void setElementSize(tlp::SizeProperty* property);

    tlp::DoubleProperty* getElementBorderWidth() const;
    void setElementBorderWidth(tlp::DoubleProperty* property);

private:

    QMap<VisualProperties, tlp::PropertyInterface*> _properties;

};

#endif // GRAPHVISUALPROPERTIES_H
