/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <portgraph/PorgyTlpGraphStructure.h>
#include <portgraph/Trace.h>
#include <portgraph/PortGraphRule.h>
#include <portgraph/PortGraphModel.h>

#include <tulip/Graph.h>
#include <tulip/ForEach.h>
#include <tulip/StringProperty.h>
#include <tulip/IntegerProperty.h>
#include <tulip/ColorProperty.h>
#include <tulip/LayoutProperty.h>
#include <tulip/SizeProperty.h>
#include <tulip/GraphMeasure.h>
#include <tulip/BooleanProperty.h>

#include <boost/lexical_cast.hpp>

using namespace std;
using namespace tlp;
using namespace boost;

tlp::Graph* PorgyTlpGraphStructure::newPorgyStructure(){

    tlp::Graph* root = tlp::newGraph();

    root->setName("root");
    root->getProperty<IntegerProperty>("viewTgtAnchorShape")->setAllEdgeValue(EdgeExtremityShape::None);
    root->getProperty<IntegerProperty>("viewSrcAnchorShape")->setAllEdgeValue(EdgeExtremityShape::None);
    root->getProperty<DoubleProperty>("viewBorderWidth")->setAllNodeValue(1.0);
    root->getProperty<StringProperty>(PorgyConstants::STATE);
    PortNodeBase::getTypeProperty(root);

    //graphs
    Graph *models_root = root->addSubGraph(PorgyConstants::GRAPHS_GRAPH_NAME);
    Graph *G0 = models_root->addSubGraph(PorgyConstants::INITIAL_GRAPH);
    models_root->getLocalProperty<BooleanProperty>("viewSelection");
    models_root->getProperty<IntegerProperty>("viewShape")->setAllNodeValue(PorgyConstants::CENTER_SHAPE);
    models_root->getLocalProperty<LayoutProperty>("viewLayout");    //must be local to each graph

    //rules
    Graph *rules_root = root->addSubGraph(PorgyConstants::RULES_GRAPH_NAME);
    rules_root->getLocalProperty<BooleanProperty>("viewSelection");
    rules_root->getProperty<IntegerProperty>("viewShape")->setAllNodeValue(PorgyConstants::CENTER_SHAPE);
    PortGraphRule::getSideProperty(rules_root);

    //derivation trees
    Graph *traceroot = root->addSubGraph(PorgyConstants::TRACE_GRAPH_NAME);
    traceroot->getLocalProperty<BooleanProperty>("viewSelection");
    Graph *tracemain = traceroot->addSubGraph(PorgyConstants::TRACEMAIN);
    root->setAttribute(PorgyConstants::TRACEMAIN, tracemain->getId());
    Trace t(tracemain);
    t.addMetaNode(G0);
    return root;
}

void PorgyTlpGraphStructure::Graph2PortGraph(Graph *graph, PluginProgress *pluginProgress) {
    //add ports
    node n;
    IntegerProperty *type = graph->getProperty<IntegerProperty>(PorgyConstants::PORTNODE_TYPE);
    forEach(n, graph->getNodes()) {
        type->setNodeValue(n, PorgyConstants::CENTER);
    }

    //find the node with the higher degree => max number of ports
    DoubleProperty deg_prop(graph);
    string err;
    graph->applyPropertyAlgorithm("Degree", &deg_prop, err, pluginProgress);
    int nb_ports=0;
    forEach(n, graph->getNodes()) {
        int tmp = deg_prop.getNodeValue(n);
        if(tmp>nb_ports)
            nb_ports=tmp;
    }

    //add ports for existing edges
    edge e;
    stableForEach(e, graph->getEdges()) {
        const pair<node, node> &ends = graph->ends(e);
        node p = graph->addNode();
        type->setNodeValue(p, PorgyConstants::PORT);
        graph->addEdge(ends.first, p);
        graph->setSource(e, p);
        p = graph->addNode();
        type->setNodeValue(p, PorgyConstants::PORT);
        graph->addEdge(ends.second, p);
        graph->setTarget(e, p);
    }

    //complete with empty ports
    PortGraphModel pg(graph);
    PortNode *pn;
    forEach(pn, pg.getPortNodes()) {
        for(int i=pn->nbPorts();i<nb_ports;i++) {
            node p = graph->addNode();
            type->setNodeValue(p, PorgyConstants::PORT);
            graph->addEdge(pn->getCenter(), p);
        }
    }
}

Graph *PorgyTlpGraphStructure::getRootGraph(const Graph* graph) {
    Graph *root=NULL;
    if(PorgyTlpGraphStructure::isTraceGraph(graph))
        root = getTraceRoot(graph);
    else if(PorgyTlpGraphStructure::isModelGraph(graph))
        root = getModelsRoot(graph);
    else if(PorgyTlpGraphStructure::isRuleGraph(graph))
        root = getRulesRoot(graph);
    assert(root!=NULL);
    return root;
}

string PorgyTlpGraphStructure::getState(Graph *g, const node &n) {
        StringProperty *s = g->getProperty<StringProperty>(PorgyConstants::STATE);
        return s->getNodeValue(n);
}

bool PorgyTlpGraphStructure::checkEdgeState(const Graph *g, const edge &rule_e, const edge &graph_e) {
    StringProperty *state = g->getRoot()->getProperty<StringProperty>(PorgyConstants::STATE);
    return state->getEdgeValue(rule_e)==state->getEdgeValue(graph_e);
}

//Remarque : prend en compte les multi-aretes
bool PorgyTlpGraphStructure::checkEdgeState(const Graph *g, const vector<edge> &rule_e, const vector<edge> &modele_e, bool exact) {
    StringProperty *state = g->getRoot()->getProperty<StringProperty>(PorgyConstants::STATE);
    bool b = true;
    set<edge> edge_set;
    if((exact && rule_e.size() != modele_e.size()) || rule_e.size() > modele_e.size() )
        return false;

    for(vector<edge>::const_iterator ite_rule = rule_e.begin(); ite_rule != rule_e.end() ; ++ite_rule)
    {
        edge e_rule = *ite_rule;
        bool edge_correct = false;
        for(vector<edge>::const_iterator ite_modele = modele_e.begin(); ite_modele != modele_e.end(); ++ite_modele)
        {
            edge e_modele = *ite_modele;
            if(state->getEdgeValue(e_rule) == state->getEdgeValue(e_modele) && edge_set.find(e_modele) == edge_set.end())
            {
                edge_set.insert(e_modele);
                edge_correct = true;
                break;
            }
        }
        if(!edge_correct)
        {
            b = false;
            break;
        }
    }

    return b;
}


void PorgyTlpGraphStructure::setPPropertyNameAttribute(Graph *g, const std::string &name) {
    g->setAttribute(PorgyConstants::P_NAME, name);
}

string PorgyTlpGraphStructure::getPPropertyNameAttribute(Graph *g) {
    string s;
    g->getAttribute(PorgyConstants::P_NAME, s);
    return s;
}

void PorgyTlpGraphStructure::setBanPropertyNameAttribute(Graph *g, const std::string &name) {
    g->setAttribute(PorgyConstants::BAN_NAME, name);
}

string PorgyTlpGraphStructure::getBanPropertyNameAttribute(Graph *g) {
    string s;
    g->getAttribute(PorgyConstants::BAN_NAME, s);
    return s;
}


//rulename set and get
void PorgyTlpGraphStructure::setRuleNameForSubGraph(tlp::Graph *g, const std::string &rulename) {
    g->setAttribute(PorgyConstants::RULENAME, rulename);
}

std::string PorgyTlpGraphStructure::getRuleNameForSubGraph(tlp::Graph *g) {
    string s;
    g->getAttribute(PorgyConstants::RULENAME, s);
    return s;
}


bool PorgyTlpGraphStructure::isPorgyGraph(const Graph* graph, string &errMsg){
    //Check if the top-level graph structure is valid.
    Graph *rules_root=getRulesRoot(graph), *models_root=getModelsRoot(graph), *traces_root=getTraceRoot(graph), *traceMain=getMainTrace(graph);
    if(rules_root == NULL) {
        errMsg = string("A required graph was not found (")+PorgyConstants::RULES_GRAPH_NAME+string(")");
        return false;
    }
    else if(models_root == NULL) {
        errMsg = string("A required graph was not found (")+PorgyConstants::GRAPHS_GRAPH_NAME+string(")");
        return false;
    }
    else if(traces_root == NULL) {
        errMsg = string("A required graph was not found (")+PorgyConstants::TRACE_GRAPH_NAME+string(")");
        return false;
    }
    else if(traceMain == NULL) {
        errMsg = string("A required graph was not found (")+PorgyConstants::TRACEMAIN+string("). Is the attribute "+PorgyConstants::TRACEMAIN+" defined for the root graph?");
        return false;
    }

    if(!graph->getRoot()->existLocalProperty(PorgyConstants::STATE)) {
        errMsg = "The property " + PorgyConstants::STATE + " must be local to the root graph";
        return false;
    }

    //node types
    IntegerProperty *type = PortNodeBase::getTypeProperty(graph->getRoot());
    if(type==NULL) {
        errMsg = "Cannot find the " + PorgyConstants::PORTNODE_TYPE + " Property";
        return false;
    }

    //verif que les propriétés sont correctement définies pour les sommets
    node n;
    forEach(n, rules_root->getNodes()) {
        if((type->getNodeValue(n)!=PorgyConstants::CENTER)&&(type->getNodeValue(n)!=PorgyConstants::PORT)) {
            errMsg = "Property "+PorgyConstants::PORTNODE_TYPE+" not defined for every nodes (at least for node " + lexical_cast<string>(n.id) + ") in " + rules_root->getName();
#ifndef NDEBUG
            errMsg += " (value: " + lexical_cast<string>(type->getNodeValue(n))+")";
#endif
            return false;
        }
    }

    IntegerProperty *side = PortGraphRule::getSideProperty(rules_root);
    if(side==NULL) {
        errMsg = "Cannot find the " + PorgyConstants::RULE_SIDE + " Property";
        return false;
    }

    edge e;
    forEach(e, rules_root->getEdges()) {
        if(side->getEdgeValue(e)==PorgyConstants::NO_SIDE) {
            errMsg = "Property "+PorgyConstants::RULE_SIDE+" not defined for every edges (at least for edge " + lexical_cast<string>(e.id) + ") in " + rules_root->getName();
            return false;
        }
    }
    forEach(n, rules_root->getNodes()) {
        if(side->getNodeValue(n)==PorgyConstants::NO_SIDE) {
            errMsg = "Property "+PorgyConstants::RULE_SIDE+" not defined for every nodes (at least for node " + lexical_cast<string>(n.id) + ") in " + rules_root->getName();
            return false;
        }
    }

    return true;
}

bool PorgyTlpGraphStructure::isTraceGraph(const Graph* g){
    assert(g!=NULL);
    Graph *traceRoot = getTraceRoot(g);
    return traceRoot == NULL ? false : traceRoot->isDescendantGraph(g);
}

Graph *PorgyTlpGraphStructure::getMainTrace(const Graph* graph){
    Graph *traceRoot = getTraceRoot(graph);
    assert(traceRoot!=NULL);
    unsigned index=0;
    graph->getRoot()->getAttribute<unsigned>(PorgyConstants::TRACEMAIN, index);
    assert(index!=0);
    return traceRoot->getDescendantGraph(index);
}

//clone du graphe courant vers le nouveau graphe
Graph *PorgyTlpGraphStructure::clone(Graph *old_graph, DataSet *ds) {
    //calcul du nom du nouveau modèle
    int num=0;
    Graph *root_models=PorgyTlpGraphStructure::getModelsRoot(old_graph);
    Graph *sub;
    //recherche numéro du dernier modèle créé
    forEach(sub, root_models->getSubGraphs()) {
        int temp = lexical_cast<int>(sub->getName().substr(1));
        if(temp>num)
            num = temp;
    }
    ++num;

    Graph *new_graph = root_models->addSubGraph(PorgyConstants::GRAPH_PREFIX+lexical_cast<string>(num));

    node n;
    forEach(n, old_graph->getNodes()) {
        new_graph->addNode(n);
    }

    edge e;
    forEach(e, old_graph->getEdges()) {
        new_graph->addEdge(e);
    }

    //recopie de viewLayout et viewSize
    LayoutProperty *layout_graph = old_graph->getProperty<LayoutProperty>("viewLayout");
    SizeProperty *size_graph = old_graph->getProperty<SizeProperty>("viewSize");
    LayoutProperty *layout_new_graph = new_graph->getLocalProperty<LayoutProperty>("viewLayout");
    SizeProperty *size_new_graph = new_graph->getLocalProperty<SizeProperty>("viewSize");
    layout_new_graph->copy(layout_graph);
    size_new_graph->copy(size_graph);

    BooleanProperty *new_element = new_graph->getLocalProperty<BooleanProperty>(PorgyConstants::NEW);
    new_element->setAllNodeValue(false);
    new_element->setAllEdgeValue(false);

    //je recopie la propriété P courante vers le nouveau modèle si les positions sont utilisées
    BooleanProperty *oldP=NULL, *oldBan=NULL;
    ds->get<BooleanProperty *>(PorgyConstants::POSITION, oldP);
    ds->get<BooleanProperty *>(PorgyConstants::BAN, oldBan);

    if(oldP!=NULL) {
        BooleanProperty *newP = new_graph->getLocalProperty<BooleanProperty>(oldP->getName());
        newP->copy(oldP);
        ds->set<BooleanProperty *>(PorgyConstants::POSITION, newP);
    }
    if(oldBan!=NULL) {
        BooleanProperty *newBan = new_graph->getLocalProperty<BooleanProperty>(oldBan->getName());
        newBan->copy(oldBan);
        ds->set<BooleanProperty *>(PorgyConstants::BAN, newBan);
    }

    return new_graph;
}

bool PorgyTlpGraphStructure::graphIsASimplePath(Graph* graph){
        map<unsigned int,node> depthMap;
        MutableContainer<unsigned int> level;
        //Compute the level for each node.
        dagLevel(graph,level);
        node n;
        forEach(n,graph->getNodes()){
                unsigned int currentDepth = level.get(n);
                map<unsigned int,node>::iterator it = depthMap.find(currentDepth);
                //An edge with the same level already exits
                if(it != depthMap.end()){
                    return false;
                }
                depthMap[currentDepth]=n;
        }
        if(depthMap.empty()){
            return false;
        }
        //Check if there is no gap between two nodes.
        for(map<unsigned int ,node >::iterator it = depthMap.begin(); it != depthMap.end();){
            node from = it->second;
            node to = (++it)->second;

            if(it != depthMap.end() ){

                //If there is no edge between two nodes a gap exists so the selection is invalid.
                if(!graph->hasEdge(from,to,true)){
                    return false;
                }
                Iterator<edge>* edges = graph->getOutEdges(from);
                unsigned int count = 0;
                bool moreThanOneEdge = false;
                //Check if there is more than one edge between the source and the destination node
                while(edges->hasNext()){
                    edge e = edges->next();
                    if(graph->target(e) == to){
                        ++count;
                        if(count > 1){
                            //More than one edge invalid structure.
                            moreThanOneEdge = true;
                            break;
                        }
                    }
                }
                delete edges;
                if(moreThanOneEdge){
                    return false;
                }
            }
        }
        return true;
}
