/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <tulip/ForEach.h>
#include <tulip/StringProperty.h>
#include <tulip/BooleanProperty.h>
#include <tulip/LayoutProperty.h>

#include <portgraph/PortNode.h>
#include <portgraph/PortGraph.h>
#include <portgraph/Port.h>
#include <portgraph/Bridge.h>

using namespace std;
using namespace tlp;

PortGraph::PortGraph (const set<node> v, Graph *g):_graph(g){

    //récupération des sommets centraux
    for(set<node>::const_iterator it=v.begin();it!=v.end();++it) {
        if(PortNodeBase::isCenter(*it,_graph))
            portnode.push_back(new PortNode(this, *it));
    }

    node n;
    Iterator<PortNode *> *it = getPortNodes();
    while(it->hasNext()) {
        PortNode *pn = it->next();
        Iterator<Port::PortPtr> *it2 = pn->getPorts();
        while(it2->hasNext()) {
            Port::PortPtr p=it2->next();
            forEach(n, _graph->getInOutNodes(p->getNode())) {
                if(PortBase::isPort(n, _graph))
                    p->connectPortTo(getPortContainer(n));
            }
        }delete it2;
    }delete it;
}

PortGraph::PortGraph (Graph *g):_graph(g){
    node n;
    //récupération des sommets centraux
    forEach(n, g->getNodes()){
        if(PortNodeBase::isCenter(n,_graph))
            portnode.push_back(new PortNode(this, n));
    }
    //tous les portnodes sont créés. je cherche les ports connectés
    Iterator<PortNode *> *it = getPortNodes();
    while(it->hasNext()) {
        PortNode *pn=it->next();
        Iterator<Port::PortPtr> *it2 = pn->getPorts();
        while(it2->hasNext()) {
            Port::PortPtr p = it2->next();
            forEach(n, _graph->getInOutNodes(p->getNode())) {
                if(PortBase::isPort(n, _graph))
                    p->connectPortTo(getPortContainer(n));
            }
        }delete it2;
    }delete it;
}

PortGraph::~PortGraph() {
    PortNode *pn;
    stableForEach(pn, getPortNodes()) {
        delete pn;
    }
}

PortNode* PortGraph::getOnePortNode() const {
    return *(portnode.begin());
}

Iterator<PortNode *> *PortGraph::getPortNodes(BooleanProperty *ban) const {     //renvoie un itérateur sur l'ensemble des portnodes
    dequeportnode p;
    if(ban==NULL)
        p=portnode;
    else {
        for(dequeportnode::const_iterator it=portnode.begin();it!=portnode.end();++it) {
            if(!ban->getNodeValue((*it)->getCenter()))
                p.push_back(*it);
        }
    }
    return new CopyStlIterator<PortNode *, dequeportnode>(p);
}

bool PortGraph::isCenter(const node n) const {
    PortNode *pn;
    forEach(pn, getPortNodes()) {
        if(pn->getCenter()==n.id)
            return true;
    }
    return false;
}

bool PortGraph::isPort(const node n) const {

    Iterator<PortNode *> *it=getPortNodes();
    while(it->hasNext()) {
        PortNode *pn = it->next();
        Port::PortPtr p;
        forEach(p, pn->getPorts()) {
            if(p->getNode()==n.id){
                delete it;
                return true;
            }
        }
    }delete it;
    return false;
}

//static method
bool PortGraph::isPortGraphEdge(const edge e, Graph *g) {
    const pair<node, node> &ends = g->ends(e);
    return (PortBase::isPort(ends.first, g)&&PortBase::isPort(ends.second, g)&&!Bridge::isBridge(ends.second, g)&&!Bridge::isBridge(ends.first, g));
}

void PortGraph::addPortContainer(const node &n, Port::PortPtr p) {
    container_port[n] = p;
}

const Port::PortPtr PortGraph::getPortContainer(const node &n) const {
    std::map<tlp::node, Port::PortPtr>::const_iterator it = container_port.find(n);
    assert(it!=container_port.end());
    return it->second;
}

const pair<PortNode*, PortNode*> PortGraph::ends(const tlp::edge e) const {
    const pair<node, node> &ends = _graph->ends(e);
    return make_pair(getContainer<PortNode*>(ends.first), getContainer<PortNode*>(ends.second));
}

PortNode* PortGraph::getPortNode(const tlp::node n) const {
    return Bridge::isBridge(n, _graph)?NULL:getContainer<PortNode *>(n);
}

void PortGraph::drawSameSizes(SizeProperty *NodeSize, LayoutProperty *layout) const {
    map <string, pair<unsigned, unsigned> > optimalSizes;
    pair<unsigned, unsigned> pair1;
    pair<unsigned, unsigned> pair2;
    PortNode *pn;
    forEach(pn, getPortNodes()) {
        if(Bridge::isBridge(pn->getCenter(), _graph))
            continue;
        if (!pn->getName().empty()) {
            pair1 = pn->getOptimalSize(layout);
            pair2 = optimalSizes[pn->getName()];

            pair2.first  = max(pair1.first,  pair2.first);
            pair2.second = max(pair1.second, pair2.second);
            optimalSizes[pn->getName()] = pair2;
        }
    }
    forEach(pn, getPortNodes()) {
        if(Bridge::isBridge(pn->getCenter(), _graph)) {
            continue;
        }
        if (optimalSizes.end() != optimalSizes.find(pn->getName())){ // found
            pn->setPositions(optimalSizes[pn->getName()].first, optimalSizes[pn->getName()].second, layout);
            pn->resizeCenter(optimalSizes[pn->getName()].first, optimalSizes[pn->getName()].second, NodeSize);
        }
        else {
            pn->setPositions(layout);
            pn->resizeCenter(NodeSize, layout);
        }
    }
}

bool PortGraph::redraw(const std::string& plugin, tlp::DataSet& parameters ,std::string &errMsg,tlp::PluginProgress* progress){
    bool res=false;
    std::list<std::string> plugins=layoutPluginsList();
    if(std::find(plugins.begin(),plugins.end(),plugin) == plugins.end()){
        errMsg = "The choosen plugin is not in the list";
    }else{
        if(getNumPortNode()!=1) {   //la règle a plus qu'un bridge
            Observable::holdObservers();
            res = _graph->applyPropertyAlgorithm(plugin, _graph->getProperty<LayoutProperty>("viewLayout"), errMsg,progress,&parameters);
            Observable::unholdObservers();
        }
        else{
            errMsg = "The rule is empty";
        }
    }
    return res;
}
