/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <tulip/StringProperty.h>
#include <tulip/BooleanProperty.h>
#include <tulip/LayoutProperty.h>
#include <tulip/ForEach.h>

#include <portgraph/PortGraphRule.h>
#include <portgraph/PortNode.h>
#include <portgraph/Bridge.h>
#include <portgraph/PorgyTlpGraphStructure.h>

#include <boost/lexical_cast.hpp>

using namespace std;
using namespace tlp;

PortGraphRule::PortGraphRule (Graph *g):PortGraph(g),bridge(NULL){
#ifndef NDEBUG
    string errMsg;
    if(!PorgyTlpGraphStructure::isPorgyGraph(g, errMsg)) {
        cerr << "Fatal Error: " << errMsg << endl;
        assert(false);
    }
#endif
    assert(PorgyTlpGraphStructure::isRuleGraph(g));

    node n;    
        forEach(n, g->getNodes()) {            
            if(PortNodeBase::isCenter(n,getGraph())&& Bridge::isBridge(n, getGraph())) {
                bridge = new Bridge(this, n);                
                break;
            }
        }
        //Be sure to find the bridge
        assert(bridge != NULL);
}

std::pair<tlp::Graph *,tlp::Graph *> PortGraphRule::extractLHSRHS(Graph *g, PluginProgress *pp) {
    IntegerProperty *side = getSideProperty(g);
    DataSet parameters;
    parameters.set("Property",side);
    StringCollection elementsType("nodes;edges;");
    elementsType.setCurrent("nodes");
    parameters.set("Type",elementsType);
    parameters.set("Connected",false);
    string err;
    g->applyAlgorithm("Equal Value",err,&parameters,pp);

    Graph *sg, *rhs = NULL, *lhs = NULL;
    forEach(sg, g->getSubGraphs()) {
        if(sg->getName()==PorgyConstants::RULE_SIDE+": "+boost::lexical_cast<string>(PorgyConstants::SIDE_RIGHT))
            rhs = sg;
        if(sg->getName()==PorgyConstants::RULE_SIDE+": "+boost::lexical_cast<string>(PorgyConstants::SIDE_LEFT))
            lhs = sg;
    }
    return make_pair(lhs, rhs);
}

void PortGraphRule::cleanLHSRHS(Graph *r) {
    Graph *g;
    stableForEach(g, r->getSubGraphs()) {
        r->delSubGraph(g);
    }
}

PortGraphRule::~PortGraphRule() {
    delete bridge;
}

tlp::BooleanProperty *PortGraphRule::getSelectionProperty() {
    return PorgyTlpGraphStructure::getRulesRoot(getGraph())->getLocalProperty<BooleanProperty>("viewSelection");
}

node PortGraphRule::findRightEquiv(const node n) const {
    return bridge->getRight(getPortContainer(n));
}

Iterator<PortNode *> *PortGraphRule::getMember(const PorgyConstants::RuleSide side) {
    IntegerProperty *sideProp = getSideProperty();
    vector<PortNode *> v;
    PortNode *pn;
    forEach(pn, PortGraph::getPortNodes()) {
        if(sideProp->getNodeValue(pn->getCenter())==side)
            v.push_back(pn);
    }
    return new CopyStlIterator<PortNode *, vector<PortNode *> >(v);
}

tlp::Iterator<PortNode *> *PortGraphRule::getPortNodes(BooleanProperty *) const {
    IntegerProperty *sideProp = getSideProperty();
    vector<PortNode *> v;
    PortNode *pn;
    forEach(pn, PortGraph::getPortNodes()) {
        if((sideProp->getNodeValue(pn->getCenter())==PorgyConstants::SIDE_RIGHT)||(sideProp->getNodeValue(pn->getCenter())==PorgyConstants::SIDE_LEFT))
            v.push_back(pn);
    }
    return new CopyStlIterator<PortNode *, vector<PortNode *> >(v);
}

Iterator<edge> *PortGraphRule::getEdges(tlp::BooleanProperty *) {
    IntegerProperty *sideProp = getSideProperty();
    vector<edge> v;
    edge e;
    forEach(e, getGraph()->getEdges()) {
        if(((sideProp->getEdgeValue(e)==PorgyConstants::SIDE_RIGHT)||(sideProp->getEdgeValue(e)==PorgyConstants::SIDE_LEFT))&&(PortGraph::isPortGraphEdge(e, getGraph()))) {
            v.push_back(e);
        }
    }
    return new CopyStlIterator<edge, vector<edge> >(v);
}

Iterator<edge> *PortGraphRule::getMemberEdges(const PorgyConstants::RuleSide side) {
    IntegerProperty *sideProp = getSideProperty();
    vector<edge> v;
    edge e;
    forEach(e, getGraph()->getEdges()) {
        if(side==PorgyConstants::SIDE_BRIDGE) {
            if((sideProp->getEdgeValue(e)==side)) { //useless?
                v.push_back(e);
            }
        }
        if((sideProp->getEdgeValue(e)==side)&&(PortGraph::isPortGraphEdge(e, getGraph()))) {
            v.push_back(e);
        }
    }
    return new CopyStlIterator<edge, vector<edge> >(v);
}

//static method
bool PortGraphRule::isMember(const node n, Graph *g, const PorgyConstants::RuleSide &sidestr) {
    assert(PorgyTlpGraphStructure::isRuleGraph(g));
    assert(g->isElement(n));
    return getSideProperty(g)->getNodeValue(n)==sidestr;
}

StringCollection PortGraphRule::getLayoutAlgorithm() {
    string layout_algo;
    getGraph()->getAttribute(PorgyConstants::LAYOUTALGORITHM,layout_algo);
    StringCollection tmp(LAYOUT_ALGO);
    for(unsigned int i=0;i<tmp.size();++i) {
        if(tmp.at(i)==layout_algo) {
            tmp.setCurrent(i);
            break;
        }
    }
    return tmp;
}

//static method
bool PortGraphRule::isMember(const edge e, Graph *g, const PorgyConstants::RuleSide &sidestr) {
    assert(PorgyTlpGraphStructure::isRuleGraph(g));
    assert(g->isElement(e));
    return getSideProperty(g)->getEdgeValue(e)==sidestr;

}

std::list<std::string> PortGraphRule::layoutPluginsList()const{
    std::list<std::string> list;
    list.push_back(PorgyConstants::REDRAW_RULE);
    return list;
}
