/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <tulip/ForEach.h>
#include <tulip/SizeProperty.h>
#include <tulip/LayoutProperty.h>
#include <tulip/ColorProperty.h>
#include <tulip/BooleanProperty.h>

#include <portgraph/PortGraph.h>
#include <portgraph/PortNode.h>
#include <portgraph/PorgyConstants.h>
#include <portgraph/Bridge.h>

using namespace std;
using namespace tlp;

PortNode::PortNode(PortGraph *pg, const node n):PortNodeBase(pg,n) {
    assert(pg->getGraph()->isElement(n));
    node v;
    pg->addContainer(n, this);
    //trouver tous les ports
    forEach(v, pg->getGraph()->getInOutNodes(n)){
        assert(PortBase::isPort(v, pg->getGraph()));
        Port::PortPtr p(new Port(v, this));
        ports.push_back(p);
        pg->addContainer(v, this);
        pg->addPortContainer(v, p);

    }
}

PortNodeBase::PortNodeType PortNode::getType() {
    return PortNodeBase::PORTNODE;
}


std::string PortNode::getState() const {
    return PorgyTlpGraphStructure::getState(ParentPortGraph->getGraph(),centerNode);
}

tlp::Iterator<const PortNode *> *PortNode::getInOutConnectedPortNodes() const {
    vector<const PortNode *> v;
    Port::PortPtr p;
    forEach(p, getPorts()) {
        Iterator<Port::PortPtr> *itp = p->getConnectedPorts();
        while(itp->hasNext()) {
            Port::PortPtr p1 = itp->next();
            v.push_back(p1->getParentPortNode());
        }
        delete itp;
    }

    return new CopyStlIterator<const PortNode *, vector<const PortNode *> >(v);
}


///////////////////////////////
unsigned PortNode::indeg() const{
    unsigned degree=0;

    Port::PortPtr p;
    forEach(p, getPorts()) {
        degree+=ParentPortGraph->getGraph()->indeg(p->getNode())-1;
    }
    return degree;
}

///////////////////////////////
unsigned PortNode::outdeg() const{
    unsigned degree=0;

    Port::PortPtr p;
    forEach(p, getPorts()) {
        degree+=ParentPortGraph->getGraph()->outdeg(p->getNode());
    }
    return degree;
}

///////////////////////////////
unsigned PortNode::deg() const{
    return indeg()+outdeg();
}

///////////////////////////////
void PortNode::setOptimalSize(const int sizeRate, node n, SizeProperty *nodeSize) const{
    float xrate = .4f;
    float yrate = 1.0f;
    float xwidth = max(xrate, yrate);
    nodeSize->setNodeValue(n, Size(sizeRate* xwidth, sizeRate* yrate, 1));
}

void PortNode::draw(SizeProperty *nodeSize, ColorProperty *viewColor, LayoutProperty *layout, tlp::ColorProperty *) const {
    // setOptimalSize(PortNode::sizeRate, centerNode, nodeSize);
    Port::PortPtr p;
    forEach(p, getPorts()) {
        setOptimalSize(PortNode::sizeRate, p->getNode(), nodeSize);
        p->setPortAlphaValue(viewColor);
    }
    setPositions(layout);
    resizeCenter(nodeSize, layout);
    setCenterAlphaValue(viewColor);
    hidePortNodeEdges(viewColor);
}

bool PortNode::check(const PortNode * rule_portnode, const PortNode * modele_portnode, const bool exact) {

#ifndef NDEBUG
    cerr << __PRETTY_FUNCTION__ << ": model portnode: " << modele_portnode->getCenter() << endl;
    cerr << __PRETTY_FUNCTION__ << ": rule portnode to compare: " << rule_portnode->getCenter() << endl;
#endif

    if(rule_portnode->getName() == modele_portnode->getName()) {
        if(rule_portnode->getState() == modele_portnode->getState()) {
            //number of ports
            if(rule_portnode->nbPorts()==modele_portnode->nbPorts()) {
                if((exact&&rule_portnode->deg() == modele_portnode->deg())||((!exact)&&(rule_portnode->deg() <= modele_portnode->deg()))) {
                    return true;
                }
#ifndef NDEBUG
                else {
                    cerr << __PRETTY_FUNCTION__ << ": portnode degrees problem" << endl;
                    cerr << __PRETTY_FUNCTION__ << ": model portnode degree: " << modele_portnode->deg() << " and rule portnode degree: " << rule_portnode->deg() << endl;
                }
#endif
            }
#ifndef NDEBUG
            else
                cerr << __PRETTY_FUNCTION__ << ": problem on the number of ports: "<<rule_portnode->nbPorts() <<" for the rules and " << modele_portnode->nbPorts() << " for the model"<< endl;
#endif
        }
#ifndef NDEBUG
        else {
            cerr << __PRETTY_FUNCTION__ << ": problem on the state of the centers" << endl;
            cerr << __PRETTY_FUNCTION__ << ": model portnode: " << modele_portnode->getState() <<" and rule portnode " << rule_portnode->getState() << endl;
        }
#endif
    }
#ifndef NDEBUG
    else {
        cerr << __PRETTY_FUNCTION__ << ": centers labelling problem" << endl;
    }
    cerr << __PRETTY_FUNCTION__ << ": model portnode: " << modele_portnode->getCenter() <<" and rule portnode " << rule_portnode->getCenter() << " not compatible" << endl;
#endif
    return false ;
 }

pair<unsigned,unsigned> PortNode::getOptimalSize(LayoutProperty *layout) const { // Call a draw_position step1 first!
    return drawPosition_aux(false, layout);
}

pair<unsigned,unsigned> PortNode::drawPosition_aux(const bool draw,  LayoutProperty *layout, const bool specifyWidths, const int xWidth, const int yWidth) const{
    assert(specifyWidths == false || (specifyWidths == true && draw == true));
    vector<Port::PortPtr> northPorts, westPorts, eastPorts, southPorts;
    vector<Port::PortPtr> freePorts;

    Port::PortPtr port;
    forEach(port, getPorts()) {
        //if any, take an arbitrary attached PortNode
        node neighbour;
        const PortNode *p=port->getOneConnectedPortNode();
        if (p != NULL){
            neighbour = p->centerNode; // take the center node
        }

        if(neighbour.isValid()) {
            Coord myPos = layout->getNodeValue(centerNode);
            Coord neighbourPos = layout->getNodeValue(neighbour);
            double deltaX = neighbourPos[0] - myPos[0];
            double deltaY = neighbourPos[1] - myPos[1];
            double ratio = 0;

            if(deltaX == 0)
                ratio = HUGE_VAL; // +INF
            else
                ratio = fabs(deltaY / deltaX);
            //NORTH
            if (ratio >= 1 && deltaY >= 0)
                northPorts.push_back(port);
            //EAST
            else if( ratio < 1 && deltaX >= 0 )
                eastPorts.push_back(port);
            //SOUTH
            else if( ratio >= 1 && deltaY < 0 )
                southPorts.push_back(port);
            //WEST
            else if( ratio < 1 && deltaX < 0 )
                westPorts.push_back(port);
        }else
            freePorts.push_back(port);
    }

    if (draw == true){
        //sort the four vectors by X or Y coord, in reversed or normal order
        if(!northPorts.empty())
            mySortPorts(northPorts, true, false, layout);
        if(!westPorts.empty())
            mySortPorts(westPorts,  false, true, layout);
        if(!southPorts.empty())
            mySortPorts(southPorts, true, false, layout);
        if(!eastPorts.empty())
            mySortPorts(eastPorts,  false, true, layout);
    }
    //add the free ports with attempting to equilibriate nr of ports per dimension
    for(unsigned i = 0; i < freePorts.size(); ++i){
        unsigned sumX = northPorts.size()+ southPorts.size();
        unsigned sumY = westPorts.size() + eastPorts.size();

        unsigned minX = min(northPorts.size(), southPorts.size());
        unsigned minY = min(westPorts.size() , eastPorts.size());

        //put free ports into west or east
        if(minX < minY || (minX == minY && sumX < sumY)){
            if(northPorts.size() > southPorts.size())
                southPorts.push_back(freePorts[i]);
            else
                northPorts.push_back(freePorts[i]);
        }
        else { //put free ports into south or north
            if(westPorts.size() > eastPorts.size())
                eastPorts.push_back(freePorts[i]);
            else
                westPorts.push_back(freePorts[i]);

        }
    }
    if (draw == false)
        return pair<unsigned, unsigned>(max(northPorts.size(), southPorts.size()),
                                        max(westPorts.size(), eastPorts.size()));

    //effectively draw ports
    if (specifyWidths == true)
        placePorts(layout, northPorts, westPorts, southPorts, eastPorts, xWidth, yWidth);
    else
        placePorts(layout, northPorts, westPorts, southPorts, eastPorts);
    return pair<unsigned, unsigned>(0,0);
}

void PortNode::setPositions(const int xWidth, const int yWidth, LayoutProperty *layout) const {
    drawPosition_aux(true, layout, true, xWidth, yWidth);
}

void PortNode::setPositions(LayoutProperty *layout) const {
    drawPosition_aux(true, layout);
}

// helper sorting method
void PortNode::mySortPorts(vector<Port::PortPtr>  & sidePorts, const bool x, const bool reverse, LayoutProperty *layout) const{
    bool b = false;

    Coord myCenter = layout->getNodeValue(centerNode);
    float delta = 0, delta_next = 0;

    while ( !b && sidePorts.size() > 1)
    {
        b = true;
        for (unsigned int i = 0; i < sidePorts.size() - 1; ++i)
        {
            node c1 = sidePorts[i]->getOneConnectedPortNode()->centerNode;
            node c2 = sidePorts[i+1]->getOneConnectedPortNode()->centerNode;

            if (c1 == c2){ // Connected to the same PortNode; puts the ports connected to main ports in c1 and c2
                c1 = (*(sidePorts[i]->connectedPortsBegin()))->getNode();
                c2 = (*(sidePorts[i+1]->connectedPortsBegin()))->getNode();
            }

            Coord pos = layout->getNodeValue(c1);
            Coord pos_next = layout->getNodeValue(c2);

            if(x)//NORTH or SOUTH -> DeltaX/DeltaY
            {
                delta = (myCenter[0] - pos[0]) / (myCenter[1] - pos[1]);
                delta_next = (myCenter[0] - pos_next[0]) / (myCenter[1] - pos_next[1]);
            }
            else//WEST or EAST -> DeltaY/DeltaX
            {
                delta = (myCenter[1] - pos[1]) / (myCenter[0] - pos[0]);
                delta_next = (myCenter[1] - pos_next[1]) / (myCenter[0] - pos_next[0]);
            }

            if ( (!reverse && delta > delta_next) || (reverse && delta < delta_next))
            {
                Port::PortPtr tmp = sidePorts[i];
                sidePorts[i] = sidePorts[i + 1];
                sidePorts[i + 1] = tmp;
                b = false;
            }
        }
    }
}

void PortNode::resizeCenter(SizeProperty *nodeSize, LayoutProperty *layout) const {
    // trouve les coordonnées de ports les plus éloignés du centre en abcisse et ordonnée
    Coord myPos = layout->getNodeValue(centerNode);
    float max_diff_y = 0, max_diff_x = 0;
    Size lg_x;
    Port::PortPtr p;
    forEach(p, getPorts()) {
        Coord pos = layout->getNodeValue(p->getNode());
        if(fabs(pos[1]-myPos[1])>max_diff_y) {
            max_diff_y=fabs(pos[1]-myPos[1]);
        }
        if(fabs(pos[0]-myPos[0])>max_diff_x) {
            max_diff_x=fabs(pos[0]-myPos[0]);
            lg_x = nodeSize->getNodeValue(p->getNode());
        }
    }
    Size centersize = nodeSize->getNodeValue(centerNode);

    //le +1 c'est pour la taille du port
    centersize.setH(2.0f*max_diff_y+1);
    centersize.setW(2.0f*max_diff_x+lg_x.getW());
    if(centersize.getW()==0)
        centersize.setW(1);
    centersize.setD(1); //obligatoire en 2D

    nodeSize->setNodeValue(centerNode,centersize);
}

void PortNode::resizeCenter(const double xWidth, const double yWidth, SizeProperty *nodeSize) const { // Bound to size max of placePorts
    nodeSize->setNodeValue(centerNode, Size(max(xWidth, .5)*PorgyConstants::SPACE +2,max(yWidth, .5)*PorgyConstants::SPACE +2,1));
    // 2 <=> 1 for decalage (offset) + 1 for PortSize
}

////////////////////////////
void PortNode::placePorts(LayoutProperty *layout, vector<Port::PortPtr> & northPorts, vector<Port::PortPtr> & westPorts,
                          vector<Port::PortPtr> & southPorts, vector<Port::PortPtr> & eastPorts) const{
    unsigned nbX = max(northPorts.size(), southPorts.size());
    unsigned nbY = max(westPorts.size(), eastPorts.size());
    placePorts(layout, northPorts, westPorts, southPorts, eastPorts, nbX, nbY);
}

void PortNode::placePorts(LayoutProperty *layout, vector<Port::PortPtr> & northPorts, vector<Port::PortPtr> & westPorts,
                          vector<Port::PortPtr> & southPorts, vector<Port::PortPtr> & eastPorts,
                          const double Xwidth, const double Ywidth) const{
    double width  = max(Xwidth, .5);
    double height = max(Ywidth, .5);
    // Think to modify recizeCenter in the same time
    assert(width >= 0 && height >= 0);
    width *= PorgyConstants::SPACE;
    height *= PorgyConstants::SPACE;

    //distance between first(last) port and corner
    double decalage = 0.5; // +PortNode::space -SizePort
    //CALCULATE DESIRED WIDTHS

    //coords of central node
    Coord center = layout->getNodeValue(centerNode);
    //PLACE PORTS
    double ratio = 0;
    node n;
    Coord newC;
    newC.setZ(0);

    //NORTH
    if(northPorts.size() != 0) {
        ratio = width / northPorts.size();
        newC.setY(center.getY() + height/2.0 + decalage);
        double X = center.getX() - width/2.0 + ratio/2.0;
        for(unsigned i = 0; i < northPorts.size(); ++i){
            n = northPorts[i]->getNode();
            newC.setX(X+i*ratio);
            layout->setNodeValue(n, newC);
        }
    }
    //EAST
    if(eastPorts.size() != 0) {
        ratio = height / eastPorts.size();
        newC.setX(center.getX() + width/2.0 + decalage);
        double Y = center.getY() + height/2.0 - ratio/2.0;
        for(unsigned i = 0; i < eastPorts.size(); ++i){
            n = eastPorts[i]->getNode();
            newC.setY(Y-i*ratio);
            layout->setNodeValue(n, newC);
        }
    }
    //SOUTH
    if(southPorts.size() != 0) {
        ratio = width / southPorts.size();
        double X = center.getX() + width/2.0 - ratio/2.0;
        newC.setY(center.getY() - height/2.0 - decalage);
        for(unsigned i = 0; i < southPorts.size(); ++i){
            n = southPorts[i]->getNode();
            newC.setX(X-i*ratio);
            layout->setNodeValue(n, newC);
        }
    }
    //WEST
    if(westPorts.size() != 0) {
        ratio = height / westPorts.size();
        newC.setX(center.getX() - width/2.0 - decalage);
        double Y = center.getY() - height/2.0 + ratio/2.0;
        for(unsigned i = 0; i < westPorts.size(); ++i){
            n = westPorts[i]->getNode();
            newC.setY(Y+i*ratio);
            layout->setNodeValue(n, newC);
        }
    }
}


//select/deselect the PortNode
void PortNode::select(const bool state) const {
    BooleanProperty *selection = getParentPortGraph()->getSelectionProperty();
    select(state,selection);
}

void PortNode::select(const bool state,BooleanProperty* resultProperty) const{
    Graph *g =    getParentPortGraph()->getGraph();
    Observable::holdObservers();
    if(!Bridge::isBridge(centerNode, g)) {
        node n;
        edge e;
        resultProperty->setNodeValue(centerNode, state);
        forEach(n, g->getInOutNodes(centerNode)) {
            resultProperty->setNodeValue(n, state);
        }
        forEach(e, g->getInOutEdges(centerNode)) {
            resultProperty->setEdgeValue(e, state);
        }
    }
    Observable::unholdObservers();
}
