/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <tulip/ColorProperty.h>
#include <tulip/StringProperty.h>
#include <tulip/IntegerProperty.h>
#include <tulip/BooleanProperty.h>
#include <tulip/DoubleProperty.h>
#include <tulip/LayoutProperty.h>
#include <tulip/ForEach.h>
#include <tulip/TulipViewSettings.h>

#include <portgraph/Trace.h>
#include <portgraph/PorgyTlpGraphStructure.h>
#include <portgraph/PortGraphModel.h>

using namespace tlp;
using namespace std;

// viewBorderWidth
class ViewborderWidthCalculator :public DoubleMinMaxProperty::MetaValueCalculator {
public:
  virtual void computeMetaValue(AbstractProperty<DoubleType, DoubleType, NumericProperty>* width, node mN,
                                Graph*, Graph*) {
    // meta node border width is 1
    width->setNodeValue(mN, 1);
  }
};

//Color properties
class ViewColCalculator :public AbstractColorProperty::MetaValueCalculator {
public:
  virtual void computeMetaValue(AbstractColorProperty* color, node mN,
                                Graph*, Graph*) {
    // meta node color is half opaque white
    color->setNodeValue(mN, Color(255, 255, 255, 0));
  }

  virtual void computeMetaValue(AbstractColorProperty* color, edge mE,
                                Iterator<edge>*itE, Graph*) {
    // meta edge color is the color of the first underlying edge
    color->setEdgeValue(mE, color->getEdgeValue(itE->next()));
  }
};

//Integer Properties
class ViewIntCalculator :public IntegerMinMaxProperty::MetaValueCalculator {
public:
  virtual void computeMetaValue(AbstractProperty<IntegerType, IntegerType, NumericProperty>* s, node mN,
                                Graph*, Graph*) {
        if(s->getName()=="viewShape")
            s->setNodeValue(mN, NodeShape::Square);
        else if(s->getName()==PorgyConstants::TYPE_PROPERTY)
            s->setNodeValue(mN, Trace::Model_Node);
  }
};

//String properties
class ViewlabelCalculator :public AbstractStringProperty::MetaValueCalculator {
public:
  virtual void computeMetaValue(AbstractStringProperty* label, node mN,
                                Graph* metagraph, Graph* ) {
    //use the metagraph name
    label->setNodeValue(mN, metagraph->getName());
  }
};

static ViewborderWidthCalculator vbordercalc;
static ViewColCalculator vcolorcalc;
static ViewIntCalculator vintcalc;
static ViewlabelCalculator vlabelCalc;

Trace::Trace(Graph *g):graph(g) {
    assert(PorgyTlpGraphStructure::isTraceGraph(g));
    graph->getProperty<DoubleProperty>("viewBorderWidth")->setMetaValueCalculator(&vbordercalc);
    graph->getProperty<ColorProperty>("viewColor")->setMetaValueCalculator(&vcolorcalc);
    graph->getProperty<IntegerProperty>("viewShape")->setMetaValueCalculator(&vintcalc);
    getTypeProperty()->setMetaValueCalculator(&vintcalc);
    graph->getProperty<StringProperty>("viewLabel")->setMetaValueCalculator(&vlabelCalc);
}

IntegerProperty* Trace::getTransformationRuleProperty()const{
    Graph *traceRoot = PorgyTlpGraphStructure::getTraceRoot(graph);
    bool initProperty = !traceRoot->existLocalProperty(PorgyConstants::TRANSFORMATION_RULE_PROPERTY);
    IntegerProperty* property = traceRoot->getLocalProperty<IntegerProperty>(PorgyConstants::TRANSFORMATION_RULE_PROPERTY);
    if(initProperty){
        property->setAllNodeValue(UINT_MAX);
        property->setAllEdgeValue(UINT_MAX);
    }
    return property;
}
IntegerProperty* Trace::getTransformationInstanceProperty()const{
    Graph *traceRoot = PorgyTlpGraphStructure::getTraceRoot(graph);
    bool initProperty = !traceRoot->existLocalProperty(PorgyConstants::TRANSFORMATION_INSTANCE_PROPERTY);
    IntegerProperty* property = traceRoot->getLocalProperty<IntegerProperty>(PorgyConstants::TRANSFORMATION_INSTANCE_PROPERTY);
    if(initProperty){
        property->setAllNodeValue(UINT_MAX);
        property->setAllEdgeValue(UINT_MAX);
    }
    return property;
}

IntegerProperty* Trace::getTypeProperty()const{
    return PorgyTlpGraphStructure::getTraceRoot(graph)->getLocalProperty<IntegerProperty>(PorgyConstants::TYPE_PROPERTY);
}

BooleanProperty *Trace::getSelectionProperty() {
    return PorgyTlpGraphStructure::getTraceRoot(graph)->getLocalProperty<BooleanProperty>("viewSelection");
}

IntegerProperty* Trace::getTransformationParallelInformationProperty()const{
    return PorgyTlpGraphStructure::getTraceRoot(graph)->getLocalProperty<IntegerProperty>(PorgyConstants::TRANSFORMATION_PARALLEL_INFORMATION_PROPERTY);
}

StringProperty* Trace::getTransformationStrategyProperty()const{
    return PorgyTlpGraphStructure::getTraceRoot(graph)->getLocalProperty<StringProperty>(PorgyConstants::TRANSFORMATION_STRATEGY_PROPERTY);
}

StringProperty* Trace::getTransformationPPropertyNameProperty()const{
    return PorgyTlpGraphStructure::getTraceRoot(graph)->getLocalProperty<StringProperty>(PorgyConstants::TRANSFORMATION_PPROPERTYNAME_PROPERTY);
}

StringProperty* Trace::getTransformationBanPropertyNameProperty() const {
    return PorgyTlpGraphStructure::getTraceRoot(graph)->getLocalProperty<StringProperty>(PorgyConstants::TRANSFORMATION_BANPROPERTYNAME_PROPERTY);
}

DoubleProperty *Trace::getTransformationProbaProperty() const {
    return PorgyTlpGraphStructure::getTraceRoot(graph)->getLocalProperty<DoubleProperty>(PorgyConstants::TRANSFORMATION_PROBA_PROPERTY);
}

bool Trace::redraw(string &errMsg, PluginProgress *pp) {
    Observable::holdObservers();    
    LayoutProperty* layout = graph->getLocalProperty<LayoutProperty>("viewLayout");
    DataSet ds;
    bool res = graph->applyPropertyAlgorithm(PorgyConstants::REDRAW_TRACE, layout, errMsg, pp, &ds);
    Observable::unholdObservers();
    return res;
}

edge Trace::addStrategyTransformationEdge(const node source, const node target,BooleanProperty* p, BooleanProperty *ban,const string& strategyCode) const{
    edge meta_e = addTransformationEdge(source,target,p,ban,Strategy_Transformation);
    //Set strategy specific var
    getTransformationStrategyProperty()->setEdgeValue(meta_e,strategyCode);
    graph->getProperty<ColorProperty>("viewColor")->setEdgeValue(meta_e, PorgyConstants::DERIVATION_STRATEGY_EDGE_COLOR);
    graph->getProperty<ColorProperty>("viewBorderColor")->setEdgeValue(meta_e, PorgyConstants::DERIVATION_STRATEGY_EDGE_COLOR);
    return meta_e;
}

edge Trace::addSimpleTransformationEdge(const node source, const node target,BooleanProperty* p, BooleanProperty *ban,Graph* instanceGraph,const int rule_graph_id)const{
    edge meta_e = addTransformationEdge(source,target,p,ban,Simple_Transformation);
    //Set simple transformation specific var
    //Instance graph can be NULL in case of an edge to failure node
    getTransformationInstanceProperty()->setEdgeValue(meta_e,instanceGraph==NULL?UINT_MAX:instanceGraph->getId());
    getTransformationRuleProperty()->setEdgeValue(meta_e,rule_graph_id);
    getTransformationParallelInformationProperty()->setEdgeValue(meta_e,No_parallel);

    return meta_e;
}

edge Trace::addProbabilisticTransformationEdge(const node source, const node target,BooleanProperty* p, BooleanProperty *ban,Graph* instanceGraph,const int rule_graph_id, const double proba)const{
    edge meta_e = addSimpleTransformationEdge(source, target, p, ban, instanceGraph, rule_graph_id);
    getTypeProperty()->setEdgeValue(meta_e,Probabilistic_Transformation);
    graph->getProperty<ColorProperty>("viewColor")->setEdgeValue(meta_e, PorgyConstants::DERIVATION_EDGE_PROBABILISTIC_OPERATION_COLOR);
    graph->getProperty<ColorProperty>("viewBorderColor")->setEdgeValue(meta_e, PorgyConstants::DERIVATION_EDGE_PROBABILISTIC_OPERATION_COLOR);
    getTransformationProbaProperty()->setEdgeValue(meta_e, proba);

    return meta_e;
}

edge Trace::addFocusTransformationEdge(const node source, const node target,BooleanProperty* p, Graph* instanceGraph, const PorgyConstants::focus_operation op, const bool isBan)const{
    edge meta_e = addTransformationEdge(source,target,isBan?NULL:p,isBan?p:NULL,isBan?Ban_Transformation:Focus_Transformation);
    //Set simple transformation specific var
    //Instance graph can be NULL in case of an edge to a failure node
    getTransformationInstanceProperty()->setEdgeValue(meta_e,instanceGraph==NULL?UINT_MAX:instanceGraph->getId());
    getTransformationParallelInformationProperty()->setEdgeValue(meta_e,No_parallel);
    getTransformationRuleProperty()->setEdgeValue(meta_e,op);

    if(op==PorgyConstants::ONENGB) {
        graph->getProperty<ColorProperty>("viewColor")->setEdgeValue(meta_e, PorgyConstants::DERIVATION_EDGE_PROBABILISTIC_OPERATION_COLOR);
        graph->getProperty<ColorProperty>("viewBorderColor")->setEdgeValue(meta_e, PorgyConstants::DERIVATION_EDGE_PROBABILISTIC_OPERATION_COLOR);
    }

    return meta_e;
}


edge Trace::addTransformationEdge(const node src, const node tgt,BooleanProperty* p, BooleanProperty *ban,TransformationType transformationType)const{
    assert(graph->isElement(src));
    assert(graph->isElement(tgt));
    edge meta_e = graph->addEdge(src,tgt);
    //Generic information
    getTypeProperty()->setEdgeValue(meta_e,transformationType);
    graph->getProperty<IntegerProperty>("viewTgtAnchorShape")->setEdgeValue(meta_e, EdgeExtremityShape::Arrow);
    if(p!=NULL)
        getTransformationPPropertyNameProperty()->setEdgeValue(meta_e,p->getName());
    if(ban!=NULL)
        getTransformationBanPropertyNameProperty()->setEdgeValue(meta_e,ban->getName());
    return meta_e;
}

node Trace::addMetaNode(Graph *content) const {
    node n;
    forEach(n, content->getNodes()) {
        graph->addNode(n);
    }

    edge e;
    forEach(e, content->getEdges()) {
        graph->addEdge(e);
    }

    node meta = graph->createMetaNode(content);
    return meta;
}

node Trace::addFailureNode(const string &name_src, const int rule_graph_id, BooleanProperty *PProperty, BooleanProperty *BanProperty) const {
    ColorProperty *col=graph->getRoot()->getProperty<ColorProperty>("viewColor");

    node src = findNode(name_src);
    //Is failure node already existing?
    node n;
    if(rule_graph_id!=-1) {
        forEach(n, graph->getOutNodes(src)) {
            if(getTransformationRuleProperty()->getNodeValue(n)== rule_graph_id)
                return n;
        }
    }
    n = graph->addNode();
    col->setNodeValue(n, Color(255,0,0,255));
    getTransformationRuleProperty()->setNodeValue(n, rule_graph_id);
    addSimpleTransformationEdge(src,n,PProperty,BanProperty,NULL,rule_graph_id);
    getTypeProperty()->setNodeValue(n,Failure_Node);

    return n;
}

bool Trace::isFailureNode(const node &n) const {
    return (getTypeProperty()->getNodeValue(n)==Failure_Node);
}

bool Trace::isModelNode(const node &n) const {
    return (getTypeProperty()->getNodeValue(n)==Model_Node);
}

string Trace::existEdge(const string &node_src, const string &edge_label) const {
    edge e;
    StringProperty *label=graph->getProperty<StringProperty>("viewLabel");
    forEach(e, graph->getOutEdges(findNode(node_src))) {
        if(label->getEdgeValue(e).compare(edge_label)==0) {
            return label->getNodeValue(graph->target(e));
        }
    }
    return string("");
}

bool Trace::existEdge(const string &node_src, const node node_dst, const string &edge_label) const {
    StringProperty *label=graph->getProperty<StringProperty>("viewLabel");
    //peut pas utiliser existEdge car il peut y avoir plusieurs arêtes entre deux sommets
    edge e;
    forEach(e, graph->getOutEdges(findNode(node_src))) {
        if(graph->target(e)==node_dst)
            if(label->getEdgeValue(e).compare(edge_label)==0)
                return true;
    }

    return false;
}

node Trace::findNode(const string &str) const {
    node n;
    StringProperty *label=graph->getProperty<StringProperty>("viewLabel");
    forEach(n, graph->getNodes()) {
        if(label->getNodeValue(n).compare(str)==0)
            break;
    }
    return n;
}

Graph* Trace::getInstanceForTransformation(edge e)const{
    //Get the source model graph.
    assert(graph->isElement(e));
    assert(graph->isMetaNode(graph->source(e)));
    return graph->getRoot()->getDescendantGraph(getTransformationInstanceProperty()->getEdgeValue(e));
}

Graph* Trace::getRuleForTransformation(tlp::edge e){
    return graph->getRoot()->getDescendantGraph(getTransformationRuleProperty()->getEdgeValue(e));
}
Graph* Trace::getRuleForTransformation(tlp::node n){
    return graph->getRoot()->getDescendantGraph(getTransformationRuleProperty()->getNodeValue(n));
}
string Trace::getRuleNameForTransformation(edge e){
    assert(graph->isElement(e));
    Graph *rule = getRuleForTransformation(e);
    return rule!=NULL?rule->getName():string();
}

bool Trace::getBanFocusPropertyForTransformation(edge e,BooleanProperty* resultProperty,EdgeExtremitiy origin, const bool isBan)const{
    assert(graph->isElement(e));
    assert(resultProperty != NULL);

    string propertyName;
    if(isBan)
        propertyName=getTransformationBanPropertyNameProperty()->getEdgeValue(e);
    else
        propertyName=getTransformationPPropertyNameProperty()->getEdgeValue(e);
    bool found = false;
    node model = origin==SourceNode?graph->source(e):graph->target(e);
    assert(graph->isMetaNode(model));
    Graph* modelGraph = graph->getNodeMetaInfo(model);
    if(!propertyName.empty() && modelGraph->existProperty(propertyName)){
        //Copy property in result.
        *resultProperty = *modelGraph->getProperty<BooleanProperty>(propertyName);
        found = true;
    }
    return found;
}

bool Trace::getLeftMembersForTransformation(edge e,tlp::BooleanProperty* result)const{
    assert(graph->isElement(e));
    //Getting instance graph
    Graph* instanceGraph = getInstanceForTransformation(e);
    bool returnResult=false;
    result->setAllNodeValue(false);
    result->setAllEdgeValue(false);
    if(instanceGraph!=NULL){
        //Copying all instance graph elements into boolean property
        node n;
        forEach(n, instanceGraph->getNodes()) {
                result->setNodeValue(n,true);
        }
        forEach(e, instanceGraph->getEdges()) {
            result->setEdgeValue(e,true);
        }
        returnResult = true;
    }
    return returnResult;
}
bool Trace::getRightMembersForTransformation(edge e,BooleanProperty* result)const{
    assert(graph->isElement(e));
    //Getting target model graph
    node target = graph->target(e);
    assert(graph->isMetaNode(target));
    Graph* targetGraph = graph->getNodeMetaInfo(target);
    //Search for the NEW property containing all the elements created during the rewriting.
    bool returnResult = false;
    if(targetGraph->existProperty(PorgyConstants::NEW)){
        //Copy properties values.
        result->copy(targetGraph->getProperty<BooleanProperty>(PorgyConstants::NEW));
        returnResult = true;
    }
    return returnResult;
}

tlp::Iterator<node> *Trace::getModelsMetaNodes() {
    vector<node> v;
    node n;
    forEach(n, graph->getNodes()) {
        if(graph->isMetaNode(n))
            v.push_back(n);
    }
    return new CopyStlIterator<node, vector<node> >(v);
}

node Trace::nodeForModel(Graph* modelGraph) const{
    node n;
    forEach(n, graph->getNodes()) {
        if(graph->isMetaNode(n) && graph->getNodeMetaInfo(n) == modelGraph){
            return n;
        }
    }
    return node();
}

string Trace::computeStrategy(node from,node to){
    list<edge> path;
    node current = to;
    while(current.isValid()){
        //While there is more nodes to treat
        if(current != from){
            vector<edge> inEdges;
            edge in;
            forEach(in,graph->getInEdges(current)){
                inEdges.push_back(in);
            }
            if(inEdges.empty()){
                //No edge there is a gap in the tree return an empty strategy.
                return "";
            }else if(inEdges.size()==1){
                //Only one in edge case of a rule application
                current = graph->source(inEdges[0]);
                path.push_front(inEdges[0]);
            }else if(inEdges.size()==2){
                //Two in edges case of a strategy application
                edge strategy = getTypeProperty()->getEdgeValue(inEdges[0])==Strategy_Transformation?inEdges[0]:inEdges[1];
                assert(getTypeProperty()->getEdgeValue(strategy)==Trace::Strategy_Transformation);
                current = graph->source(strategy);
                path.push_front(strategy);
            }else{
                //A model cannot have more than 2 in edges.
                return "";
            }

        }else{
            current = node();
        }
    }
    return computeStrategy(path);
}

string Trace::computeStrategy(const list<edge>& path){
    if(path.empty())
        return "";

    node lastNode;
    string strategy;
    for( list<edge>::const_iterator it = path.begin();it!= path.end();++it){
        if((*it).isValid() && !graph->isElement(*it))
            return "";

        //Checks if the last edge target node is the new edge source.
        if(lastNode.isValid() && graph->source(*it) != lastNode){
            return "";
        }
        string command;
        int operationType = getTypeProperty()->getEdgeValue(*it);
        switch(operationType){
        case Simple_Transformation:
            command = getRuleNameForTransformation(*it);            
            break;
        case Focus_Transformation:
            case Ban_Transformation:
            //This is the same treatment for focus and ban transformation only the key word change
            if(operationType == Focus_Transformation){
            command = string("setPos(");
            }else{
                command = string("setBan(");
            }
            switch(getTransformationRuleProperty()->getEdgeValue(*it)){
            case PorgyConstants::CRTGRAPH:
                command.append("CrtGraph");
                break;
            case PorgyConstants::CRTPOS:
                command.append("CrtPos");
                break;
            case PorgyConstants::PROPERTY:
                //Todo invalid tokens implements this functionnality.
                command.append("Property()");
                cerr<< __PRETTY_FUNCTION__ << ":" << __LINE__ << ": Missing details for property"<<endl;
                break;
            case PorgyConstants::PPROPERTY:
                command.append("Prop(");
                if(operationType == Focus_Transformation){
                    command.append(getTransformationPPropertyNameProperty()->getEdgeValue(*it));
                }else{
                    command.append(getTransformationBanPropertyNameProperty()->getEdgeValue(*it));
                }
                command.append(")");
                break;
            case PorgyConstants::NEXTNGB:
                command.append("NextNgb");
                break;
            case PorgyConstants::ALLNGB:
                command.append("AllNgb");
                break;
            case PorgyConstants::ONENGB:
                command.append("OneNgb");
                break;
            default:
                cerr << __PRETTY_FUNCTION__ << ":" << __LINE__ << ": TODO !!!" << endl;
                break;
            }
            command.append(")");
            break;
        case Strategy_Transformation:
            command = getTransformationStrategyProperty()->getEdgeValue(*it);
            break;
        default:
            break;
        }
        if(!command.empty()){
            if(!strategy.empty()){
                strategy.push_back(';');
            }
            strategy.append(command);
        }
        lastNode = graph->target(*it);
    }
    return strategy;
}
string Trace::computeStrategy(tlp::BooleanProperty* selection,node* sourceNode,node* destinationNode){
    graph->push(false);
    Graph* subgraph = graph->addSubGraph(selection);
    string strategy;
    node firstNode;
    node lastNode;
    //Test if the induced subgraph of selected elements is a simple path.
    if(PorgyTlpGraphStructure::graphIsASimplePath(subgraph)){
        //Find the first node.
        node n;
        forEach(n,subgraph->getNodes()){
            Iterator<edge>*inEdges = subgraph->getInEdges(n);
            bool isFirst = !inEdges->hasNext();
            delete inEdges;
            Iterator<edge>* outEdges = subgraph->getOutEdges(n);
            bool isLast = !outEdges->hasNext();
            delete outEdges;
            if(isFirst){
                firstNode = n;
            }
            if(isLast){
                lastNode = n;
            }
            if(firstNode.isValid() && lastNode.isValid()){
                break;
            }
        }
        Trace trace(subgraph);
        strategy = trace.computeStrategy(firstNode,lastNode);
    }
    if(sourceNode != NULL){
        *sourceNode=firstNode;
    }
    if(destinationNode!= NULL){
        *destinationNode=lastNode;
    }

    graph->pop(false);
    return strategy;
}

Trace::TransformationType Trace::getTransformationType(const edge& transformation)const{
    assert(graph->isElement(transformation));
    return static_cast<Trace::TransformationType>(getTypeProperty()->getEdgeValue(transformation));
}

Trace::TraceType Trace::type()const{
    if(graph->attributeExist(TRACETYPEATTRIBUTE)){
        unsigned tt=0;
        graph->getAttribute<unsigned int>(TRACETYPEATTRIBUTE, tt);
        return static_cast<TraceType>(tt);
    }else{
        return DerivationTree;
    }
}

vector<string> Trace::followedPortnodes()const{
    vector<string> v;
    if(graph->attributeExist(PORTNODENAMEATTRIBUTE))
        graph->getAttribute< vector<string> >(PORTNODENAMEATTRIBUTE,v);

   return v;
}
vector<IntegerProperty*> Trace::followedPortnodesProperties(){
    vector<string> portnodes;
    graph->getAttribute< vector<string> >(PORTNODENAMEATTRIBUTE,portnodes);
    vector<IntegerProperty*> properties;
    properties.reserve(portnodes.size());
    for(vector<string>::const_iterator it = portnodes.begin() ; it != portnodes.end() ; ++it){
        properties.push_back(graph->getLocalProperty<IntegerProperty>(*it));
    }
    return properties;
}

IntegerProperty* Trace::followPortNode(const string& portnodeName){
    //Create the new property
    IntegerProperty *prop = graph->getLocalProperty<IntegerProperty>(portnodeName);
    //add nodes
    node n;
    forEach(n, graph->getNodes()) {
        unsigned count=0;
        PortGraphModel pg(graph->getNodeMetaInfo(n));
        Iterator<PortNode *> *it2 = pg.getPortNodes();
        while(it2->hasNext()) {
            PortNode *pn = it2->next();
            if(pn->getName()==portnodeName)
                count++;
        }delete it2;
        prop->setNodeValue(n, count);
    }
    vector<string> portnodes = followedPortnodes();
    portnodes.push_back(portnodeName);
    graph->setAttribute<vector<string> >(PORTNODENAMEATTRIBUTE,portnodes);
    graph->setAttribute<unsigned int>(TRACETYPEATTRIBUTE,Histogram);    
    return prop;
}

IntegerProperty* Trace::depthProperty(){
    if(!graph->existLocalProperty(DEPTHPROPERTY)){
        Observable::holdObservers();
        string err;
        DoubleProperty depth(graph);
        graph->applyPropertyAlgorithm("Dag Level", &depth, err);
        IntegerProperty *depthInteger= graph->getLocalProperty<IntegerProperty>(DEPTHPROPERTY);
        node n;
        forEach(n,graph->getNodes()){
            depthInteger->setNodeValue(n,static_cast<int>(depth.getNodeValue(n)));
        }
        Observable::unholdObservers();
        return depthInteger;
    }else{
        return graph->getLocalProperty<IntegerProperty>(DEPTHPROPERTY);
    }
}

node Trace::rootNode() const{

    node root;
    node n;
    forEach(n,graph->getNodes()){
        Iterator<edge>* inEdges = graph->getInEdges(n);
        bool hasInEdges = inEdges->hasNext();
        delete inEdges;
        if(!hasInEdges){
            //If the node has no in edges this is the root node
            root = n;
            break;
        }
    }
    return root;
}

edge Trace::getEdge(Graph* srcModel, Graph* tgtModel,bool directed)const{
    node src = nodeForModel(srcModel);
    node tgt = nodeForModel(tgtModel);
    return graph->existEdge(src,tgt,directed);
}
