/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef PORGYGRAPHSTRUCTURE_H
#define PORGYGRAPHSTRUCTURE_H

#include <string>

#include <portgraph/PorgyConstants.h>

#include <tulip/Iterator.h>
#include <tulip/StlIterator.h>

namespace tlp {
class BooleanProperty;
}

/**
 * @brief Iterate on a copy of a stl container.
 **/
template<typename VALUE, typename CONTAINER>
class CopyStlIterator: public tlp::Iterator<VALUE> {
    CONTAINER container;
    tlp::StlIterator<VALUE, typename CONTAINER::iterator> stlIterator;

public:
    CopyStlIterator(const CONTAINER &container):container(container), stlIterator(this->container.begin(), this->container.end()) {
    }

    VALUE next() {
        return stlIterator.next();
    }
    ///Return true if it exists a next element
    bool hasNext() {
        return stlIterator.hasNext();
    }
};

struct PorgyTlpGraphStructure {

    static tlp::Graph* newPorgyStructure();

    static void Graph2PortGraph(tlp::Graph *graph, tlp::PluginProgress *pluginProgress);

    static bool isPorgyGraph(const tlp::Graph* graph, std::string&);
    /**
         * @brief Returns the associate root Graph (i.e. TraceRoot, ModelsRoot or RulesRoot)
         * @param graph: a graph of a Porgy graph structure
         */
    static tlp::Graph *getRootGraph(const tlp::Graph* graph);

    static bool isTraceGraph(const tlp::Graph*);

    static inline  tlp::Graph* getModelsRoot(const tlp::Graph* g) {
        tlp::Graph *r = g->getRoot();
        return (r==NULL)?NULL:r->getDescendantGraph(PorgyConstants::GRAPHS_GRAPH_NAME);
    }

    static inline bool isModelGraph(const tlp::Graph* g){
        return getModelsRoot(g)->isDescendantGraph(g);
    }

    static inline tlp::Graph* findModel(const tlp::Graph* g, const std::string &name) {
        return getModelsRoot(g)->getDescendantGraph(name);
    }

    static inline tlp::Graph* getRulesRoot(const tlp::Graph* g) {
        tlp::Graph *r = g->getRoot();
        return (r==NULL)?NULL:r->getDescendantGraph(PorgyConstants::RULES_GRAPH_NAME);
    }

    static inline tlp::Graph* findRule(const tlp::Graph *g, const std::string &name) {
        tlp::Graph *r = getRulesRoot(g);
        return (r==NULL)?NULL:r->getDescendantGraph(name);
    }

    static inline bool isRuleGraph(const tlp::Graph* g){
        tlp::Graph *r = getRulesRoot(g);
        return r!=NULL&&r->isDescendantGraph(g);
    }

    static inline tlp::Graph* getTraceRoot(const tlp::Graph* g){
        tlp::Graph *r = g->getRoot();
        return (r==NULL)?NULL:r->getRoot()->getDescendantGraph(PorgyConstants::TRACE_GRAPH_NAME);
    }

    static bool checkEdgeState(const tlp::Graph *root, const tlp::edge &rule_e, const tlp::edge &graph_e);
    static bool checkEdgeState(const tlp::Graph *root, const std::vector<tlp::edge> &rule_e, const std::vector<tlp::edge> &graph_e, bool exact);

    static std::string getState(tlp::Graph *g, const tlp::node &n);

    //P property set and get
    static void setPPropertyNameAttribute(tlp::Graph *g, const std::string &name);
    static std::string getPPropertyNameAttribute(tlp::Graph *g);

    static void setBanPropertyNameAttribute(tlp::Graph *g, const std::string &name);
    static std::string getBanPropertyNameAttribute(tlp::Graph *g);

    //rulename set and get
    static void setRuleNameForSubGraph(tlp::Graph *g, const std::string &);
    static std::string getRuleNameForSubGraph(tlp::Graph *g);


    /**
         * @brief Find the main trace graph in the given graph hierachy and return it.
  *
  * @param graph A graph of a Porgy graph structure.
         * @return A tlp::Graph * if the traces graph is found.
  **/
    static tlp::Graph *getMainTrace(const tlp::Graph *graph);

    static tlp::Graph * clone(tlp::Graph * old_graph, tlp::DataSet *ds);

    static bool graphIsASimplePath(tlp::Graph* graph);

};
#endif // PORGYGRAPHSTRUCTURE_H
