/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef _PORTGRAPH_H
#define _PORTGRAPH_H

#include <boost/variant.hpp>

#include <portgraph/Port.h>
#include <portgraph/PortNode.h>

namespace tlp {
class Graph;
class SizeProperty;
class LayoutProperty;
}

class QWidget;

class Bridge;

class PortGraph {

    friend class Bridge;
    friend class PortNode;

public:

    typedef std::deque<PortNode *> dequeportnode;

    PortGraph (tlp::Graph *);
    PortGraph (const std::set<tlp::node> v, tlp::Graph *g);
    virtual ~PortGraph();

    virtual tlp::BooleanProperty *getSelectionProperty()=0;

    void addPortContainer(const tlp::node &, Port::PortPtr);
    const Port::PortPtr getPortContainer(const tlp::node &) const;

    bool isPort(const tlp::node n) const;
    bool isCenter(const tlp::node n) const;

    //test pour savoir si le sommet fait partie du portgraphe
    inline bool isPortGraphElement(const tlp::node n) const { return (isCenter(n)||isPort(n)); }
    inline unsigned getNumPortNode() const {return portnode.size();}
    /**
     * @brief getPortNodes
     * @param ban
     * @return an iterator on all portnodes of the graph
     */
    virtual tlp::Iterator<PortNode *> *getPortNodes(tlp::BooleanProperty *ban=NULL) const;
    inline tlp::Graph *getGraph() const {return _graph;}

    void drawSameSizes(tlp::SizeProperty *nodeSize, tlp::LayoutProperty *layout) const;

    static bool isPortGraphEdge(const tlp::edge e, tlp::Graph *g);

    virtual tlp::Iterator<tlp::edge> *getEdges(tlp::BooleanProperty *ban=NULL)=0;

    /**
     * @brief redraw Redraw the model using the given plugin. If errors occurs, they are store in the errMsg variable.
     * The plugin must be in the redrawPlugins list.
     * @param errMsg The error message if the redraw process fail.     
     * @return True if the redraw was done correctly.
     */
    virtual bool redraw(const std::string& plugin, tlp::DataSet& parameters ,std::string &errMsg,tlp::PluginProgress* progress=NULL);
    /**
     * @brief layoutPluginsList returns the list of the plugins capable to redraw a port graph. Call the redraw function with the selected plugins and it's parameters.
     * @return
     */
    virtual std::list<std::string> layoutPluginsList()const=0;

    /**
     * @brief gives the portnode associated to a tulip node
     * @param n The tulip node
     * @return the portnode in which the tulip node takes place
     */
    PortNode *getPortNode(const tlp::node n) const;

    //retourne le premier portnode
    PortNode *getOnePortNode() const;

    const std::pair<PortNode *, PortNode *> ends(const tlp::edge e) const;

protected:

    tlp::Graph *_graph;

    inline void setPortNode(dequeportnode v) {portnode = v;}

    template<typename TYPE>
    void addContainer(const tlp::node &, TYPE);
    template<typename TYPE>
    TYPE getContainer(const tlp::node &) const;

   private:

    dequeportnode portnode;

    std::map<tlp::node, boost::variant<PortNode *, Bridge *> > container;	//relation entre node et PortNode *
    std::map<tlp::node, Port::PortPtr> container_port;   //relation entre node et boost::shared_ptr<Port>
};


template<typename TYPE>
TYPE PortGraph::getContainer(const tlp::node &n) const {
    assert(container.find(n) != container.end());
    return boost::get<TYPE>(container.at(n));
}

template<typename TYPE>
void PortGraph::addContainer(const tlp::node &n, TYPE pn) {
    container[n] = pn;
}

#endif
