/**
 *
 * This file is part of Porgy (https://gforge.inria.fr/projects/porgy/)
 *
 * from LaBRI, University of Bordeaux 1, Inria Bordeaux - Sud Ouest and King's College London
 *
 * Porgy is free software;  you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Porgy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this Porgy.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef TRACE_H_
#define TRACE_H_

#include <portgraph/PorgyConstants.h>

namespace tlp {
class Color;
class BooleanProperty;
class IntegerProperty;
class StringProperty;
class DoubleProperty;
class PluginProgress;
struct node;
struct edge;
template <typename TYPE> struct iterator;
}

const static std::string TRACETYPEATTRIBUTE = "Type";
const static std::string PORTNODENAMEATTRIBUTE = "Portnode";
const static std::string DEPTHPROPERTY = "Depth";

class Trace {

    tlp::Graph *graph;

public:
    enum TransformationType{Simple_Transformation, Probabilistic_Transformation, Strategy_Transformation, Focus_Transformation, Ban_Transformation};
    enum NodeType{Failure_Node,Model_Node};
    enum TransformationParallelInformation{No_parallel,Lazy_Parallel,Real_Parallel};

    enum TraceType{DerivationTree,Histogram};

    Trace(tlp::Graph *);
    bool redraw(std::string &errMsg, tlp::PluginProgress *pp);
    //Access methods
    tlp::node findNode(const std::string &) const;
    bool existEdge(const std::string &, const tlp::node, const std::string &) const;
    inline tlp::Graph *getGraph() const {return graph;}
    std::string existEdge(const std::string &, const std::string &) const;
    tlp::Iterator<tlp::node> *getModelsMetaNodes();
    tlp::BooleanProperty *getSelectionProperty();

    /**
         * @brief getEdge search for an edge between the two given model. Search for the nodes corresponding to the given graphs and return the result of existEdge function of the graph. If there is no edge between the two models return an invalid edge.
         * @param srcModel The source model
         * @param tgtModel The target model
         * @return The edge between the two models or an invalid
         */
    tlp::edge getEdge(tlp::Graph* srcModel, tlp::Graph* tgtModel,bool directed)const;


    /**
         * @brief rootNode searh and return the root node. The root node is the node with no in edges.
         * @return the initial graph node.
         */
    tlp::node rootNode() const;


    /**
          * @brief Find the node corresponding to the given model graph in the trace.
          *
          * @return The node corresponding the model or an invalid node.
          **/
    tlp::node nodeForModel(tlp::Graph* modelGraph) const;

    /**
          * @brief Return the property containing the node type or transformation type for edges.
          *
          * See TransformationType and NodeType for available types.
          **/
    tlp::IntegerProperty* getTypeProperty()const;

    //Properties for simple transformations
    /**
          * @brief Return the property containing the parallel information for the transformation.
          **/
    tlp::IntegerProperty* getTransformationParallelInformationProperty()const;
    /**
          * @brief Return the property containing the rule used for transformations.
          **/
    tlp::IntegerProperty* getTransformationRuleProperty()const;
    /**
          * @brief Return the property containing the instance used for transformations.
          **/
    tlp::IntegerProperty* getTransformationInstanceProperty()const;

    //Properties for strategies transformations
    /**
          * @brief Return the property containing the strategy code for strategy transformations
          **/
    tlp::StringProperty* getTransformationStrategyProperty()const;


    /**
          * @brief Return the property containing the p property name for transformations
          **/
    tlp::StringProperty* getTransformationPPropertyNameProperty()const;
    tlp::StringProperty* getTransformationBanPropertyNameProperty() const;
    tlp::DoubleProperty* getTransformationProbaProperty() const;

    /**
          * @brief Convinience function get the type of the transformation for the transformation edge. The transformation edge must be an element of the trace graph.
          **/
    TransformationType getTransformationType(const tlp::edge& transformation)const;


    /**
          *
          **/
    tlp::Graph* getRuleForTransformation(tlp::edge e);
    tlp::Graph* getRuleForTransformation(tlp::node n);

    /**
          * @brief Find the rule used in the transformation e.
          **/
    std::string getRuleNameForTransformation(tlp::edge e);

    /**
     * @brief Return the instance of the source model used to rewrite it into the destination model.
     **/
    tlp::Graph* getInstanceForTransformation(tlp::edge e)const;

    /**
          * @brief Search the source P property used in the transformation e.
          **/
    inline bool getSourcePPropertyForTransformation(tlp::edge e,tlp::BooleanProperty* result)const{
        return getBanFocusPropertyForTransformation(e,result,SourceNode,false);
    }

    /**
          * @brief Search the target P property used in the transformation e.
          **/
    inline bool getTargetPPropertyForTransformation(tlp::edge e,tlp::BooleanProperty* result)const{
        return getBanFocusPropertyForTransformation(e,result,TargetNode,false);
    }

    /**
          * @brief Search the source Ban property used in the transformation e.
          **/
    inline bool getSourceBanPropertyForTransformation(tlp::edge e,tlp::BooleanProperty* result)const{
        return getBanFocusPropertyForTransformation(e,result,SourceNode,true);
    }

    /**
          * @brief Search the target Ban property used in the transformation e.
          **/
    inline bool getTargetBanPropertyForTransformation(tlp::edge e,tlp::BooleanProperty* result)const{
        return getBanFocusPropertyForTransformation(e,result,TargetNode,true);
    }

    /**
     * @brief Get the left members used in the transformation e and fill the result property with them.
     **/
    bool getLeftMembersForTransformation(tlp::edge e,tlp::BooleanProperty* result)const;
    /**
     * @brief Get the created members with the transformation e.
     **/
    bool getRightMembersForTransformation(tlp::edge e,tlp::BooleanProperty* result)const;

    //Modification methods.
    /**
          * @brief Create a new strategy transformation edge and set properties.
          **/
    tlp::edge addStrategyTransformationEdge(const tlp::node, const tlp::node,tlp::BooleanProperty* p,tlp::BooleanProperty *ban, const std::string &strategyCode) const;
    /**
          * @brief Create a new simple transformation edge and set properties.
          **/
    tlp::edge addSimpleTransformationEdge(const tlp::node, const tlp::node,tlp::BooleanProperty* p,tlp::BooleanProperty *ban,tlp::Graph *instanceGraph, const int rule_graph_id) const;
    tlp::edge addProbabilisticTransformationEdge(const tlp::node, const tlp::node,tlp::BooleanProperty* p,tlp::BooleanProperty *ban,tlp::Graph *instanceGraph, const int rule_graph_id, const double proba) const;


    tlp::edge addFocusTransformationEdge(const tlp::node source, const tlp::node target,tlp::BooleanProperty* p,tlp::Graph *instanceGraph,const PorgyConstants::focus_operation op, const bool isBan)const;

    tlp::node addMetaNode(tlp::Graph *) const;
    /**
          * @brief Add a failure node
          **/
    tlp::node addFailureNode(const std::string &, const int rule_graph_id, tlp::BooleanProperty *PProperty, tlp::BooleanProperty *BanProperty) const;

    bool isFailureNode(const tlp::node &n) const;
    bool isModelNode(const tlp::node &n) const;


    std::string computeStrategy(tlp::node from,tlp::node to);

    /**
          * @brief Compute the strategy for the path given in parameters. If one of the node in the path don't exists or if an edge between two next elements don't exists return an empty string.
          **/
    std::string computeStrategy(const std::list<tlp::edge>& path);

    /**
          * @brief Try to compute the strategy code from a selection. The sourceNode and destinationNode paramters are used to get the source and destination nodes of the computed strategy.
          * If a strategy can be computed from the selection this two values will be filled.
          **/
    std::string computeStrategy(tlp::BooleanProperty* selection,tlp::node* sourceNode = NULL,tlp::node* destinationNode=NULL);

    tlp::IntegerProperty* depthProperty();

    /**
          * @brief Return the type of the graph.
          **/
    TraceType type()const;

    std::vector<std::string> followedPortnodes()const;
    std::vector<tlp::IntegerProperty*> followedPortnodesProperties();

    /**
          * @brief Follow the given portnode : count all
          **/
    tlp::IntegerProperty* followPortNode(const std::string& portnodeName);


private:

    enum EdgeExtremitiy{ SourceNode,TargetNode};

    /**
          * @brief Create an edge between two states and set generals information on newly created edge.
          **/
    tlp::edge addTransformationEdge(const tlp::node, const tlp::node,tlp::BooleanProperty* p,tlp::BooleanProperty *ban, TransformationType transformationType)const;

    /**
         * @brief Search the P property used in the transformation e. You can search P either in source or in destination model using the gPInSourceModel flag.
         * @return The property with the P value or NULL if the transformation don't exists.
         **/
    bool getBanFocusPropertyForTransformation(tlp::edge e,tlp::BooleanProperty* result,EdgeExtremitiy origin, const bool isBan)const;
};

#endif /* TRACE_H_ */
